/*
(c) 2010 Perttu Ahola <celeron55@gmail.com>
*/

#ifndef MAP_HEADER
#define MAP_HEADER

#include <irrlicht.h>
#include <jmutex.h>
#include <jthread.h>
#include <iostream>
#include <malloc.h>

#ifdef _WIN32
	#include <windows.h>
	#define sleep(x) Sleep(x)
#else
	#include <unistd.h>
#endif

#include "heightmap.h"

using namespace irr;
typedef core::vector3df v3f;
typedef core::vector3d<s16> v3s16;
typedef core::vector2d<s16> v2s16;

/*

TODO: Automatically unload blocks from memory and save on disk
	  when they are far away

RULES OF LIGHT:

light = 1.0 (0.999-1.001) is SUNLIGHT. It goes downwards infinitely from
infinite highness and stops when it first hits a non-transparent node.
The lighting of the node it hits is 0.

Other LIGHT SOURCES have a light value of UNDER 0.999.

Light diminishes at a constant factor between nodes.

*/

#define BS 10

#define MATERIALS_COUNT 4

// This is the common material through which the player can walk
// and the light can propagate
#define MATERIAL_AIR 254
// This is completely ignored. It doesn't create faces with anything.
#define MATERIAL_IGNORE 255

//#define MAP_BLOCKSIZE 20
//#define MAP_BLOCKSIZE 32
#define MAP_BLOCKSIZE 16

typedef f32 light_t;

#define LIGHT_MAX 1.0
#define LIGHT_MIN 0.0
// If lighting value is under this, it can be assumed that
// there is no light
#define NO_LIGHT_MAX 0.03

void limitBox(core::aabbox3d<s16> & box, core::aabbox3d<s16> & limits);

#define LIGHTING_RADIUS 10

/*
	MapNode flags
*/
// TODO: none at this moment

struct MapNode
{
	//TODO: block type to differ from material (e.g. grass edges)
	u8 d; // block type
	light_t light;
	//u8 flags;

	MapNode(const MapNode & n)
	{
		*this = n;
	}

	MapNode(u8 data=MATERIAL_AIR, light_t a_light=LIGHT_MIN/*, u8 a_flags=0*/)
	{
		d = data;
		light = a_light;
		/*flags = a_flags;*/
	}

	/*
		If true, the node allows light propagation
	*/
	inline bool transparent()
	{
		return (d == MATERIAL_AIR);
	}
};

// Named by looking towards z+
enum{
	FACE_BACK=0,
	FACE_TOP,
	FACE_RIGHT,
	FACE_FRONT,
	FACE_BOTTOM,
	FACE_LEFT
};

struct FastFace
{
	u8 material;
	/*u16 x;
	u16 y;
	u16 z;*/
	video::S3DVertex vertices[4]; // Precalculated vertices
	//u8 direction; // Index in the precalc buffer
};

struct LightBorder
{
	v3s16 pos;
	v3s16 dir;

	LightBorder(v3s16 &a_pos, v3s16 &a_dir)
	{
		pos = a_pos;
		dir = a_dir;
	}
};

class Map;

class MapBlock
{
	v3s16 size;
	v3s16 posRelative;
	core::array<MapNode> data;
	Map *map;
	bool changed;
public:
	core::list<FastFace*> *fastfaces;
	JMutex fastfaces_mutex;
	
	MapBlock(Map *a_map, v3s16 a_size, v3s16 a_posRelative)
		: posRelative(a_posRelative), map(a_map), changed(true)
	{
		resize(a_size);
		fastfaces = new core::list<FastFace*>;
		fastfaces_mutex.Init();
	}

	inline bool getChangedFlag()
	{
		return changed;
	}

	inline void resetChangedFlag()
	{
		/*v3s16 p = getPosRelative();
		std::cout<<"MapBlock("<<p.X<<","<<p.Y<<","<<p.Z<<")"
				<<"::resetChangedFlag()"
				<<std::endl;*/
				
		changed = false;
	}

	inline void setChangedFlag()
	{
		/*v3s16 p = getPosRelative();
		std::cout<<"MapBlock("<<p.X<<","<<p.Y<<","<<p.Z<<")"
				<<"::setChangedFlag()"
				<<std::endl;*/
				
		changed = true;
	}

	/*// Not used
	inline bool nodeAtEdges(v3s16 p, core::list<v3s16> & edges)
	{
		v3s16 dirs[3] = {
			v3s16(0,0,1), // back
			v3s16(0,1,0), // top
			v3s16(1,0,0), // right
		};
		for(u16 i=0; i<3; i++)
		{
			//f32 nodemax = nodebox.MaxEdge.dotProduct(dirs[i]);
			s16 p0 = p.dotProduct(dirs[i]);
			s16 s0 = getSizeNodes().dotProduct(dirs[i]);
			
			if(p0 == 0)
				edges.push_back(dirs[i] * -1);
			if(p0 == s0 - 1)
				edges.push_back(dirs[i]);
		}

		return (edges.getSize() > 0);
	}*/

	inline v3s16 & getSizeNodes()
	{
		return size;
	}

	inline v3s16 & getPosRelative()
	{
		return posRelative;
	}
		
	inline core::aabbox3d<s16> getBox()
	{
		return core::aabbox3d<s16>(getPosRelative(),
				getPosRelative() + getSizeNodes() - v3s16(1,1,1));
	}
	
	void resize(v3s16 a_size)
	{
		data.clear();
		u32 l = a_size.X * a_size.Y * a_size.Z;
		data.reallocate(l);
		for(u32 i=0; i<l; i++){
			data.push_back(MapNode(MATERIAL_AIR, 1.0));
		}
		size = a_size;
	}

	inline MapNode * getNodePtr(s16 x, s16 y, s16 z)
	{
		if(x < 0 || x >= size.X) throw 1;
		if(y < 0 || y >= size.Y) throw 1;
		if(z < 0 || z >= size.Z) throw 1;
		return &data[z*size.X*size.Y + y*size.X + x];
	}
	
	inline MapNode * getNodePtr(v3s16 p)
	{
		return getNodePtr(p.X, p.Y, p.Z);
	}
	
	inline MapNode & getNode(s16 x, s16 y, s16 z)
	{
		if(x < 0 || x >= size.X) throw 1;
		if(y < 0 || y >= size.Y) throw 1;
		if(z < 0 || z >= size.Z) throw 1;
		return data[z*size.X*size.Y + y*size.X + x];
	}
	
	inline MapNode & getNode(v3s16 p)
	{
		return getNode(p.X, p.Y, p.Z);
	}
	
	inline void setNode(s16 x, s16 y, s16 z, MapNode & n)
	{
		if(x < 0 || x >= size.X) throw 1;
		if(y < 0 || y >= size.Y) throw 1;
		if(z < 0 || z >= size.Z) throw 1;
		data[z*size.X*size.Y + y*size.X + x] = n;
	}
	
	inline void setNode(v3s16 p, MapNode & n)
	{
		setNode(p.X, p.Y, p.Z, n);
	}

	/*
		Uses the Map's function to get a node on any block
		with relative coordinates to this block.
	*/
	MapNode * getNodePtrMap(v3s16 p);

	/*
		Uses the Map's function to set a node on any block
		with relative coordinates to this block.
	*/
	void setNodeMap(v3s16 p, MapNode & n);

	void drawbox(s16 x0, s16 y0, s16 z0, s16 w, s16 h, s16 d, MapNode node)
	{
		for(u16 z=0; z<d; z++)
			for(u16 y=0; y<h; y++)
				for(u16 x=0; x<w; x++)
					setNode(x0+x, y0+y, z0+z, node);
	}

	static FastFace * makeFastFace(u8 material, light_t light, v3f p,
			v3f dir, v3f scale, v3f posRelative_f)
	{
		FastFace *f = new FastFace;
		
		// Position is at the center of the cube.
		v3f pos = p * BS;
		posRelative_f *= BS;

		v3f vertex_pos[4];
		// If looking towards z+, this is the face that is behind
		// the center point, facing towards z+.
		vertex_pos[0] = v3f( BS/2,-BS/2,BS/2);
		vertex_pos[1] = v3f(-BS/2,-BS/2,BS/2);
		vertex_pos[2] = v3f(-BS/2, BS/2,BS/2);
		vertex_pos[3] = v3f( BS/2, BS/2,BS/2);
		
		core::CMatrix4<f32> m;
		m.buildRotateFromTo(v3f(0,0,1), dir);
		
		for(u16 i=0; i<4; i++){
			m.rotateVect(vertex_pos[i]);
			vertex_pos[i].X *= scale.X;
			vertex_pos[i].Y *= scale.Y;
			vertex_pos[i].Z *= scale.Z;
			vertex_pos[i] += pos + posRelative_f;
		}

		f32 abs_scale = 1.;
		if     (scale.X < 0.999 || scale.X > 1.001) abs_scale = scale.X;
		else if(scale.Y < 0.999 || scale.Y > 1.001) abs_scale = scale.Y;
		else if(scale.Z < 0.999 || scale.Z > 1.001) abs_scale = scale.Z;

		v3f zerovector = v3f(0,0,0);
		
		//u8 li = (1.0+light)/2 * 255.0; //DEBUG
		u8 li = light * 255.0;
		//u8 li = light / 128;
		video::SColor c = video::SColor(255,li,li,li);

		//video::SColor c = video::SColor(255,255,255,255);
		//video::SColor c = video::SColor(255,64,64,64);
		
		f->vertices[0] = video::S3DVertex(vertex_pos[0], zerovector, c,
				core::vector2d<f32>(0,1));
		f->vertices[1] = video::S3DVertex(vertex_pos[1], zerovector, c,
				core::vector2d<f32>(abs_scale,1));
		f->vertices[2] = video::S3DVertex(vertex_pos[2], zerovector, c,
				core::vector2d<f32>(abs_scale,0));
		f->vertices[3] = video::S3DVertex(vertex_pos[3], zerovector, c,
				core::vector2d<f32>(0,0));

		f->material = material;
		//f->direction = direction;

		return f;
	}
	
	light_t getFaceLight(v3s16 p, v3s16 face_dir);
	
	/*
		Gets node material from any place relative to block.
		Returns MATERIAL_AIR if doesn't exist.
	*/
	u8 getNodeMaterial(v3s16 p)
	{
		MapNode *n = getNodePtrMap(p);
		if(n == NULL)
			return MATERIAL_IGNORE;
		else
			return n->d;
	}

	/*
		startpos:
		translate_dir: unit vector with only one of x, y or z
		face_dir: unit vector with only one of x, y or z
	*/
	void updateFastFaceRow(v3s16 startpos,
			u16 length,
			v3s16 translate_dir,
			v3s16 face_dir,
			core::list<FastFace*> &dest);

	/*
		Propagates sunlight down through the block.

		If there is a block above, continues from it.

		If there is no block above, assumes there is sunlight.

		Returns true if some sunlight touches the bottom of the block.
	*/

	bool propagateSunlight()
	{
		bool sunlight_touches_bottom = false;

		for(s16 x=0; x<size.X; x++)
		{
			for(s16 z=0; z<size.Z; z++)
			{
				// Check if node above block has sunlight
				MapNode *n = getNodePtrMap(v3s16(x, size.Y, z));
				if(n != NULL){
					if(n->light < 0.999)
						// No sunlight here
						continue;
				}
				
				// Continue spreading sunlight downwards
				s16 y = size.Y-1;
				for(; y >= 0; y--){
					v3s16 pos(x, y, z);
					
					MapNode *n = getNodePtr(pos);
					if(n == NULL)
						break;

					if(n->transparent())
					{
						n->light = 1.0;
					}
					else{
						break;
					}
				}
				if(y == -1)
					sunlight_touches_bottom = true;
			}
		}

		return sunlight_touches_bottom;
	}

	/*
		Find all faces being between material and air
	*/
	void updateFastFaces();

};

class MapUpdateThread : public JThread
{
	bool run;
	JMutex run_mutex;

	Map *map;
	JMutex map_mutex;

public:

	MapUpdateThread() : JThread(), run(true), map(NULL)
	{
		run_mutex.Init();
		map_mutex.Init();
	}

	void * Thread();

	bool getRun()
	{
		run_mutex.Lock();
		bool run_cached = run;
		run_mutex.Unlock();
		return run_cached;
	}
	void setRun(bool a_run)
	{
		run_mutex.Lock();
		run = a_run;
		run_mutex.Unlock();
	}

	void setMap(Map *a_map)
	{
		map_mutex.Lock();
		map = a_map;
		map_mutex.Unlock();
	}
};

class LoadStatus
{
	bool ready;
	JMutex ready_mutex;
	
	u32 done;
	JMutex done_mutex;

	u32 todo;
	JMutex todo_mutex;

	wchar_t *text;
	JMutex text_mutex;

public:

	LoadStatus(bool a_ready=false, u32 a_done=0, u32 a_todo=0)
	{
		ready = a_ready;
		done = a_done;
		todo = a_todo;
		text = NULL;
		ready_mutex.Init();
		done_mutex.Init();
		todo_mutex.Init();
		text_mutex.Init();
	}

	void setReady(bool a_ready)
	{
		ready_mutex.Lock();
		ready = a_ready;
		ready_mutex.Unlock();
	}

	bool getReady(void)
	{
		ready_mutex.Lock();
		bool a_ready = ready;
		ready_mutex.Unlock();
		return a_ready;
	}

	void setDone(u32 a_done)
	{
		done_mutex.Lock();
		done = a_done;
		done_mutex.Unlock();
	}

	u32 getDone(void)
	{
		done_mutex.Lock();
		u32 a_done = done;
		done_mutex.Unlock();
		return a_done;
	}

	void setTodo(u32 a_todo)
	{
		todo_mutex.Lock();
		todo = a_todo;
		todo_mutex.Unlock();
	}

	u32 getTodo(void)
	{
		todo_mutex.Lock();
		u32 a_todo = todo;
		todo_mutex.Unlock();
		return a_todo;
	}

	/*
		Copies the text if not NULL,
		If NULL; sets text to NULL.
	*/
	void setText(const wchar_t *a_text)
	{
		text_mutex.Lock();
		if(text != NULL)
			free(text);
		if(a_text == NULL){
			text = NULL;
			text_mutex.Unlock();
			return;
		}
		u32 len = wcslen(a_text);
		text = (wchar_t*)malloc(sizeof(wchar_t) * (len+1));
		if(text == NULL) throw;
		swprintf(text, len+1, L"%ls", a_text);
		text_mutex.Unlock();
	}
	
	/*
		Return value must be free'd
		Return value can be NULL
	*/
	wchar_t * getText()
	{
		text_mutex.Lock();
		if(text == NULL){
			text_mutex.Unlock();
			return NULL;
		}
		u32 len = wcslen(text);
		wchar_t *b_text = (wchar_t*)malloc(sizeof(wchar_t) * (len+1));
		if(b_text == NULL) throw;
		swprintf(b_text, len+1, L"%ls", text);
		text_mutex.Unlock();
		return b_text;
	}
	
	/*
		Return value must be free'd
	*/
	wchar_t * getNiceText()
	{
		const wchar_t *defaulttext = L"Loading";
		wchar_t *t = getText();
		u32 maxlen = 20; // " (%i/%i)"
		if(t != NULL)
			maxlen += wcslen(t);
		else
			maxlen += wcslen(defaulttext);
		wchar_t *b_text = (wchar_t*)malloc(sizeof(wchar_t) * (maxlen+1));
		if(b_text == NULL) throw;
		if(t != NULL)
			swprintf(b_text, maxlen+1, L"%ls (%i/%i)",
					t, getDone(), getTodo());
		else
			swprintf(b_text, maxlen+1, L"%ls (%i/%i)",
					defaulttext, getDone(), getTodo());
		if(t != NULL)
			free(t);
		return b_text;
	}
};

class Map : public HeightmapSource
{
	/*
		TODO: Dynamic block loading
		Add a filename to the constructor, in which the map will be
		automatically stored
	*/

	core::map<v3s16, MapBlock*> blocks;
	JMutex block_create_mutex;

	v3f camera_position;
	v3f camera_direction;
	JMutex camera_mutex;

	MapUpdateThread updater;

	bool master_heightmap_generated;
	JMutex master_heightmap_generated_mutex;

	bool generateblock_running;

	Heightmap master_heightmap;
	core::map<v2s16, f32> groundheights;
	
public:
	v3s16 drawoffset;

	LoadStatus status;
	
	/*Map(Player *a_player)
		: player(a_player), drawoffset(0,0,0)*/
	Map():
		camera_position(0,0,0),
		camera_direction(0,0,1),
		drawoffset(0,0,0)
	{
		block_create_mutex.Init();
		camera_mutex.Init();

		master_heightmap_generated = false;
		generateblock_running = false;
	
		updater.setMap(this);
		updater.Start();
	}

	~Map()
	{
		/*
			Stop updater thread
		*/
		updater.setRun(false);
		while(updater.IsRunning())
			sleep(1);

		/*
			Free all MapBlocks.
			TODO: Flush to disk
		*/
		core::map<v3s16, MapBlock*>::Iterator i = blocks.getIterator();
		for(; i.atEnd() == false; i++)
		{
			MapBlock *block = i.getNode()->getValue();
			delete block;
		}
	}

	void updateCamera(v3f pos, v3f dir)
	{
		camera_mutex.Lock();
		camera_position = pos;
		camera_direction = dir;
		camera_mutex.Unlock();
	}

	/*
		Returns integer position of the node in given
		floating point position.
	*/
	static inline v3s16 floatToInt(v3f p)
	{
		v3s16 p2(
			(p.X + (p.X>0 ? BS/2 : -BS/2))/BS,
			(p.Y + (p.Y>0 ? BS/2 : -BS/2))/BS,
			(p.Z + (p.Z>0 ? BS/2 : -BS/2))/BS);
		return p2;
	}

	static inline v3f intToFloat(v3s16 p)
	{
		v3f p2(
			p.X * BS,
			p.Y * BS,
			p.Z * BS
		);
		return p2;
	}

	static inline core::aabbox3d<f32> getNodeBox(v3s16 p)
	{
		return core::aabbox3d<f32>(
			(float)p.X * BS - 0.5*BS,
			(float)p.Y * BS - 0.5*BS,
			(float)p.Z * BS - 0.5*BS,
			(float)p.X * BS + 0.5*BS,
			(float)p.Y * BS + 0.5*BS,
			(float)p.Z * BS + 0.5*BS
		);
	}
	
	// Returns the block or NULL if it doesn't already exist
	MapBlock * getBlockNoCreate(v3s16 p);
	// Returns the block; creates one if doesn't exist
	MapBlock * getBlock(v3s16 p);

	/*
		Returns the position of the block where the node is located
	*/
	inline v3s16 getNodeBlockPos(v3s16 p)
	{
		return v3s16(
				(p.X>=0 ? p.X : p.X-MAP_BLOCKSIZE+1) / MAP_BLOCKSIZE,
				(p.Y>=0 ? p.Y : p.Y-MAP_BLOCKSIZE+1) / MAP_BLOCKSIZE,
				(p.Z>=0 ? p.Z : p.Z-MAP_BLOCKSIZE+1) / MAP_BLOCKSIZE);
	}

	inline MapBlock * getNodeBlock(v3s16 p)
	{
		v3s16 blockpos = getNodeBlockPos(p);
		return getBlock(blockpos);
	}

	inline MapNode & getNode(v3s16 p)
	{
		v3s16 blockpos = getNodeBlockPos(p);
		MapBlock *block = getBlock(blockpos);
		if(block == NULL) throw 1;
		v3s16 relpos = p - blockpos*MAP_BLOCKSIZE;
		return block->getNode(relpos);
	}

	inline MapNode & getNode(s16 x, s16 y, s16 z)
	{
		return getNode(v3s16(x,y,z));
	}

	inline MapNode * getNodePtrNoCreate(v3s16 p)
	{
		v3s16 blockpos = getNodeBlockPos(p);
		MapBlock *block = getBlockNoCreate(blockpos);
		if(block == NULL) return NULL;
		v3s16 relpos = p - blockpos*MAP_BLOCKSIZE;
		return block->getNodePtr(relpos);
	}

	void setNode(v3s16 p, MapNode & n)
	{
		v3s16 blockpos = getNodeBlockPos(p);
		MapBlock *block = getBlock(blockpos);
		if(block == NULL) throw 1;
		v3s16 relpos = p - blockpos*MAP_BLOCKSIZE;
		block->setNode(relpos, n);
	}

	inline void setNode(s16 x, s16 y, s16 z, MapNode & n)
	{
		setNode(v3s16(x,y,z), n);
	}
	
	
	inline MapNode & getNode(v3f p)
	{
		return getNode(floatToInt(p));
	}

	void drawbox(s16 x0, s16 y0, s16 z0, s16 w, s16 h, s16 d, MapNode node)
	{
		x0 += drawoffset.X;
		y0 += drawoffset.Y;
		z0 += drawoffset.Z;
		for(u16 z=0; z<d; z++)
			for(u16 y=0; y<h; y++)
				for(u16 x=0; x<w; x++)
					setNode(x0+x, y0+y, z0+z, node);
	}
	
	void drawslope(s16 x0, s16 y0, s16 z0, s16 w, s16 h, s16 d, s16 dy_x, s16 dy_z, MapNode node)
	{
		x0 += drawoffset.X;
		y0 += drawoffset.Y;
		z0 += drawoffset.Z;
		for(u16 z=0; z<d; z++){
			for(u16 x=0; x<w; x++){
				for(u16 y=0; y < h + z*dy_z + x*dy_x; y++){
					setNode(x0+x, y0+y, z0+z, node);
				}
			}
		}
	}

	f32 getGroundHeight(v2s16 p);
	void setGroundHeight(v2s16 p, f32 h);
	
	void unLightNeighbors(v3s16 pos, f32 oldlight,
			core::list<v3s16> & light_sources,
			core::map<v3s16, MapBlock*> & modified_blocks);
	
	void lightNeighbors(v3s16 pos, f32 oldlight,
			core::map<v3s16, MapBlock*> & modified_blocks);

	v3s16 getBrightestNeighbour(v3s16 p);

	s16 propagateSunlight(v3s16 start,
			core::map<v3s16, MapBlock*> & modified_blocks);
	
	void updateLighting(core::list<MapBlock*> & a_blocks,
			core::map<v3s16, MapBlock*> & modified_blocks);
			
	void nodeAddedUpdate(v3s16 p, f32 light);
	void nodeRemovedUpdate(v3s16 p);

	core::aabbox3d<s16> getDisplayedBlockArea();

	void generateBlock(MapBlock *block);
	void generateMaster();

	bool updateChangedVisibleArea();
	
	void render(video::IVideoDriver* driver,
		video::SMaterial *materials);
		
};

#endif

