/*
(c) 2010 Perttu Ahola <celeron55@gmail.com>
*/

#ifndef MAP_HEADER
#define MAP_HEADER

#include <jmutex.h>
#include <jthread.h>
#include <iostream>
#include <malloc.h>

#ifdef _WIN32
	#include <windows.h>
	#define sleep(x) Sleep(x)
#else
	#include <unistd.h>
#endif

#include "common_irrlicht.h"
#include "heightmap.h"
#include "loadstatus.h"
#include "mapnode.h"
#include "mapblock.h"
#include "mapsector.h"

/*

TODO: Automatically unload blocks from memory and save on disk
	  when they are far away

RULES OF LIGHT:

light = 1.0 (0.999-1.001) is SUNLIGHT. It goes downwards infinitely from
infinite highness and stops when it first hits a non-transparent node.
The lighting of the node it hits is 0.

Other LIGHT SOURCES have a light value of UNDER 0.999.

Light diminishes at a constant factor between nodes.

*/

//void limitBox(core::aabbox3d<s16> & box, core::aabbox3d<s16> & limits);

#define LIGHTING_RADIUS 10

class Map;

class MapUpdateThread : public JThread
{
	bool run;
	JMutex run_mutex;

	Map *map;
	JMutex map_mutex;

public:

	MapUpdateThread() : JThread(), run(true), map(NULL)
	{
		run_mutex.Init();
		map_mutex.Init();
	}

	void * Thread();

	bool getRun()
	{
		run_mutex.Lock();
		bool run_cached = run;
		run_mutex.Unlock();
		return run_cached;
	}
	void setRun(bool a_run)
	{
		run_mutex.Lock();
		run = a_run;
		run_mutex.Unlock();
	}

	void setMap(Map *a_map)
	{
		map_mutex.Lock();
		map = a_map;
		map_mutex.Unlock();
	}
};

class Map : public HeightmapSource, public NodeContainer
{
	/*
		TODO: Dynamic block loading
		Add a filename to the constructor, in which the map will be
		automatically stored - or something.
	*/

	core::map<v2s16, MapSector*> m_sectors;
	JMutex m_getsector_mutex;
	JMutex m_gensector_mutex;

	v3f camera_position;
	v3f camera_direction;
	JMutex camera_mutex;

	MapUpdateThread updater;

public:

	v3s16 drawoffset;

	LoadStatus status;
	
	Map():
		camera_position(0,0,0),
		camera_direction(0,0,1),
		drawoffset(0,0,0)
	{
		m_getsector_mutex.Init();
		m_gensector_mutex.Init();
		camera_mutex.Init();
		assert(m_getsector_mutex.IsInitialized());
		assert(m_gensector_mutex.IsInitialized());
		assert(camera_mutex.IsInitialized());
	}

	~Map()
	{
		/*
			Stop updater thread
		*/
		updater.setRun(false);
		while(updater.IsRunning())
			sleep(1);

		/*
			Free all MapSectors.
		*/
		core::map<v2s16, MapSector*>::Iterator i = m_sectors.getIterator();
		for(; i.atEnd() == false; i++)
		{
			MapSector *sector = i.getNode()->getValue();
			delete sector;
		}
	}

	void updateCamera(v3f pos, v3f dir)
	{
		camera_mutex.Lock();
		camera_position = pos;
		camera_direction = dir;
		camera_mutex.Unlock();
	}

	void StartUpdater()
	{
		updater.setMap(this);
		updater.Start();
	}

	void StopUpdater()
	{
		updater.setRun(false);
		while(updater.IsRunning())
			sleep(1);
	}

	bool UpdaterIsRunning()
	{
		return updater.IsRunning();
	}

	/*
		Returns integer position of the node in given
		floating point position.
	*/
	static v3s16 floatToInt(v3f p)
	{
		v3s16 p2(
			(p.X + (p.X>0 ? BS/2 : -BS/2))/BS,
			(p.Y + (p.Y>0 ? BS/2 : -BS/2))/BS,
			(p.Z + (p.Z>0 ? BS/2 : -BS/2))/BS);
		return p2;
	}

	static v3f intToFloat(v3s16 p)
	{
		v3f p2(
			p.X * BS,
			p.Y * BS,
			p.Z * BS
		);
		return p2;
	}

	static core::aabbox3d<f32> getNodeBox(v3s16 p)
	{
		return core::aabbox3d<f32>(
			(float)p.X * BS - 0.5*BS,
			(float)p.Y * BS - 0.5*BS,
			(float)p.Z * BS - 0.5*BS,
			(float)p.X * BS + 0.5*BS,
			(float)p.Y * BS + 0.5*BS,
			(float)p.Z * BS + 0.5*BS
		);
	}

	Ref<MapSector> getSector(v2s16 p);
	Ref<MapSector> generateSector(v2s16 p2d);
	
	Ref<MapBlock> getBlockNoCreate(v3s16 p);
	Ref<MapBlock> getBlock(v3s16 p);

	f32 getGroundHeight(v2s16 p);
	void setGroundHeight(v2s16 p, f32 h);
	
	bool isValidPosition(v3s16 p)
	{
		v3s16 blockpos = getNodeBlockPos(p);
		Ref<MapBlock> blockref = getBlockNoCreate(blockpos);
		if(blockref.get() == NULL)
			return false;
		v3s16 relpos = p - blockpos*MAP_BLOCKSIZE;
		return blockref.get()->isValidPosition(relpos);
	}
	
	/*
		Returns the position of the block where the node is located
	*/
	v3s16 getNodeBlockPos(v3s16 p)
	{
		return v3s16(
				(p.X>=0 ? p.X : p.X-MAP_BLOCKSIZE+1) / MAP_BLOCKSIZE,
				(p.Y>=0 ? p.Y : p.Y-MAP_BLOCKSIZE+1) / MAP_BLOCKSIZE,
				(p.Z>=0 ? p.Z : p.Z-MAP_BLOCKSIZE+1) / MAP_BLOCKSIZE);
	}

	v2s16 getNodeSectorPos(v2s16 p)
	{
		return v2s16(
				(p.X>=0 ? p.X : p.X-MAP_BLOCKSIZE+1) / MAP_BLOCKSIZE,
				(p.Y>=0 ? p.Y : p.Y-MAP_BLOCKSIZE+1) / MAP_BLOCKSIZE);
	}

	s16 getNodeBlockY(s16 y)
	{
		return (y>=0 ? y : y-MAP_BLOCKSIZE+1) / MAP_BLOCKSIZE;
	}

	Ref<MapBlock> getNodeBlock(v3s16 p)
	{
		v3s16 blockpos = getNodeBlockPos(p);
		return getBlock(blockpos);
	}

	MapNode getNode(v3s16 p)
	{
		v3s16 blockpos = getNodeBlockPos(p);
		Ref<MapBlock> blockref = getBlockNoCreate(blockpos);
		if(blockref.get() == NULL)
			throw exception;
		v3s16 relpos = p - blockpos*MAP_BLOCKSIZE;
		return blockref.get()->getNode(relpos);
	}

	MapNode getNode(s16 x, s16 y, s16 z)
	{
		return getNode(v3s16(x,y,z));
	}

	void setNode(v3s16 p, MapNode & n)
	{
		v3s16 blockpos = getNodeBlockPos(p);
		Ref<MapBlock> blockref = getBlockNoCreate(blockpos);
		if(blockref.get() == NULL)
			throw exception;
		v3s16 relpos = p - blockpos*MAP_BLOCKSIZE;
		blockref.get()->setNode(relpos, n);
	}

	void setNode(s16 x, s16 y, s16 z, MapNode & n)
	{
		setNode(v3s16(x,y,z), n);
	}
	
	
	MapNode getNode(v3f p)
	{
		return getNode(floatToInt(p));
	}

	void drawbox(s16 x0, s16 y0, s16 z0, s16 w, s16 h, s16 d, MapNode node)
	{
		x0 += drawoffset.X;
		y0 += drawoffset.Y;
		z0 += drawoffset.Z;
		for(u16 z=0; z<d; z++)
			for(u16 y=0; y<h; y++)
				for(u16 x=0; x<w; x++)
					setNode(x0+x, y0+y, z0+z, node);
	}
	
	void drawslope(s16 x0, s16 y0, s16 z0, s16 w, s16 h, s16 d, s16 dy_x, s16 dy_z, MapNode node)
	{
		x0 += drawoffset.X;
		y0 += drawoffset.Y;
		z0 += drawoffset.Z;
		for(u16 z=0; z<d; z++){
			for(u16 x=0; x<w; x++){
				for(u16 y=0; y < h + z*dy_z + x*dy_x; y++){
					setNode(x0+x, y0+y, z0+z, node);
				}
			}
		}
	}

	void unLightNeighbors(v3s16 pos, f32 oldlight,
			core::list<v3s16> & light_sources,
			core::map<v3s16, Ref<MapBlock> > & modified_blocks);
	
	void lightNeighbors(v3s16 pos, f32 oldlight,
			core::map<v3s16, Ref<MapBlock> > & modified_blocks);

	v3s16 getBrightestNeighbour(v3s16 p);

	s16 propagateSunlight(v3s16 start,
			core::map<v3s16, Ref<MapBlock> > & modified_blocks);
	
	void updateLighting(core::list< Ref<MapBlock> > & a_blocks,
			core::map<v3s16, Ref<MapBlock> > & modified_blocks);
			
	void nodeAddedUpdate(v3s16 p, f32 light);
	void nodeRemovedUpdate(v3s16 p);

	core::aabbox3d<s16> getDisplayedBlockArea();

	/*void generateBlock(MapBlock *block);
	void generateMaster();*/

	bool updateChangedVisibleArea();
	
	void render(video::IVideoDriver* driver,
		video::SMaterial *materials);
		
};

#endif

