#include "test.h"
#include "common_irrlicht.h"

#include "map.h"
#include "player.h"
#include "main.h"
#include "referencecounted.h"

struct TestRefcount
{
	class TC : public ReferenceCounted
	{
	public:
		TC() : ReferenceCounted("TC")
		{
		}
	};

	void Run()
	{
		TC tc;

		tc.grab();
		assert(tc.getRefcount() == 1);

		tc.drop();
		assert(tc.getRefcount() == 0);

		bool exception = false;
		try
		{
			tc.drop();
		}
		catch(ReferenceCounted::Exception & e)
		{
			exception = true;
		}
		assert(exception == true);
		
		TC tc2;
		{
			Ref<TC> ref(&tc2);
			assert(ref.get() == &tc2);
			assert(tc2.getRefcount() == 1);
		}
		assert(tc2.getRefcount() == 0);
	}
};

struct TestMapNode
{
	void Run()
	{
		MapNode n;

		// Default values
		assert(n.d == MATERIAL_AIR);
		assert(n.light == 0.0);
		
		// Transparency
		n.d = MATERIAL_AIR;
		assert(n.transparent() == true);
		n.d = 0;
		assert(n.transparent() == false);
	}
};

struct TestMapBlock
{
	class TC : public NodeContainer
	{
	public:

		MapNode node;
		bool position_valid;

		TC()
		{
			position_valid = true;
		}

		virtual bool isValidPosition(v3s16 p)
		{
			return position_valid;
		}

		virtual MapNode getNode(v3s16 p)
		{
			if(position_valid == false)
				throw exception;
			return node;
		}

		virtual void setNode(v3s16 p, MapNode & n)
		{
			if(position_valid == false)
				throw exception;
		};
	};

	void Run()
	{
		TC parent;
		
		MapBlock b(&parent, v3s16(1,1,1));
		v3s16 relpos(MAP_BLOCKSIZE, MAP_BLOCKSIZE, MAP_BLOCKSIZE);

		assert(b.getPosRelative() == relpos);

		assert(b.getBox().MinEdge.X == MAP_BLOCKSIZE);
		assert(b.getBox().MaxEdge.X == MAP_BLOCKSIZE*2-1);
		assert(b.getBox().MinEdge.Y == MAP_BLOCKSIZE);
		assert(b.getBox().MaxEdge.Y == MAP_BLOCKSIZE*2-1);
		assert(b.getBox().MinEdge.Z == MAP_BLOCKSIZE);
		assert(b.getBox().MaxEdge.Z == MAP_BLOCKSIZE*2-1);
		
		/*
			TODO: this method should probably be removed
			if the block size isn't going to be set variable
		*/
		assert(b.getSizeNodes() == v3s16(MAP_BLOCKSIZE,
				MAP_BLOCKSIZE, MAP_BLOCKSIZE));
		
		// Changed flag should be initially set
		assert(b.getChangedFlag() == true);
		b.resetChangedFlag();
		assert(b.getChangedFlag() == false);

		// All nodes should have been set to
		// .d=MATERIAL_AIR and .light < 0.001
		for(u16 z=0; z<MAP_BLOCKSIZE; z++)
			for(u16 y=0; y<MAP_BLOCKSIZE; y++)
				for(u16 x=0; x<MAP_BLOCKSIZE; x++){
					assert(b.getNode(v3s16(x,y,z)).d == MATERIAL_AIR);
					assert(b.getNode(v3s16(x,y,z)).light < 0.001);
				}
		
		/*
			Parent fetch functions
		*/
		parent.position_valid = false;
		parent.node.d = 5;

		MapNode n;
		
		// Positions in the block should still be valid
		assert(b.isValidPositionParent(v3s16(0,0,0)) == true);
		assert(b.isValidPositionParent(v3s16(MAP_BLOCKSIZE-1,MAP_BLOCKSIZE-1,MAP_BLOCKSIZE-1)) == true);
		n = b.getNodeParent(v3s16(0,MAP_BLOCKSIZE-1,0));
		assert(n.d == MATERIAL_AIR);

		// ...but outside the block they should be invalid
		assert(b.isValidPositionParent(v3s16(-121,2341,0)) == false);
		assert(b.isValidPositionParent(v3s16(-1,0,0)) == false);
		assert(b.isValidPositionParent(v3s16(MAP_BLOCKSIZE-1,MAP_BLOCKSIZE-1,MAP_BLOCKSIZE)) == false);

		bool exception_thrown = false;
		try{
			// This should throw an exception
			MapNode n = b.getNodeParent(v3s16(0,0,-1));
		}
		catch(NodeContainer::Exception &e)
		{
			exception_thrown = true;
		}
		assert(exception_thrown);

		parent.position_valid = true;
		// Now the positions outside should be valid
		assert(b.isValidPositionParent(v3s16(-121,2341,0)) == true);
		assert(b.isValidPositionParent(v3s16(-1,0,0)) == true);
		assert(b.isValidPositionParent(v3s16(MAP_BLOCKSIZE-1,MAP_BLOCKSIZE-1,MAP_BLOCKSIZE)) == true);
		n = b.getNodeParent(v3s16(0,0,MAP_BLOCKSIZE));
		assert(n.d == 5);

		/*
			Set a node
		*/
		v3s16 p(1,2,0);
		n.d = 4;
		b.setNode(p, n);
		assert(b.getNode(p).d == 4);
		assert(b.getNodeMaterial(p) == 4);
		assert(b.getNodeMaterial(v3s16(-1,-1,0)) == 5);
		
		/*
			propagateSunlight()
		*/
		// Set lighting of all nodes to 0
		for(u16 z=0; z<MAP_BLOCKSIZE; z++){
			for(u16 y=0; y<MAP_BLOCKSIZE; y++){
				for(u16 x=0; x<MAP_BLOCKSIZE; x++){
					MapNode n = b.getNode(v3s16(x,y,z));
					n.light = 0.0;
					b.setNode(v3s16(x,y,z), n);
				}
			}
		}
		parent.position_valid = false;
		assert(b.propagateSunlight() == true);
		assert(b.getNode(v3s16(1,4,0)).light > 0.999);
		assert(b.getNode(v3s16(1,3,0)).light > 0.999);
		assert(b.getNode(v3s16(1,2,0)).light < 0.001);
		assert(b.getNode(v3s16(1,1,0)).light < 0.001);
		assert(b.getNode(v3s16(1,0,0)).light < 0.001);
		assert(b.getNode(v3s16(1,2,3)).light > 0.999);
		assert(b.getFaceLight(p, v3s16(0,1,0)) > 0.999);
		assert(b.getFaceLight(p, v3s16(0,-1,0)) < 0.001);
		assert(fabs(b.getFaceLight(p, v3s16(0,0,1))-0.7) < 0.001);
		parent.position_valid = true;
		parent.node.light = 0.500;
		assert(b.propagateSunlight() == false);
		// Should not touch blocks that are not affected (that is, all of them)
		assert(b.getNode(v3s16(1,2,3)).light > 0.999);
	}
};

struct TestMap
{
	void Run()
	{
		Map map;
	}
};

#define TEST(X)\
{\
	X x;\
	std::cout<<"Running " #X <<std::endl;\
	x.Run();\
}

void run_tests()
{
	std::cout<<"run_tests() started"<<std::endl;
	TEST(TestRefcount);
	TEST(TestMapNode);
	TEST(TestMapBlock);
	TEST(TestMap);
	std::cout<<"run_tests() passed"<<std::endl;
}

