#include "test.h"
#include "common_irrlicht.h"

#include "map.h"
#include "player.h"
#include "main.h"
#include "referencecounted.h"
#include "heightmap.h"

/*
	Asserts that the exception occurs
*/
#define EXCEPTION_CHECK(EType, code)\
{\
	bool exception_thrown = false;\
	try{ code; }\
	catch(EType &e) { exception_thrown = true; }\
	assert(exception_thrown);\
}
		

struct TestRefcount
{
	class TC : public ReferenceCounted
	{
	public:
		s16 value;
		TC() : ReferenceCounted("TC")
		{
			value = 0;
		}
	};

	Ref<TC> TestReturn()
	{
		static TC tc;
		tc.value = 5;
		return &tc;
	}

	void Run()
	{
		TC tc;

		tc.grab();
		assert(tc.getRefcount() == 1);

		tc.drop();
		assert(tc.getRefcount() == 0);

		bool exception = false;
		try
		{
			tc.drop();
		}
		catch(ReferenceCounted::Exception & e)
		{
			exception = true;
		}
		assert(exception == true);
		
		TC tc2;
		{
			Ref<TC> ref(&tc2);
			assert(ref.get() == &tc2);
			assert(tc2.getRefcount() == 1);
		}
		assert(tc2.getRefcount() == 0);

		Ref<TC> ref2 = TestReturn();
		assert(ref2.get()->value == 5);
		assert(ref2.get()->getRefcount() == 1);
	}
};

struct TestMapNode
{
	void Run()
	{
		MapNode n;

		// Default values
		assert(n.d == MATERIAL_AIR);
		assert(n.light == 0.0);
		
		// Transparency
		n.d = MATERIAL_AIR;
		assert(n.transparent() == true);
		n.d = 0;
		assert(n.transparent() == false);
	}
};

struct TestMapBlock
{
	class TC : public NodeContainer
	{
	public:

		MapNode node;
		bool position_valid;

		TC()
		{
			position_valid = true;
		}

		virtual bool isValidPosition(v3s16 p)
		{
			return position_valid;
		}

		virtual MapNode getNode(v3s16 p)
		{
			if(position_valid == false)
				throw InvalidPositionException();
			return node;
		}

		virtual void setNode(v3s16 p, MapNode & n)
		{
			if(position_valid == false)
				throw InvalidPositionException();
		};
	};

	void Run()
	{
		TC parent;
		
		MapBlock b(&parent, v3s16(1,1,1));
		v3s16 relpos(MAP_BLOCKSIZE, MAP_BLOCKSIZE, MAP_BLOCKSIZE);

		assert(b.getPosRelative() == relpos);

		assert(b.getBox().MinEdge.X == MAP_BLOCKSIZE);
		assert(b.getBox().MaxEdge.X == MAP_BLOCKSIZE*2-1);
		assert(b.getBox().MinEdge.Y == MAP_BLOCKSIZE);
		assert(b.getBox().MaxEdge.Y == MAP_BLOCKSIZE*2-1);
		assert(b.getBox().MinEdge.Z == MAP_BLOCKSIZE);
		assert(b.getBox().MaxEdge.Z == MAP_BLOCKSIZE*2-1);
		
		assert(b.isValidPosition(v3s16(0,0,0)) == true);
		assert(b.isValidPosition(v3s16(-1,0,0)) == false);
		assert(b.isValidPosition(v3s16(-1,-142,-2341)) == false);
		assert(b.isValidPosition(v3s16(-124,142,2341)) == false);
		assert(b.isValidPosition(v3s16(MAP_BLOCKSIZE-1,MAP_BLOCKSIZE-1,MAP_BLOCKSIZE-1)) == true);
		assert(b.isValidPosition(v3s16(MAP_BLOCKSIZE-1,MAP_BLOCKSIZE,MAP_BLOCKSIZE-1)) == false);

		/*
			TODO: this method should probably be removed
			if the block size isn't going to be set variable
		*/
		assert(b.getSizeNodes() == v3s16(MAP_BLOCKSIZE,
				MAP_BLOCKSIZE, MAP_BLOCKSIZE));
		
		// Changed flag should be initially set
		assert(b.getChangedFlag() == true);
		b.resetChangedFlag();
		assert(b.getChangedFlag() == false);

		// All nodes should have been set to
		// .d=MATERIAL_AIR and .light < 0.001
		for(u16 z=0; z<MAP_BLOCKSIZE; z++)
			for(u16 y=0; y<MAP_BLOCKSIZE; y++)
				for(u16 x=0; x<MAP_BLOCKSIZE; x++){
					assert(b.getNode(v3s16(x,y,z)).d == MATERIAL_AIR);
					assert(b.getNode(v3s16(x,y,z)).light < 0.001);
				}
		
		/*
			Parent fetch functions
		*/
		parent.position_valid = false;
		parent.node.d = 5;

		MapNode n;
		
		// Positions in the block should still be valid
		assert(b.isValidPositionParent(v3s16(0,0,0)) == true);
		assert(b.isValidPositionParent(v3s16(MAP_BLOCKSIZE-1,MAP_BLOCKSIZE-1,MAP_BLOCKSIZE-1)) == true);
		n = b.getNodeParent(v3s16(0,MAP_BLOCKSIZE-1,0));
		assert(n.d == MATERIAL_AIR);

		// ...but outside the block they should be invalid
		assert(b.isValidPositionParent(v3s16(-121,2341,0)) == false);
		assert(b.isValidPositionParent(v3s16(-1,0,0)) == false);
		assert(b.isValidPositionParent(v3s16(MAP_BLOCKSIZE-1,MAP_BLOCKSIZE-1,MAP_BLOCKSIZE)) == false);

		bool exception_thrown = false;
		try{
			// This should throw an exception
			MapNode n = b.getNodeParent(v3s16(0,0,-1));
		}
		catch(InvalidPositionException &e)
		{
			exception_thrown = true;
		}
		assert(exception_thrown);

		parent.position_valid = true;
		// Now the positions outside should be valid
		assert(b.isValidPositionParent(v3s16(-121,2341,0)) == true);
		assert(b.isValidPositionParent(v3s16(-1,0,0)) == true);
		assert(b.isValidPositionParent(v3s16(MAP_BLOCKSIZE-1,MAP_BLOCKSIZE-1,MAP_BLOCKSIZE)) == true);
		n = b.getNodeParent(v3s16(0,0,MAP_BLOCKSIZE));
		assert(n.d == 5);

		/*
			Set a node
		*/
		v3s16 p(1,2,0);
		n.d = 4;
		b.setNode(p, n);
		assert(b.getNode(p).d == 4);
		assert(b.getNodeMaterial(p) == 4);
		assert(b.getNodeMaterial(v3s16(-1,-1,0)) == 5);
		
		/*
			propagateSunlight()
		*/
		// Set lighting of all nodes to 0
		for(u16 z=0; z<MAP_BLOCKSIZE; z++){
			for(u16 y=0; y<MAP_BLOCKSIZE; y++){
				for(u16 x=0; x<MAP_BLOCKSIZE; x++){
					MapNode n = b.getNode(v3s16(x,y,z));
					n.light = 0.0;
					b.setNode(v3s16(x,y,z), n);
				}
			}
		}
		parent.position_valid = false;
		assert(b.propagateSunlight() == true);
		assert(b.getNode(v3s16(1,4,0)).light > 0.999);
		assert(b.getNode(v3s16(1,3,0)).light > 0.999);
		assert(b.getNode(v3s16(1,2,0)).light < 0.001);
		assert(b.getNode(v3s16(1,1,0)).light < 0.001);
		assert(b.getNode(v3s16(1,0,0)).light < 0.001);
		assert(b.getNode(v3s16(1,2,3)).light > 0.999);
		assert(b.getFaceLight(p, v3s16(0,1,0)) > 0.999);
		assert(b.getFaceLight(p, v3s16(0,-1,0)) < 0.001);
		assert(fabs(b.getFaceLight(p, v3s16(0,0,1))-0.7) < 0.001);
		parent.position_valid = true;
		parent.node.light = 0.500;
		assert(b.propagateSunlight() == false);
		// Should not touch blocks that are not affected (that is, all of them)
		assert(b.getNode(v3s16(1,2,3)).light > 0.999);
	}
};

struct TestMapSector
{
	class TC : public NodeContainer
	{
	public:

		MapNode node;
		bool position_valid;

		TC()
		{
			position_valid = true;
		}

		virtual bool isValidPosition(v3s16 p)
		{
			return position_valid;
		}

		virtual MapNode getNode(v3s16 p)
		{
			if(position_valid == false)
				throw InvalidPositionException();
			return node;
		}

		virtual void setNode(v3s16 p, MapNode & n)
		{
			if(position_valid == false)
				throw InvalidPositionException();
		};
	};
	
	void Run()
	{
		TC parent;
		parent.position_valid = false;

		DummyHeightmap dummyheightmap;

		MapSector sector(&parent, v2s16(1,1), &dummyheightmap);
		
		EXCEPTION_CHECK(InvalidPositionException, sector.getBlock(0));
		EXCEPTION_CHECK(InvalidPositionException, sector.getBlock(1));

		MapBlock * bref = sector.createBlankBlock(-2);
		
		EXCEPTION_CHECK(InvalidPositionException, sector.getBlock(0));
		assert(sector.getBlock(-2) == bref);
		
		//TODO: Check for AlreadyExistsException

		/*bool exception_thrown = false;
		try{
			sector.getBlock(0);
		}
		catch(InvalidPositionException &e){
			exception_thrown = true;
		}
		assert(exception_thrown);*/

	}
};

struct TestHeightmap
{
	void TestSingleFixed()
	{
		const s16 BS1 = 4;
		OneChildHeightmap hm1(BS1);
		
		// Test that it is filled with < GROUNDHEIGHT_VALID_MINVALUE
		for(s16 y=0; y<=BS1; y++){
			for(s16 x=0; x<=BS1; x++){
				v2s16 p(x,y);
				assert(hm1.m_child.getGroundHeight(p)
					< GROUNDHEIGHT_VALID_MINVALUE);
			}
		}

		hm1.m_child.setGroundHeight(v2s16(1,0), 2.0);
		//hm1.m_child.print();
		assert(fabs(hm1.getGroundHeight(v2s16(1,0))-2.0)<0.001);
		hm1.setGroundHeight(v2s16(0,1), 3.0);
		assert(fabs(hm1.m_child.getGroundHeight(v2s16(0,1))-3.0)<0.001);
		
		// Fill with -1.0
		for(s16 y=0; y<=BS1; y++){
			for(s16 x=0; x<=BS1; x++){
				v2s16 p(x,y);
				hm1.m_child.setGroundHeight(p, -1.0);
			}
		}

		f32 corners[] = {0.0, 0.0, 1.0, 1.0};
		hm1.m_child.generateContinued(0.0, 0.0, corners);
		
		hm1.m_child.print();
		assert(fabs(hm1.m_child.getGroundHeight(v2s16(1,0))-0.2)<0.05);
		assert(fabs(hm1.m_child.getGroundHeight(v2s16(4,3))-0.7)<0.05);
		assert(fabs(hm1.m_child.getGroundHeight(v2s16(4,4))-1.0)<0.05);
	}

	void TestUnlimited()
	{
		//g_heightmap_debugprint = true;
		const s16 BS1 = 4;
		UnlimitedHeightmap hm1(BS1, 0.0, 0.0, 5.0);
		// Go through it so it generates itself
		for(s16 y=0; y<=BS1; y++){
			for(s16 x=0; x<=BS1; x++){
				v2s16 p(x,y);
				hm1.getGroundHeight(p);
			}
		}
		// Print it
		std::cout<<"UnlimitedHeightmap hm1:"<<std::endl;
		hm1.print();
		/*for(s16 y=0; y<=BS1; y++){
			for(s16 x=0; x<=BS1; x++){
				v2s16 p(x,y);
				f32 h = hm1.getGroundHeight(p);
				printf("% 2.1f ", h);
			}
			std::cout<<std::endl;
		}*/
		
		std::cout<<"testing UnlimitedHeightmap set/get"<<std::endl;
		v2s16 p1(0,3);
		f32 v1(234.01);
		// Get first heightmap and try setGroundHeight
		FixedHeightmap * href = hm1.getHeightmap(v2s16(0,0));
		href->setGroundHeight(p1, v1);
		// Read from UnlimitedHeightmap
		assert(fabs(hm1.getGroundHeight(p1)-v1)<0.001);
	}
	
	void Random()
	{
		std::cout<<"Running random code"<<std::endl;
		std::cout<<"rand() values: ";
		for(u16 i=0; i<5; i++)
			std::cout<<(u16)rand()<<" ";
		std::cout<<std::endl;

		const s16 BS1 = 4;
		//UnlimitedHeightmap hm1(BS1, 0.0, 0.0, 5.0);
		UnlimitedHeightmap hm1(BS1, 2.5, 0.5, 0.0);

		// Force hm1 to generate a some heightmap
		hm1.getGroundHeight(v2s16(0,0));
		hm1.getGroundHeight(v2s16(BS1,0));
		hm1.print();

		// Get the (0,0) and (1,0) heightmaps
		/*FixedHeightmap * hr00 = hm1.getHeightmap(v2s16(0,0));
		FixedHeightmap * hr01 = hm1.getHeightmap(v2s16(1,0));
		f32 corners[] = {1.0, 1.0, 1.0, 1.0};
		hr00->generateContinued(0.0, 0.0, corners);
		hm1.print();*/
	}

	void Run()
	{
		//srand(7); // Get constant random
		srand(time(0)); // Get better random

		TestSingleFixed();
		TestUnlimited();
		Random();

		g_heightmap_debugprint = false;
		//g_heightmap_debugprint = true;

		//assert(0); // Uncomment when not interested in running the game

#if 0
		/*
			Length of side per block. The actual size of a heightmap
			is one higher.
		*/
		const s16 MASTER_BS = 2;
		const s16 CHILD_BS = 8;
		
		// This will contain the corner heights for the child heightmaps
		// Use constant height of 4.0
		UnlimitedHeightmap master(MASTER_BS, 0.0, 0.0, 4.0);

		//std::cout<<"Master:"<<std::endl;
		for(s16 y=0; y<=MASTER_BS; y++){
			for(s16 x=0; x<=MASTER_BS; x++){
				v2s16 p(x,y);
				f32 h = master.getGroundHeight(p);
				//printf("% 2.1f ", h);
			}
			//std::cout<<std::endl;
		}
		// Check contents
		for(s16 y=0; y<=MASTER_BS; y++){
			for(s16 x=0; x<=MASTER_BS; x++){
				v2s16 p(x,y);
				assert(abs(master.getGroundHeight(p) - 4.0)<0.001);
			}
		}

		/*
			Now, what should we do? Let's take four directions.
		*/
		v2s16 dirs[] = {
			v2s16(1,0),
			v2s16(-1,0),
			v2s16(0,1),
			v2s16(0,-1)
		};
		for(u16 i=0; i<4; i++){
			v2s16 dir = dirs[i];
			/*
				Fill master with -1.0
			*/
			for(s16 y=0; y<=MASTER_BS; y++){
				for(s16 x=0; x<=MASTER_BS; x++){
					v2s16 p(x,y);
					master.setGroundHeight(p, -1.0);
					assert(abs(master.getGroundHeight(p) - (-1.0))<0.001);
				}
			}
			// Center point of master
			v2s16 ms(MASTER_BS/2, MASTER_BS/2);
			/*
				Get a vector that is pointing counter-clockwise from
				the position defined by dir. That is, dir turned 90
				decrees counter-clockwise.
			*/
			v3s16 dir3(dir.X, dir.Y, 0);
			v3s16 dir3_90 = dir3.crossProduct(v3s16(0,0,-1));
			v2s16 dir90(dir3_90.X, dir3_90.Y);
			std::cout<<"dir=("<<dir.X<<","<<dir.Y<<")"<<std::endl;
			std::cout<<"dir90=("<<dir90.X<<","<<dir90.Y<<")"<<std::endl;
			
		}
		assert(0);

		for(s16 y=0; y<=MASTER_BS; y++){
			for(s16 x=MASTER_BS; x>=0; x--){
				v2s16 p(x,y);
				assert(abs(master.getGroundHeight(p) - (-1.0))<0.001);
				master.setGroundHeight(p, 5.0);
				assert(abs(master.getGroundHeight(p) - 5.0)<0.001);
			}
		}
		// Actually, set the x=0 border to 0.0

		/*
			Create child heightmaps.

			These should now align from their corners to the master map
		*/
		FixedHeightmap child00(&master, v2s16(0,0), CHILD_BS);
		FixedHeightmap child01(&master, v2s16(0,1), CHILD_BS);
		FixedHeightmap child11(&master, v2s16(1,1), CHILD_BS);
		FixedHeightmap child10(&master, v2s16(1,0), CHILD_BS);
#endif	
	}
};

#define TEST(X)\
{\
	X x;\
	std::cout<<"Running " #X <<std::endl;\
	x.Run();\
}

void run_tests()
{
	std::cout<<"run_tests() started"<<std::endl;
	TEST(TestRefcount);
	TEST(TestMapNode);
	TEST(TestMapBlock);
	TEST(TestMapSector);
	TEST(TestHeightmap);
	std::cout<<"run_tests() passed"<<std::endl;
}

