#ifndef CONNECTION_HEADER
#define CONNECTION_HEADER

#include <assert.h>
#include "common_irrlicht.h"
#include "socket.h"
#include "utility.h"
#include "exceptions.h"
#include <iostream>

namespace con
{

#define PROTOCOL_ID 0x4f457403

class NotImplementedException : public BaseException
{
public:
	NotImplementedException(const char *s):
		BaseException(s)
	{}
};

class NotFoundException : public BaseException
{
public:
	NotFoundException(const char *s):
		BaseException(s)
	{}
};

class ConnectionException : public BaseException
{
public:
	ConnectionException(const char *s):
		BaseException(s)
	{}
};

class InvalidIncomingDataException : public BaseException
{
public:
	InvalidIncomingDataException(const char *s):
		BaseException(s)
	{}
};

class InvalidOutgoingDataException : public BaseException
{
public:
	InvalidOutgoingDataException(const char *s):
		BaseException(s)
	{}
};

class NoIncomingDataException : public BaseException
{
public:
	NoIncomingDataException(const char *s):
		BaseException(s)
	{}
};

void writeU32(u8 *data, u32 i);
void writeU16(u8 *data, u16 i);
void writeU8(u8 *data, u8 i);
u32 readU32(u8 *data);
u16 readU16(u8 *data);
u8 readU8(u8 *data);

inline u16 readPeerId(u8 *packetdata)
{
	return readU16(&packetdata[4]);
}
inline u16 readSeqNumber(u8 *packetdata)
{
	return readU16(&packetdata[6]);
}
inline u8 readType(u8 *packetdata)
{
	return readU8(&packetdata[8]);
}

struct BufferedPacket
{
	float time; // Seconds from buffering the packet
	Address address; // Sender or destination
	u8 *data; // Data of the packet, including headers
	s32 size;
};

/*
	Connection packet header format:
		u32 protocol_id [0-3]
		u16 peer_id [4-5]
		u16 sequence_num [6-7]
		u8 type [8]
	
	peer_id:
		Unique to each peer.
		value 0 is reserved for making new connections
		value 1 is reserved for server
	sequence_num:
		different for reliable and unreliable packets.
	
	ACK packet
	- sequence_num is unique
	- ACKed sequence_num is stored in first bytes of payload data.

	RELIABLE packet
	- After receiving, an ACK packet is sent.
	- Reliable packets are delivered in the order they were sent.
	- If a reliable packet is re-sent, the sequence number is kept
	  the same.

	PEER_ID packet
	- Sets client peer_id
	- New peer id is stored in first two bytes of data.
*/

#define HEADER_SIZE (4+2+2+1)

#define TYPE_NORMAL 0
#define TYPE_RELIABLE 1
#define TYPE_ACK 2
#define TYPE_PEER_ID 3

#define PEER_ID_NEW 0
#define PEER_ID_SERVER 1

void makePacket(u8 *packetdata, u8 *data, u32 datasize,
		u32 protocol_id, u16 peer_id, u16 seqnum, u8 type);

//#define SEQNUM_INITIAL 65500
#define SEQNUM_INITIAL 0x10

struct Peer
{
	Peer()
	{
		m_next_seqnum_unreliable = SEQNUM_INITIAL;
		m_next_seqnum_reliable = SEQNUM_INITIAL;
		m_next_incoming_reliable_seqnum = SEQNUM_INITIAL;
		m_timeout_counter = 0.0;
	}
	~Peer()
	{
		core::list<BufferedPacket>::Iterator i;
		i = m_ack_timeout_queue.begin();
		for(; i != m_ack_timeout_queue.end(); i++)
		{
			delete[] i->data;
		}
		i = m_incoming_reliable_queue.begin();
		for(; i != m_incoming_reliable_queue.end(); i++)
		{
			delete[] i->data;
		}
	}
	u16 m_next_seqnum_unreliable;
	u16 m_next_seqnum_reliable;
	// seqnum of the next incoming reliable packet
	u16 m_next_incoming_reliable_seqnum;
	// A list of sent reliables that haven't received an ACK.
	// when an ack is received, the corresponding item is removed.
	// If no ack is received in some time, the packet is re-sent.
	core::list<BufferedPacket> m_ack_timeout_queue;
	// A list of received reliables whose sequence number is higher than
	// what is expected next.
	core::list<BufferedPacket> m_incoming_reliable_queue;
	// Address of the peer
	Address m_address;
	// Unique id of the peer
	u16 m_id;
	// Seconds from last receive
	float m_timeout_counter;
};

class Connection
{
public:
	Connection(u32 protocol_id);
	~Connection();
	// Start being a server
	void Serve(unsigned short port);
	// Connect to a server
	void Connect(Address address);
	bool Connected();
	u32 Receive(u16 &peer_id, u8 *data, s32 datasize, u16 &seqnumber);
	void Send(u16 peer_id, u8 *data, u32 datasize, bool reliable=false);
	void ProtocolSend(Address &destination, u8 *data, u32 datasize,
			u16 peer_id, u16 seqnum, u8 type);
	void RawSend(Address &destination, u8 *data, u32 datasize);
	void RunTimeouts(float dtime);
	Peer* GetPeer(u16 peer_id);
private:
	void PrintInfo(); // For debug printing
	u32 m_protocol_id;
	core::map<u16, Peer*> m_peers;
	u16 m_peer_id;
	bool m_waiting_new_peer_id;
	UDPSocket m_socket;
};

} // namespace

#endif

