#include "connection.h"

namespace con
{

void makePacket(u8 *packetdata, u8 *data, u32 datasize,
		u32 protocol_id, u16 peer_id, u16 seqnum, u8 type)
{
	writeU32(&packetdata[0], protocol_id);
	writeU16(&packetdata[4], peer_id);
	writeU16(&packetdata[6], seqnum);
	writeU8(&packetdata[8], type);

	memcpy(&packetdata[HEADER_SIZE], data, datasize);
}

Connection::Connection(u32 protocol_id)
{
	m_protocol_id = protocol_id;
	m_peer_id = PEER_ID_NEW;
	m_waiting_new_peer_id = false;
}

Connection::~Connection()
{
	// Clear peers
	core::map<u16, Peer*>::Iterator j;
	j = m_peers.getIterator();
	for(; j.atEnd() == false; j++)
	{
		Peer *peer = j.getNode()->getValue();
		delete peer;
	}
}

void Connection::Serve(unsigned short port)
{
	m_socket.Bind(port);
	m_peer_id = PEER_ID_SERVER;
}

void Connection::Connect(Address address)
{
	core::map<u16, Peer*>::Node *node = m_peers.find(PEER_ID_SERVER);
	if(node != NULL){
		throw ConnectionException("Already connected to a server");
	}

	Peer *peer = new Peer();
	peer->m_address = address;
	peer->m_id = PEER_ID_SERVER;
	m_peers.insert(peer->m_id, peer);
	
	m_socket.Bind(0);

	// Send a dummy packet to server with peer_id = PEER_ID_NEW
	m_peer_id = PEER_ID_NEW;
	Send(PEER_ID_SERVER, NULL, 0, true);
	//ProtocolSend(address, NULL, 0, PEER_ID_NEW, 

	m_waiting_new_peer_id = true;
}

bool Connection::Connected()
{
	if(m_peers.size() != 1)
		return false;
		
	core::map<u16, Peer*>::Node *node = m_peers.find(PEER_ID_SERVER);
	if(node == NULL)
		return false;
	
	if(m_peer_id == PEER_ID_NEW)
		return false;
	
	return true;
}

u32 Connection::Receive(u16 &peer_id, u8 *data, s32 datasize, u16 &seqnumber)
{
	for(;;)
	{
		s32 headersize = HEADER_SIZE;
		/*
			Check if there is a suitable packet in the reliable
			packet queue of some peer
		*/
		core::map<u16, Peer*>::Iterator j;
		j = m_peers.getIterator();
		for(; j.atEnd() == false; j++)
		{
			Peer *peer = j.getNode()->getValue();
			assert(j.getNode()->getKey() == peer->m_id);
			
			core::list<BufferedPacket>::Iterator i;
check_first_on_queue:
			// Get first item of buffer
			i = peer->m_incoming_reliable_queue.begin();
			// If queue is empty, continue to next peer
			if(i == peer->m_incoming_reliable_queue.end()){
				//std::cout<<"Queue is empty"<<std::endl;
				// Queue is empty
				continue;
			}
		
			PrintInfo();
			std::cout<<"Receive(): Processing queued packets for peer_id="
					<<peer->m_id<<"... ";

			std::cout<<"next seqnum should be "
					<<peer->m_next_incoming_reliable_seqnum
					<<"; ";

			// Read the sequence number of the first one
			u16 firstseqnum = readSeqNumber(i->data);
			std::cout<<"First seqnum is "<<firstseqnum<<"; ";

			// If it is old, dump it and start again
			if(firstseqnum < peer->m_next_incoming_reliable_seqnum)
			{
				std::cout<<"First one is old, removing; ";
				delete[] i->data;
				peer->m_incoming_reliable_queue.erase(i);
				goto check_first_on_queue;
			}
			// If it has the right sequence number, remove it from
			// the queue and return the packet.
			else if(firstseqnum ==
					peer->m_next_incoming_reliable_seqnum)
			{
				std::cout<<"Taking the first one from queue"<<std::endl;

				peer->m_next_incoming_reliable_seqnum++;

				assert(datasize >= i->size - headersize);
				u32 received_datasize = i->size - headersize;
				memcpy(data, &i->data[headersize], received_datasize);

				peer_id = readPeerId(i->data);
				seqnumber = readSeqNumber(i->data);

				delete[] i->data;
				peer->m_incoming_reliable_queue.erase(i);
				
				return received_datasize;
			}
			else{
				std::cout<<"The first one in queue comes later"<<std::endl;
			}
		}

		/*
			Receive a packet from the network
		*/

		s32 packet_maxsize = datasize + headersize;
		//u8 packetdata[packet_maxsize];
		u8 *packetdata = new u8[packet_maxsize];

		Address sender;

		s32 received_size = m_socket.Receive(sender, packetdata, packet_maxsize);
	
		//std::cout<<"received_size="<<received_size<<std::endl;
		if(received_size <= 0)
			throw NoIncomingDataException("No incoming data");
		if(received_size < headersize)
			throw InvalidIncomingDataException("No full header received");
		if(readU32(&packetdata[0]) != m_protocol_id)
			throw InvalidIncomingDataException("Invalid protocol id");
		
		peer_id = readPeerId(packetdata);
		seqnumber = readSeqNumber(packetdata);
		u8 type = readType(packetdata);

		if(peer_id == PEER_ID_NEW){
			// Somebody wants to make a new connection

			// Get a unique peer id (2 or higher)
			u16 peer_id_new = 2;
			core::map<u16, Peer*>::Iterator j;
			j = m_peers.getIterator();
			for(; j.atEnd() == false; j++)
			{
				Peer *peer = j.getNode()->getValue();
				if(peer_id_new <= peer->m_id)
					peer_id_new = peer->m_id + 1;
			}

			PrintInfo();
			std::cout<<"Receive(): Got a PEER_ID_NEW, giving peer_id="
					<<peer_id_new<<std::endl;

			// Create a peer
			Peer *peer = new Peer();
			peer->m_address = sender;
			peer->m_id = peer_id_new;
			m_peers.insert(peer->m_id, peer);

			// Create packet contents
			u8 data[2];
			writeU16(&data[0], peer_id_new);
			// Seqnum, type
			u16 seqnum = peer->m_next_seqnum_unreliable;
			peer->m_next_seqnum_unreliable++;
			u8 type = TYPE_PEER_ID;
			// Send it
			ProtocolSend(sender, data, 2, m_peer_id, seqnum, type);
		
			// We're now talking to a valid peer_id
			peer_id = peer_id_new;
			
			/*// Process more packets
			delete packetdata;
			continue;*/
		}

		/*PrintInfo();
		std::cout<<"peer_id="<<peer_id<<std::endl;*/
		
		core::map<u16, Peer*>::Node *node = m_peers.find(peer_id);

		if(node == NULL){
			// Peer not found
			throw InvalidIncomingDataException("Peer not found (possible timeout)");
		}

		Peer *peer = node->getValue();

		peer->m_timeout_counter = 0.0;
		
		if(type == TYPE_PING)
		{
			// Send PONG
			u16 packetseqnum = peer->m_next_seqnum_unreliable;
			peer->m_next_seqnum_unreliable++;
			u8 packettype = TYPE_PONG;
			ProtocolSend(sender, NULL, 0, m_peer_id, packetseqnum, packettype);
		}
		else if(type == TYPE_PONG)
		{
		}
		else if(type == TYPE_ACK)
		{
			/*
				TODO:
				The packet is an ACK.
				Remove corresponding packet from ack timeout queue
			*/

			u16 ackseqnum = readU16(&packetdata[headersize]);

			PrintInfo();
			std::cout<<"Receive(): Got an ACK packet for seqnum="
					<<ackseqnum<<" from peer_id="
					<<peer_id<<std::endl;
			
			//std::cout<<"Going through ack_timeout_queue"<<std::endl;

			core::list<BufferedPacket>::Iterator i;
			i = peer->m_ack_timeout_queue.begin();
			for(; i != peer->m_ack_timeout_queue.end(); i++)
			{
				u16 seqnum = readSeqNumber(i->data);
				//std::cout<<"seqnum="<<seqnum<<std::endl;
				if(seqnum == ackseqnum){
					PrintInfo();
					std::cout<<"Receive(): Found ACKed packet in buffer"
							". Removing it."<<std::endl;
					
					// Decrease resend timeout
					peer->m_resend_timeout /= 2.0;
					if(peer->m_resend_timeout < RESEND_TIMEOUT_MINIMUM)
						peer->m_resend_timeout = RESEND_TIMEOUT_MINIMUM;

					// Found packet in buffer
					// delete it
					delete[] i->data;
					peer->m_ack_timeout_queue.erase(i);
					break;
				}
			}

			// Process more packets
			delete packetdata;
			continue;
		}
		else if(type == TYPE_PEER_ID)
		{
			/*
				We don't allow changing our peer id if we alredy have one
				Also, happens in this (harmless) glitch:
				- Client sends reliable packet with peer_id=0
				- Server sends TYPE_ACK
				- Server sends TYPE_PEER_ID
				- Client gets TYPE_PEER_ID
				- Client doesn't get ACK, sends a peer_id=0 packet again
				- Client gets TYPE_PEER_ID again -> goes here
			*/
			if(m_peer_id != PEER_ID_NEW){
				throw InvalidIncomingDataException(
						"Got a PEER_ID packet but we already have one");
			}

			u16 new_peer_id = readU16(&packetdata[headersize]);
			PrintInfo();
			std::cout<<"Receive(): Changing peer id to "<<new_peer_id<<std::endl;
			m_peer_id = new_peer_id;

			// Process more packets
			delete packetdata;
			continue;
		}
		else if(type == TYPE_RELIABLE)
		{
			/*
				The packet is a reliable packet:
				- Send an ACK
				- If the packet is the one expected next, return it.
				- If the packet is to be delivered later, queue it and get a
				  new packet.
			*/

			PrintInfo();
			std::cout<<"Receive(): Received a reliable packet, seqnumber="
					<<seqnumber<<", sending ACK."<<std::endl;
			
			// Send ACK
			u8 ackdata[2];
			writeU16(&ackdata[0], seqnumber);
			u16 ackseqnum = peer->m_next_seqnum_unreliable;
			peer->m_next_seqnum_unreliable++;
			u8 acktype = TYPE_ACK;
			ProtocolSend(sender, ackdata, 2, m_peer_id, ackseqnum, acktype);
			
			// If it is an old packet, dump it.
			if(seqnumber < peer->m_next_incoming_reliable_seqnum)
			{
				PrintInfo();
				std::cout<<"Receive(): Packet is old"
						<<" (seqnumber="<<seqnumber<<")"<<std::endl;
				// Process more packets
				delete packetdata;
				continue;
			}
			
			// Deliver the packet if it is the one expected next
			if(seqnumber == peer->m_next_incoming_reliable_seqnum)
			{
				PrintInfo();
				std::cout<<"Receive(): Packet has the right seqnum"<<std::endl;
				// It will get returned if we do nothing here.
				// The next one is going to be one higher.
				peer->m_next_incoming_reliable_seqnum++;
			}
			// If the packet is not the one expected next, queue it
			// and get a new one.
			else
			{
				PrintInfo();
				std::cout<<"Receive(): Queueing incoming packet "
						<<"(expected seqnum="
						<<peer->m_next_incoming_reliable_seqnum
						<<", got seqnum="<<seqnumber<<")"
						<<std::endl;

				// Make packet with copied data
				BufferedPacket p;
				p.time = 0.0;
				p.data = new u8[received_size];
				memcpy(p.data, packetdata, received_size);
				p.size = received_size;

				// Find the appropriate place in the receive queue
				core::list<BufferedPacket>::Iterator i;
				i = peer->m_incoming_reliable_queue.begin();
				bool added = false;
				// If queue is empty
				if(i == peer->m_incoming_reliable_queue.end())
				{
					PrintInfo();
					std::cout<<"Receive(): Queue is empty, adding to"
							" beginning"<<std::endl;
					peer->m_incoming_reliable_queue.push_front(p);
					added = true;
				}
				/* If the sequence number of the first packet in queue is
				   higher than that of the packet we got, add the packet
				   to the beginning of the list */
				else if(readSeqNumber(i->data) > seqnumber){
					PrintInfo();
					std::cout<<"Receive(): First has higher number,"
							" adding to beginning"
							<<std::endl;
					peer->m_incoming_reliable_queue.push_front(p);
					added = true;
				}
				/*
					Else find the right place and add it there.
				*/
				else{
					u16 count = 1;
					for(;;){
						if(readSeqNumber(i->data) < seqnumber){
							PrintInfo();
							std::cout<<"Receive(): Adding after index "<<count
									<<std::endl;
							peer->m_incoming_reliable_queue.insert_after(i, p);
							added = true;
							break;
						}
						i++;
						count++;
					}
				}
				assert(added == true);

				// Process more packets
				delete packetdata;
				continue;
			}
		}

		assert(datasize >= received_size - headersize);
		u32 received_datasize = received_size - headersize;
		memcpy(data, &packetdata[headersize], received_datasize);

		delete[] packetdata;

		return received_datasize;
	}
}

void Connection::Send(u16 peer_id, u8 *data, u32 datasize, bool reliable)
{
	PrintInfo();
	std::cout<<"Send(): peer_id="<<peer_id<<", datasize="<<datasize
			<<", reliable="<<reliable<<std::endl;

	core::map<u16, Peer*>::Node *node = m_peers.find(peer_id);

	if(node == NULL){
		// Peer not found
		throw InvalidOutgoingDataException("Peer not found (possible timeout)");
	}

	Peer *peer = node->getValue();
	
	// A reality check
	assert(peer->m_id == peer_id);

	u8 type = TYPE_NORMAL;
	u16 seqnum;
	if(reliable){
		type = TYPE_RELIABLE;
		seqnum = peer->m_next_seqnum_reliable;
		peer->m_next_seqnum_reliable++;
	}
	else{
		seqnum = peer->m_next_seqnum_unreliable;
		peer->m_next_seqnum_unreliable++;
	}

	Address destination = peer->m_address;

	/*
		If it is a reliable packet, add it to m_ack_timeout_queue
	*/
	if(reliable == true)
	{
		BufferedPacket b;
		b.time = 0.0;
		b.address = destination;
		u32 packet_size = datasize + HEADER_SIZE;
		b.data = new u8[packet_size];
		b.size = packet_size;
		makePacket(b.data, data, datasize, m_protocol_id,
				m_peer_id, seqnum, type);
		peer->m_ack_timeout_queue.push_back(b);
	}
	
	ProtocolSend(destination, data, datasize, m_peer_id, seqnum, type);
}

void Connection::ProtocolSend(Address &destination, u8 *data, u32 datasize,
		u16 peer_id, u16 seqnum, u8 type)
{
	u32 packet_size = datasize + HEADER_SIZE;
	u8 packetdata[packet_size];
	makePacket(packetdata, data, datasize, m_protocol_id, peer_id,
			seqnum, type);

	RawSend(destination, packetdata, packet_size);
}

void Connection::RawSend(Address &destination, u8 *data, u32 datasize)
{
	m_socket.Send(destination, data, datasize);
}

void Connection::RunTimeouts(float dtime)
{
	/*
		1) Remove timed-out peers
		2) Send timed-out packets in m_ack_timeout_queue
	*/
	core::list<u16> timeouted_peers;
	core::map<u16, Peer*>::Iterator j;
	j = m_peers.getIterator();
	for(; j.atEnd() == false; j++)
	{
		Peer *peer = j.getNode()->getValue();

		u16 peer_id = j.getNode()->getKey();
		assert(peer_id == peer->m_id);

		// Check peer timeout
		peer->m_timeout_counter += dtime;
		// TODO: Get the constant timeout from somewhere else
		if(peer->m_timeout_counter > 10.0)
		{
			PrintInfo();
			std::cout<<"RunTimeouts(): Peer "<<peer_id
					<<" has timeouted."<<std::endl;
			// Add peer to the list
			timeouted_peers.push_back(peer_id);
			// Don't bother going through the buffers of this one
			continue;
		}

		// Ping peer at constant intervals
		peer->m_ping_timer += dtime;
		// TODO: Get the constant interval from somewhere else
		if(peer->m_ping_timer > 3.0)
		{
			// Send PING
			u16 packetseqnum = peer->m_next_seqnum_unreliable;
			peer->m_next_seqnum_unreliable++;
			u8 packettype = TYPE_PING;
			ProtocolSend(peer->m_address, NULL, 0, m_peer_id,
					packetseqnum, packettype);
		}

		// Check buffer timeouts
		core::list<BufferedPacket>::Iterator i;
		i = peer->m_ack_timeout_queue.begin();
		for(; i != peer->m_ack_timeout_queue.end(); i++)
		{
			i->time += dtime;
			// If it has run over timeout, re-send it
			// TODO: Get the constant timeout from somewhere else
			if(i->time > peer->m_resend_timeout)
			{
				PrintInfo();
				std::cout<<"RunTimeouts(): Re-sent packet."<<std::endl;
				RawSend(i->address, i->data, i->size);
				i->time = 0.0;

				peer->m_resend_timeout *= 2.0;
				if(peer->m_resend_timeout > RESEND_TIMEOUT_MAXIMUM)
					peer->m_resend_timeout = RESEND_TIMEOUT_MAXIMUM;
			}
		}
	}
	// Remove timeouted peers
	core::list<u16>::Iterator i = timeouted_peers.begin();
	for(; i != timeouted_peers.end(); i++)
	{
		PrintInfo();
		std::cout<<"RunTimeouts(): Removing peer "<<(*i)<<std::endl;
		m_peers.remove(*i);
	}
}

Peer* Connection::GetPeer(u16 peer_id)
{
	core::map<u16, Peer*>::Node *node = m_peers.find(peer_id);

	if(node == NULL){
		// Peer not found
		throw NotFoundException("Peer not found (possible timeout)");
	}

	return node->getValue();
}

void Connection::PrintInfo()
{
	std::cout<<"Connection(h="<<m_socket.GetHandle()<<", id="
			<<m_peer_id<<"): ";
}

} // namespace

