#include "client.h"
#include "utility.h"
#include <iostream>
#include "clientserver.h"
#include "jmutexautolock.h"

#ifdef _WIN32
	#include <windows.h>
	#define sleep_ms(x) Sleep(x)
#else
	#include <unistd.h>
	#define sleep_ms(x) usleep(x*1000)
#endif

void * ClientUpdateThread::Thread()
{
	ThreadStarted();

	while(getRun())
	{
		IncomingPacket packet = m_client->getPacket();
		u8 *data = packet.m_data;
		u32 datasize = packet.m_datalen;

		if(data == NULL){
			sleep_ms(200);
			continue;
		}
		
		if(datasize < 2)
			continue;
		
		ToClientCommand command = (ToClientCommand)readU16(&data[0]);
		switch(command)
		{
		case TOCLIENT_BLOCKDATA:{
			// Ignore too small packet
			if(datasize < 8 + MapBlock::serializedLength())
				continue;
				
			v3s16 p;
			p.X = readS16(&data[2]);
			p.Y = readS16(&data[4]);
			p.Z = readS16(&data[6]);
			
			std::cout<<"Client: Thread: BLOCKDATA for ("
					<<p.X<<","<<p.Y<<","<<p.Z<<")"
					<<std::endl;

			v2s16 p2d(p.X, p.Z);
			MapSector *sector = m_client->m_env.getMap().getSector(p2d);

			try{
				MapBlock *block = sector->getBlockNoCreate(p.Y);
				block->deSerialize(&data[8]);
			}
			catch(InvalidPositionException &e)
			{
				MapBlock *block = new MapBlock(&m_client->m_env.getMap(), p);
				block->deSerialize(&data[8]);
				sector->insertBlock(block);
			}
			
			}break;
		default:
			std::cout<<"Client: Thread: Ingoring unknown command "
					<<command<<std::endl;
		}

	}

	return NULL;
}

Client::Client():
	m_env(new SlaveMap(this)),
	m_thread(this),
	m_con(PROTOCOL_ID)
{
	//m_block_fetch_queue_mutex.Init();
	m_fetchblock_mutex.Init();
	m_incoming_queue_mutex.Init();
	m_con_mutex.Init();

	m_fetchblock_counter = 0;

	m_thread.Start();
}

Client::~Client()
{
	m_thread.setRun(false);
	while(m_thread.IsRunning())
		sleep_s(1);
}

void Client::connect(Address address)
{
	m_con.Connect(address);
}

void Client::step(float dtime)
{
	m_env.step(dtime);
	m_con.RunTimeouts(dtime);
	
	ReceiveAll();

	/*
		Clear old entries from fetchblock history
	*/
	{
		JMutexAutoLock lock(m_fetchblock_mutex);
		
		core::list<v3s16> remove_queue;
		core::map<v3s16, float>::Iterator i;
		i = m_fetchblock_history.getIterator();
		for(; i.atEnd() == false; i++)
		{
			float value = i.getNode()->getValue();
			value += dtime;
			i.getNode()->setValue(value);
			// Older than 5 seconds
			if(value >= 5.0)
				remove_queue.push_back(i.getNode()->getKey());
		}
		core::list<v3s16>::Iterator j;
		j = remove_queue.begin();
		for(; j != remove_queue.end(); j++)
		{
			m_fetchblock_history.remove(*j);
		}

		m_fetchblock_counter = 0;
	}

	// fetch queued blocks
	fetchBlockQueue();
}

void Client::ReceiveAll()
{
	for(;;){
		try{
			Receive();
		}
		catch(con::NoIncomingDataException &e)
		{
			break;
		}
		catch(con::InvalidIncomingDataException &e)
		{
			std::cout<<"Client::ReceiveAll(): "
					"InvalidIncomingDataException: what()="
					<<e.what()<<std::endl;
		}
	}
}

void Client::Receive()
{
	u32 data_maxsize = 10000;
	Buffer<u8> data(data_maxsize);
	u16 peer_id;
	u16 seqnumber;
	u32 datasize;
	{
		JMutexAutoLock lock(m_con_mutex);
		datasize = m_con.Receive(peer_id, *data, data_maxsize, seqnumber);
	}
	ProcessData(*data, datasize, peer_id, seqnumber);
}

void Client::ProcessData(u8 *data, u32 datasize, u16 peer_id, u16 seqnumber)
{
	// Ignore packets that don't even fit a command
	if(datasize < 2)
		return;

	ToClientCommand command = (ToClientCommand)readU16(&data[0]);
	switch(command)
	{
	// Execute fast commands straight away
	case TOCLIENT_REMOVENODE:{
		if(datasize < 8)
			return;
		v3s16 p;
		p.X = readS16(&data[2]);
		p.Y = readS16(&data[4]);
		p.Z = readS16(&data[6]);
		m_env.getMap().removeNodeAndUpdate(p);
		}
		break;

	case TOCLIENT_ADDNODE:{
		if(datasize < 8 + MapNode::serializedLength())
			return;

		v3s16 p;
		p.X = readS16(&data[2]);
		p.Y = readS16(&data[4]);
		p.Z = readS16(&data[6]);

		MapNode n;
		n.deSerialize(&data[8]);

		f32 light = m_env.getMap().getNode(p).light;
		m_env.getMap().setNode(p, n);
		m_env.getMap().nodeAddedUpdate(p, light);

		}
		break;

	// Default to queueing it (for slow commands)
	default:
	{
		JMutexAutoLock lock(m_incoming_queue_mutex);
		
		IncomingPacket packet(data, datasize);
		m_incoming_queue.push_back(packet);
	}
	}
}

void Client::Send(u8 *data, u32 datasize, bool reliable)
{
	JMutexAutoLock lock(m_con_mutex);
	m_con.Send(PEER_ID_SERVER, data, datasize, reliable);
}

void Client::fetchBlock(v3s16 p)
{
	{
		JMutexAutoLock lock2(m_con_mutex);
		if(m_con.Connected() == false)
			return;
	}

	{
		JMutexAutoLock lock(m_fetchblock_mutex);

		// If fetch request was recently sent, cancel
		if(m_fetchblock_history.find(p) != NULL)
			return;
			
		//TODO: Check that it isn't already in the queue
		m_fetchblock_queue.push_back(p);
		
		m_fetchblock_history.insert(p, 0.0);
	}

	// fetch queued blocks
	fetchBlockQueue();
}

void Client::fetchBlockQueue()
{
	JMutexAutoLock lock(m_fetchblock_mutex);

	for(;;){
		if(m_fetchblock_queue.empty())
			break;

		// Avoid flooding the connection
		if(m_fetchblock_counter >= 3)
			break;

		core::list<v3s16>::Iterator i = m_fetchblock_queue.begin();
		v3s16 p = *i;
		m_fetchblock_queue.erase(i);

		m_fetchblock_counter++;
		fetchBlockRaw(p);
	}
}

void Client::fetchBlockRaw(v3s16 p)
{
	u8 data[8];
	writeU16(&data[0], TOSERVER_GETBLOCK);
	writeS16(&data[2], p.X);
	writeS16(&data[4], p.Y);
	writeS16(&data[6], p.Z);
	Send(data, sizeof(data), false);
}

IncomingPacket Client::getPacket()
{
	JMutexAutoLock lock(m_incoming_queue_mutex);
	
	core::list<IncomingPacket>::Iterator i;
	// Refer to first one
	i = m_incoming_queue.begin();

	// If queue is empty, return empty packet
	if(i == m_incoming_queue.end()){
		IncomingPacket packet;
		return packet;
	}
	
	// Pop out first packet and return it
	IncomingPacket packet = *i;
	m_incoming_queue.erase(i);
	return packet;
}

void Client::removeNode(v3s16 nodepos)
{
	u8 data[8];
	writeU16(&data[0], TOSERVER_REMOVENODE);
	writeS16(&data[2], nodepos.X);
	writeS16(&data[4], nodepos.Y);
	writeS16(&data[6], nodepos.Z);
	Send(data, sizeof(data), true);
}

void Client::addNode(v3s16 nodepos, MapNode n)
{
	u8 datasize = 8 + MapNode::serializedLength();
	//u8 data[datasize];
	Buffer<u8> data(datasize);
	writeU16(&data[0], TOSERVER_ADDNODE);
	writeS16(&data[2], nodepos.X);
	writeS16(&data[4], nodepos.Y);
	writeS16(&data[6], nodepos.Z);
	n.serialize(&data[8]);
	Send(*data, datasize, false);
}


