#ifndef CONNECTION_HEADER
#define CONNECTION_HEADER

#include <assert.h>
#include "common_irrlicht.h"
#include "socket.h"
#include "utility.h"
#include "exceptions.h"
#include <iostream>

namespace con
{

class NotFoundException : public BaseException
{
public:
	NotFoundException(const char *s):
		BaseException(s)
	{}
};

class ConnectionException : public BaseException
{
public:
	ConnectionException(const char *s):
		BaseException(s)
	{}
};

class InvalidIncomingDataException : public BaseException
{
public:
	InvalidIncomingDataException(const char *s):
		BaseException(s)
	{}
};

class InvalidOutgoingDataException : public BaseException
{
public:
	InvalidOutgoingDataException(const char *s):
		BaseException(s)
	{}
};

class NoIncomingDataException : public BaseException
{
public:
	NoIncomingDataException(const char *s):
		BaseException(s)
	{}
};

inline u16 readPeerId(u8 *packetdata)
{
	return readU16(&packetdata[4]);
}
inline u16 readSeqNumber(u8 *packetdata)
{
	return readU16(&packetdata[6]);
}
inline u8 readType(u8 *packetdata)
{
	return readU8(&packetdata[8]);
}

struct BufferedPacket
{
	BufferedPacket(u8 *a_data, u32 a_size):
		data(a_data, a_size), time(0.0)
	{}
	BufferedPacket(u32 a_size):
		data(a_size), time(0.0)
	{}
	SharedBuffer<u8> data; // Data of the packet, including headers
	float time; // Seconds from buffering the packet
	Address address; // Sender or destination
};

BufferedPacket makePacket(Address &address, u8 *data, u32 datasize,
		u32 protocol_id, u16 peer_id, u16 seqnum, u8 type);

struct IncomingSplitPacket
{
	// Key is chunk number, value is data without headers
	core::map<u16, SharedBuffer<u8> > chunks;
	u32 chunk_count;

	bool allReceived()
	{
		return (chunks.size() == chunk_count);
	}
};

/*
	Connection packet header format:
		u32 protocol_id [0-3]
		u16 peer_id [4-5]
		u16 sequence_num [6-7]
		u8 type [8]
	
	peer_id:
		Unique to each peer.
		value 0 is reserved for making new connections
		value 1 is reserved for server
	sequence_num:
		different for reliable and unreliable packets.
	
	ACK packet
	- sequence_num is unique
	- ACKed sequence_num is stored in first bytes of payload data.

	RELIABLE packet
	- After receiving, an ACK packet is sent.
	- Reliable packets are delivered in the order they were sent.
	- If a reliable packet is re-sent, the sequence number is kept
	  the same.

	PEER_ID packet
	- Sets client peer_id
	- New peer id is stored in first two bytes of data.
*/

#define HEADER_SIZE (4+2+2+1)

#define TYPE_NORMAL 0
#define TYPE_RELIABLE 1
#define TYPE_ACK 2
#define TYPE_PEER_ID 3
#define TYPE_PING 4
#define TYPE_PONG 5
#define TYPE_SPLIT 6
#define TYPE_SPLIT_ACK 7

#define PEER_ID_NEW 0
#define PEER_ID_SERVER 1

//#define SEQNUM_INITIAL 65500
#define SEQNUM_INITIAL 0x10

#define RESEND_TIMEOUT_MINIMUM 0.2
#define RESEND_TIMEOUT_MAXIMUM 5.0

struct Peer
{
	Peer()
	{
		m_next_seqnum_unreliable = SEQNUM_INITIAL;
		m_next_seqnum_reliable = SEQNUM_INITIAL;
		m_next_seqnum_split = SEQNUM_INITIAL;
		m_next_incoming_reliable_seqnum = SEQNUM_INITIAL;
		m_timeout_counter = 0.0;
		m_resend_timeout = RESEND_TIMEOUT_MINIMUM;
	}
	~Peer()
	{
		core::map<u16, IncomingSplitPacket*>::Iterator i;
		i = m_incoming_split_queue.getIterator();
		for(; i.atEnd() == false; i++)
		{
			delete i.getNode()->getValue();
		}
	}
	u16 m_next_seqnum_unreliable;
	u16 m_next_seqnum_reliable;
	u16 m_next_seqnum_split;

	// seqnum of the next incoming reliable packet
	u16 m_next_incoming_reliable_seqnum;

	// A list of sent reliables that haven't received an ACK.
	// when an ack is received, the corresponding item is removed.
	// If no ack is received in some time, the packet is re-sent.
	core::list<BufferedPacket> m_ack_timeout_queue;
	// A list of received reliables whose sequence number is higher than
	// what is expected next.
	core::list<BufferedPacket> m_incoming_reliable_queue;
	// A list of sent split packet chunks that haven't received an ACK
	core::list<BufferedPacket> m_split_ack_timeout_queue;
	// key is seqnum
	// TODO: Incomplete ones are never cleared ATM
	core::map<u16, IncomingSplitPacket*> m_incoming_split_queue;

	// Address of the peer
	Address m_address;
	// Unique id of the peer
	u16 m_id;
	// Seconds from last receive
	float m_timeout_counter;
	// Ping timer
	float m_ping_timer;
	// This is changed dynamically
	float m_resend_timeout;
};

class Connection
{
public:
	Connection(u32 protocol_id);
	~Connection();
	// Start being a server
	void Serve(unsigned short port);
	// Connect to a server
	void Connect(Address address);
	bool Connected();
	u32 Receive(u16 &peer_id, u8 *data, s32 datasize, u16 &seqnumber);
	void Send(u16 peer_id, u8 *data, u32 datasize, bool reliable=false);
	void SendToAll(u8 *data, u32 datasize, bool reliable=false);
	void SendSplit(u16 peer_id, u8 *data, u32 datasize);
	void ProtocolSend(Address &destination, u8 *data, u32 datasize,
			u16 peer_id, u16 seqnum, u8 type);
	//void RawSend(Address &destination, u8 *data, u32 datasize);
	void RawSend(BufferedPacket &packet);
	void RunTimeouts(float dtime);
	Peer* GetPeer(u16 peer_id);
private:
	void PrintInfo(); // For debug printing
	u32 m_protocol_id;
	core::map<u16, Peer*> m_peers;
	u16 m_peer_id;
	bool m_waiting_new_peer_id;
	UDPSocket m_socket;
};

} // namespace

#endif

