#include "server.h"
#include "utility.h"
#include <iostream>
#include "clientserver.h"
#include "map.h"
#include "jmutexautolock.h"

#ifdef _WIN32
	#include <windows.h>
	#define sleep_ms(x) Sleep(x)
#else
	#include <unistd.h>
	#define sleep_ms(x) usleep(x*1000)
#endif

#define THREADED 0

void * ServerNetworkThread::Thread()
{
	ThreadStarted();

	if(THREADED == false)
		return NULL;
	
	while(getRun())
	{
		m_server->RunConTimeouts();
		
		try{
			//std::cout<<"Running m_server->Receive()"<<std::endl;
			m_server->Receive();
		}
		catch(con::NoIncomingDataException &e)
		{
		}
		catch(std::exception &e)
		{
			std::cout<<"ServerNetworkThread: Some exception: "
					<<e.what()<<std::endl;
		}
	}
	

	return NULL;
}


Server::Server():
	m_env(new MasterMap()),
	m_con(PROTOCOL_ID, 512),
	m_thread(this)
{
	m_env_mutex.Init();
	m_con_mutex.Init();
	m_con_dtime_mutex.Init();
	m_con_dtime = 0.0;
}

Server::~Server()
{
	stop();
}

void Server::start(unsigned short port)
{
	stop();
	if(THREADED)
		m_con.setTimeoutMs(200);
	else
		m_con.setTimeoutMs(0);
	m_con.Serve(port);
	m_thread.setRun(true);
	if(THREADED)
		m_thread.Start();
}

void Server::stop()
{
	m_thread.setRun(false);
	while(m_thread.IsRunning())
		sleep_ms(100);
}

void Server::step(float dtime)
{
	{
		JMutexAutoLock lock(m_env_mutex);
		m_env.step(dtime);
	}
	{
		JMutexAutoLock lock(m_con_dtime_mutex);
		m_con_dtime += dtime;
	}
	
	if(THREADED == false)
	{
		// Test code for testing non-threaded use

		RunConTimeouts();
		
		for(;;)
		{
			try{
				//std::cout<<"Running Receive()"<<std::endl;
				Receive();
			}
			catch(con::NoIncomingDataException &e)
			{
				break;
			}
			catch(std::exception &e)
			{
				std::cout<<"Server Receive: Some exception: "
						<<e.what()<<std::endl;
				break;
			}
		}
	}
}

void Server::RunConTimeouts()
{
	float dtime;
	{
		JMutexAutoLock lock1(m_con_dtime_mutex);
		dtime = m_con_dtime;
		m_con_dtime = 0.0;
	}
	JMutexAutoLock lock2(m_con_mutex);
	m_con.RunTimeouts(dtime);
}

void Server::Receive()
{
	u32 data_maxsize = 10000;
	Buffer<u8> data(data_maxsize);
	u16 peer_id;
	u32 datasize;
	try{
		{
			JMutexAutoLock lock(m_con_mutex);
			datasize = m_con.Receive(peer_id, *data, data_maxsize);
		}
		ProcessData(*data, datasize, peer_id);
	}
	catch(con::InvalidIncomingDataException &e)
	{
		std::cout<<"Server::Receive(): "
				"InvalidIncomingDataException: what()="
				<<e.what()<<std::endl;
	}
}

void Server::ProcessData(u8 *data, u32 datasize, u16 peer_id)
{
	JMutexAutoLock envlock(m_env_mutex);
	JMutexAutoLock conlock(m_con_mutex);
	
	try
	{

	if(datasize < 2)
		return;

	ToServerCommand command = (ToServerCommand)readU16(&data[0]);

	switch(command)
	{
	case TOSERVER_GETBLOCK:{
		//throw NotImplementedException("");

		// Check for too short data
		if(datasize < 8)
			return;
		std::cout<<"Server::ProcessData(): GETBLOCK"
				<<std::endl;
		/*
			Get block data and send it
		*/
		v3s16 p;
		p.X = readS16(&data[2]);
		p.Y = readS16(&data[4]);
		p.Z = readS16(&data[6]);
		MapBlock *block = m_env.getMap().getBlock(p);
	
		u32 replysize = 8 + MapBlock::serializedLength();
		SharedBuffer<u8> reply(replysize);
		writeU16(&reply[0], TOCLIENT_BLOCKDATA);
		writeS16(&reply[2], p.X);
		writeS16(&reply[4], p.Y);
		writeS16(&reply[6], p.Z);
		block->serialize(&reply[8]);
		// Send as unreliable
		//m_con.Send(peer_id, 1, reply, false);
		m_con.Send(peer_id, 1, reply, true);
		}break;
	case TOSERVER_REMOVENODE:{
		if(datasize < 8)
			return;
			
		v3s16 p;
		p.X = readS16(&data[2]);
		p.Y = readS16(&data[4]);
		p.Z = readS16(&data[6]);

		MapNode n;
		n.d = MATERIAL_AIR;
		m_env.getMap().setNode(p, n);

		u32 replysize = 8;
		//u8 reply[replysize];
		SharedBuffer<u8> reply(replysize);
		writeU16(&reply[0], TOCLIENT_REMOVENODE);
		writeS16(&reply[2], p.X);
		writeS16(&reply[4], p.Y);
		writeS16(&reply[6], p.Z);
		// Send as reliable
		m_con.SendToAll(0, reply, true);
		}break;
	case TOSERVER_ADDNODE:{
		if(datasize < 8 + MapNode::serializedLength())
			return;

		v3s16 p;
		p.X = readS16(&data[2]);
		p.Y = readS16(&data[4]);
		p.Z = readS16(&data[6]);

		MapNode n;
		n.deSerialize(&data[8]);
		m_env.getMap().setNode(p, n);

		u32 replysize = 8 + MapNode::serializedLength();
		//u8 reply[replysize];
		SharedBuffer<u8> reply(replysize);
		writeU16(&reply[0], TOCLIENT_ADDNODE);
		writeS16(&reply[2], p.X);
		writeS16(&reply[4], p.Y);
		writeS16(&reply[6], p.Z);
		n.serialize(&reply[8]);
		// Send as reliable
		m_con.SendToAll(0, reply, true);
	}break;
	default:
		std::cout<<"Server::ProcessData(): Ingoring unknown command "
				<<command<<std::endl;
	} //switch
	
	} //try
	/*catch(con::ThrottlingException &e)
	{
		std::cout<<"Server::ProcessData(): Cancelling due to "
				<<"bandwidth throttling"<<std::endl;
	}*/
	catch(SendFailedException &e)
	{
		std::cout<<"Server::ProcessData(): SendFailedException: "
				<<"what="<<e.what()
				<<std::endl;
	}
}


