/*
(c) 2010 Perttu Ahola <celeron55@gmail.com>
*/

#include "map.h"
//#include "player.h"
#include "main.h"
#include "jmutexautolock.h"
#include "client.h"
#include "filesys.h"
#include "utility.h"

#ifdef _WIN32
	#include <windows.h>
	#define sleep_ms(x) Sleep(x)
#else
	#include <unistd.h>
	#define sleep_ms(x) usleep(x*1000)
#endif

void * MapUpdateThread::Thread()
{
	if(map == NULL)
		return NULL;
		
	ThreadStarted();

	while(getRun())
	{
		//std::cout<<DTIME<<"UpdateThread running"<<std::endl;
		
		bool did_something;
		did_something = map->updateChangedVisibleArea();
		
		if(did_something == false)
			sleep_ms(500);
	}

	//std::cout<<DTIME<<"UpdateThread stopped"<<std::endl;

	return NULL;
}

Map::Map(std::ostream &dout):
	m_dout(dout),
	camera_position(0,0,0),
	camera_direction(0,0,1),
	updater(this),
	m_sector_cache(NULL),
	m_hwrapper(this),
	drawoffset(0,0,0)
{
	m_getsector_mutex.Init();
	m_gensector_mutex.Init();
	camera_mutex.Init();
	assert(m_getsector_mutex.IsInitialized());
	assert(m_gensector_mutex.IsInitialized());
	assert(camera_mutex.IsInitialized());
	
	// Get this so that the player can stay on it at first
	//getSector(v2s16(0,0));
}

Map::~Map()
{
	/*
		Stop updater thread
	*/
	updater.setRun(false);
	while(updater.IsRunning())
		sleep_s(1);

	/*
		Free all MapSectors.
	*/
	core::map<v2s16, MapSector*>::Iterator i = m_sectors.getIterator();
	for(; i.atEnd() == false; i++)
	{
		MapSector *sector = i.getNode()->getValue();
		delete sector;
	}
}

/*
	TODO: These mutexes need rethinking. They don't work properly
	      at the moment. Maybe. Or do they?
*/

/*bool Map::sectorExists(v2s16 p)
{
	JMutexAutoLock lock(m_getsector_mutex);
	core::map<v2s16, MapSector*>::Node *n = m_sectors.find(p);
	return (n != NULL);
}*/

MapSector * Map::getSectorNoGenerate(v2s16 p)
{
	//JMutexAutoLock lock(m_getsector_mutex);

	if(m_sector_cache != NULL && p == m_sector_cache_p){
		MapSector * sector = m_sector_cache;
		// Reset inactivity timer
		sector->usage_timer = 0.0;
		return sector;
	}
	
	core::map<v2s16, MapSector*>::Node *n = m_sectors.find(p);
	// If sector doesn't exist, throw an exception
	if(n == NULL)
	{
		/*
			TODO: Check if sector is stored on disk
			      and load dynamically
		*/
		//sector = NULL;
		throw InvalidPositionException();
	}
	
	MapSector *sector = n->getValue();
	
	// Cache the last result
	m_sector_cache_p = p;
	m_sector_cache = sector;

	//MapSector * ref(sector);
	
	return sector;
}

/*
	This is overloaded by ClientMap and ServerMap to allow
	their differing fetch methods.
*/
MapSector * Map::getSector(v2s16 p2d, bool generate)
{
	return getSectorNoGenerate(p2d);
}

MapBlock * Map::getBlockNoCreate(v3s16 p3d)
{	
	/*m_dout<<DTIME<<"Map::getBlockNoCreate(("
			<<p3d.X<<","<<p3d.Y<<","<<p3d.Z
			<<")"<<std::endl;*/
	v2s16 p2d(p3d.X, p3d.Z);
	//v2s16 sectorpos = getNodeSectorPos(p2d);
	MapSector * sector = getSectorNoGenerate(p2d);

	//JMutexAutoLock(sector->mutex);

	return sector->getBlockNoCreate(p3d.Y);
}

MapBlock * Map::getBlock(v3s16 p3d, bool generate)
{
	v2s16 p2d(p3d.X, p3d.Z);
	MapSector * sector = getSector(p2d, generate);

	if(sector == NULL)
		throw InvalidPositionException();

	//JMutexAutoLock(sector->mutex);

	MapBlock * blockref = sector->getBlock(p3d.Y, generate);
	
	/*m_dout<<DTIME<<"Map::GetBlock(): blockref->getRefcount()="
			<<blockref->getRefcount()
			<<std::endl;*/

	/*m_dout<<DTIME<<"Map::GetBlock(): "
			<<"Got block. Now returning."<<std::endl;*/
	
	return blockref;
}

f32 Map::getGroundHeight(v2s16 p, bool generate)
{
	/*m_dout<<DTIME<<"Map::getGroundHeight(("
			<<p.X<<","<<p.Y<<"))"<<std::endl;*/
	try{
		v2s16 sectorpos = getNodeSectorPos(p);
		MapSector * sref = getSectorNoGenerate(sectorpos);
		v2s16 relpos = p - sectorpos * MAP_BLOCKSIZE;
		//sref->mutex.Lock();
		f32 y = sref->getGroundHeight(relpos);
		//sref->mutex.Unlock();
		//m_dout<<DTIME<<"[found]"<<std::endl;
		return y;
	}
	catch(InvalidPositionException &e)
	{
		//m_dout<<DTIME<<"[not found]"<<std::endl;
		return GROUNDHEIGHT_NOTFOUND_SETVALUE;
	}
}

void Map::setGroundHeight(v2s16 p, f32 y, bool generate)
{
	/*m_dout<<DTIME<<"Map::setGroundHeight(("
			<<p.X<<","<<p.Y
			<<"), "<<y<<")"<<std::endl;*/
	v2s16 sectorpos = getNodeSectorPos(p);
	MapSector * sref = getSectorNoGenerate(sectorpos);
	v2s16 relpos = p - sectorpos * MAP_BLOCKSIZE;
	//sref->mutex.Lock();
	sref->setGroundHeight(relpos, y);
	//sref->mutex.Unlock();
}

bool Map::isNodeUnderground(v3s16 p)
{
	v3s16 blockpos = getNodeBlockPos(p);
	try{
		MapBlock * block = getBlockNoCreate(blockpos);
		return block->getIsUnderground();
	}
	catch(InvalidPositionException &e)
	{
		return false;
	}
}

/*
	TODO: Check the order of changing lighting and recursing in
	these functions (choose the faster one)

	TODO: It has to done like this: cache the neighbours that were
	changed; then recurse to them.
*/

/*
	Goes recursively through the neighbours of the node.

	Alters only transparent nodes.

	If the lighting of the neighbour is lower than the lighting of
	the node was (before changing it to 0 at the step before), the
	lighting of the neighbour is set to 0 and then the same stuff
	repeats for the neighbour.

	Some things are made strangely to make it as fast as possible.

	Usage:
	core::list<v3s16> light_sources;
	core::map<v3s16, MapBlock*> modified_blocks;
	f32 oldlight = node_at_pos.light;
	node_at_pos.light = 0;
	unLightNeighbors(pos, oldlight, light_sources, modified_blocks);
*/
void Map::unLightNeighbors(v3s16 pos, f32 oldlight,
		core::list<v3s16> & light_sources,
		core::map<v3s16, MapBlock*>  & modified_blocks)
{
	v3s16 dirs[6] = {
		v3s16(0,0,1), // back
		v3s16(0,1,0), // top
		v3s16(1,0,0), // right
		v3s16(0,0,-1), // front
		v3s16(0,-1,0), // bottom
		v3s16(-1,0,0), // left
	};
	
	// Loop through 6 neighbors
	for(u16 i=0; i<6; i++){
		// Get the position of the neighbor node
		v3s16 n2pos = pos + dirs[i];
		
		// Get the block where the node is located
		v3s16 blockpos = getNodeBlockPos(n2pos);
		MapBlock *block;
		try{
			block = getBlockNoCreate(blockpos);
		}
		catch(InvalidPositionException &e)
		{
			// If block is inexistent, move to next one.
			continue;
		}
		
		// Find if block is in cache container
		core::map<v3s16, MapBlock * >::Node *cacheblocknode;
		cacheblocknode = modified_blocks.find(blockpos);
		
		// If the block is not found in the cache
		if(cacheblocknode == NULL)
		{
			/*
				Add block to cache container. It could be a 'set' but
				there is no such thing in Irrlicht.
				
				We can use the pointer as the value of a map just fine,
				it gets nicely cached that way, too.
			*/
			modified_blocks.insert(blockpos, block);
		}

		// Calculate relative position in block
		v3s16 relpos = n2pos - blockpos * MAP_BLOCKSIZE;
		// Get node straight from the block
		//MapNode *n2 = block->getNodePtr(relpos);
		MapNode n2 = block->getNode(relpos);
		
		/*
			If the neighbor is dimmer than what was specified
			as oldlight (the light of the previous node)...
		*/
		/*if(n2->light < oldlight - 0.001){
			if(n2->transparent()){
				light_t current_light = n2->light;
				n2->light = 0.0;*/
		if(n2.light < oldlight - 0.001){
			if(n2.transparent()){
				light_t current_light = n2.light;
				n2.light = 0.0;
				block->setNode(relpos, n2);
				unLightNeighbors(n2pos, current_light,
						light_sources, modified_blocks);
			}
		}
		else{
			light_sources.push_back(n2pos);
		}
	}
}

/*
	Goes recursively through the neighbours of the node.

	Alters only transparent nodes.

	If the lighting of the neighbour is lower than the calculated
	lighting coming from the current node to it, the lighting of
	the neighbour is set and the same thing repeats.

	Some things are made strangely to make it as fast as possible.

	Usage:
	core::map<v3s16, MapBlock*> modified_blocks;
	lightNeighbors(pos, node_at_pos.light, modified_blocks);
*/
void Map::lightNeighbors(v3s16 pos,
		core::map<v3s16, MapBlock*> & modified_blocks)
{
	v3s16 dirs[6] = {
		v3s16(0,0,1), // back
		v3s16(0,1,0), // top
		v3s16(1,0,0), // right
		v3s16(0,0,-1), // front
		v3s16(0,-1,0), // bottom
		v3s16(-1,0,0), // left
	};
	
	core::list<v3s16> neighbour_cache;
	
	/*
		Initialize block cache
		(using center node as a starting position)
	*/
	v3s16 blockpos_last = getNodeBlockPos(pos);
	MapBlock *block = NULL;
	try{
		block = getBlockNoCreate(blockpos_last);
	}
	catch(InvalidPositionException &e)
	{
		return;
	}
	
	// Calculate relative position in block
	v3s16 relpos = pos - blockpos_last * MAP_BLOCKSIZE;
	// Get node straight from the block
	//MapNode *n = block->getNodePtr(relpos);
	MapNode n = block->getNode(relpos);

	//f32 oldlight = n->light;
	f32 oldlight = n.light;
	f32 newlight = oldlight * LIGHT_DIMINISH_FACTOR;

	// Loop through 6 neighbors
	for(u16 i=0; i<6; i++){
		// Get the position of the neighbor node
		v3s16 n2pos = pos + dirs[i];
		
		// Get the block where the node is located
		v3s16 blockpos = getNodeBlockPos(n2pos);

		try
		{
			/*
				Only fetch a new block if the block position has changed
			*/
			if(blockpos != blockpos_last){
				block = getBlockNoCreate(blockpos);
			}
			blockpos_last = blockpos;

			// Find if block is in cache container
			core::map<v3s16, MapBlock * >::Node *cacheblocknode;
			cacheblocknode = modified_blocks.find(blockpos);
			
			// If the block is not found in the cache
			if(cacheblocknode == NULL)
			{
				/*
					Add block to cache container. It could be a 'set' but
					there is no such thing in Irrlicht.
					
					We can use the pointer as the value of a map just fine,
					it gets nicely cached that way, too.
				*/
				modified_blocks.insert(blockpos, block);
			}
			
			// Calculate relative position in block
			v3s16 relpos = n2pos - blockpos * MAP_BLOCKSIZE;
			// Get node straight from the block
			//MapNode *n2 = block->getNodePtr(relpos);
			MapNode n2 = block->getNode(relpos);
			
			/*
				If the neighbor is much brighter than the current node,
				add to cache
			*/
			/*if(n2->light_source() > oldlight / LIGHT_DIMINISH_FACTOR + 0.01)
			{
				n2->light = n2->light_source();*/
			if(n2.light_source() > oldlight / LIGHT_DIMINISH_FACTOR + 0.01)
			{
				n2.light = n2.light_source();
				block->setNode(relpos, n2);
				neighbour_cache.push_back(n2pos);
			}
			/*
				If the neighbor is dimmer than what was specified
				as oldlight (the light of the previous node)...
			*/
			/*if(n2->light < newlight - 0.001)
			{
				if(n2->transparent())
				{
					n2->light = newlight;*/
			if(n2.light < newlight - 0.001)
			{
				if(n2.transparent())
				{
					n2.light = newlight;
					block->setNode(relpos, n2);
					// Cache and recurse at last step for maximum speed
					neighbour_cache.push_back(n2pos);
				}
			}
		}
		catch(InvalidPositionException &e)
		{
			continue;
		}
	}
	
	// Light the neighbors of every block in cache
	core::list<v3s16>::Iterator j = neighbour_cache.begin();
	for(; j != neighbour_cache.end(); j++){
		lightNeighbors(*j, modified_blocks);
	}
}

v3s16 Map::getBrightestNeighbour(v3s16 p)
{
	v3s16 dirs[6] = {
		v3s16(0,0,1), // back
		v3s16(0,1,0), // top
		v3s16(1,0,0), // right
		v3s16(0,0,-1), // front
		v3s16(0,-1,0), // bottom
		v3s16(-1,0,0), // left
	};
	
	f32 brightest_light = -1.0;
	v3s16 brightest_pos(0,0,0);

	// Loop through 6 neighbors
	for(u16 i=0; i<6; i++){
		// Get the position of the neighbor node
		v3s16 n2pos = p + dirs[i];
		MapNode n2;
		try{
			n2 = getNode(n2pos);
		}
		catch(InvalidPositionException &e)
		{
			continue;
		}
		if(n2.light > brightest_light){
			brightest_light = n2.light;
			brightest_pos = n2pos;
		}
	}
	if(brightest_light < -0.001)
		throw;
	return brightest_pos;
}

/*
	Propagates sunlight down from a node.
	Starting point gets sunlight.

	Returns the lowest y value of where the sunlight went.
*/
s16 Map::propagateSunlight(v3s16 start,
		core::map<v3s16, MapBlock*> & modified_blocks)
{
	s16 y = start.Y;
	for(; ; y--)
	{
		v3s16 pos(start.X, y, start.Z);
		
		v3s16 blockpos = getNodeBlockPos(pos);
		MapBlock *block;
		try{
			block = getBlockNoCreate(blockpos);
		}
		catch(InvalidPositionException &e)
		{
			break;
		}

		v3s16 relpos = pos - blockpos*MAP_BLOCKSIZE;
		//MapNode *n = block->getNodePtr(relpos);
		MapNode n = block->getNode(relpos);

		/*if(n->transparent())
		{
			n->light = 1.0;*/

		if(n.transparent())
		{
			n.light = 1.0;
			block->setNode(relpos, n);

			modified_blocks.insert(blockpos, block);
		}
		else{
			break;
		}
	}
	return y + 1;
}
		
void Map::updateLighting(core::list< MapBlock*> & a_blocks,
		core::map<v3s16, MapBlock*> & modified_blocks)
{
	m_dout<<DTIME<<"Map::updateLighting(): "
			<<a_blocks.getSize()<<" blocks... ";

	core::list< MapBlock * >::Iterator i = a_blocks.begin();
	
	/*m_dout<<DTIME<<"Map::updateLighting(): "
			"Flooding direct sunlight"<<std::endl;*/
	
	/*
		Propagate sunlight through the blocks
	*/
	core::map<v3s16, MapBlock*> propagated_blocks;
	for(; i != a_blocks.end(); i++)
	{
		MapBlock *block = *i;
		for(;;)
		{
			v3s16 pos = block->getPosRelative();
			modified_blocks.insert(pos, block);
			bool bottom_valid = block->propagateSunlight(propagated_blocks);
			// If bottom is valid, we're done.
			if(bottom_valid)
				break;
			// Else get the block below and loop to it, if it exists.
			pos.Y--;
			try{
				block = getBlockNoCreate(pos);
			}
			catch(InvalidPositionException &e)
			{
				break;
			}
		}
	}

	// Add other light sources here somewhere (or in the latter loop)

	/*for(; i != a_blocks.end(); i++)
	{
		MapBlock *block = *i;
	}*/
	
	//

	/*m_dout<<DTIME<<"Map::updateLighting(): "
			"Spreading light to remaining air"<<std::endl;*/

	/*
		Spread light to air that hasn't got light
		
		TODO: This shouldn't have to be done to every block
		      Though it doesn't hurt much i guess.
		
		TODO: Would adding one to every edge fix the problem
		      of getting dark places in recently loaded blocks?
	*/
	
#if 0
	for(; i != a_blocks.end(); i++)
	{
		MapBlock *block = *i;
#endif
#if 1
	/*
		TODO: It might be a wise idea to use a separate modified_blocks2
		for propagateSunlight() above which would only contain the blocks
		that it did modify.
		NOTE: This has the problem that then a newly added block cannot
		propagate light to a block that already was added.
	*/
	core::map<v3s16, MapBlock * >::Iterator i2;
	i2 = propagated_blocks.getIterator();
	m_dout<<DTIME<<"propagated_blocks= ";
	for(; i2.atEnd() == false; i2++)
	{
		MapBlock *block = i2.getNode()->getValue();
#endif
		v3s16 pos = block->getPosRelative();
		m_dout<<"("<<pos.X<<","<<pos.Y<<","<<pos.Z<<") ";
		/*s16 xmin = pos.X;
		s16 zmin = pos.Z;
		s16 ymin = pos.Y;
		s16 xmax = pos.X + MAP_BLOCKSIZE - 1;
		s16 zmax = pos.Z + MAP_BLOCKSIZE - 1;
		s16 ymax = pos.Y + MAP_BLOCKSIZE - 1;*/
		s16 xmin = pos.X - 1;
		s16 zmin = pos.Z - 1;
		s16 ymin = pos.Y - 1;
		s16 xmax = pos.X + MAP_BLOCKSIZE - 1 + 1;
		s16 zmax = pos.Z + MAP_BLOCKSIZE - 1 + 1;
		s16 ymax = pos.Y + MAP_BLOCKSIZE - 1 + 1;
		for(s16 z=zmin; z<=zmax; z++){
			for(s16 x=xmin; x<=xmax; x++){
				for(s16 y=ymax; y>=ymin; y--){
					v3s16 pos(x, y, z);
					MapNode n;
					try{
						n = getNode(pos);
					}
					catch(InvalidPositionException &e)
					{
						continue;
					}
					/*
						If it sources brighter light than there already is,
						update the lightning.
					*/
					if(n.light < n.light_source())
					{
						n.light = n.light_source();
						setNode(pos, n);
					}
					/*
						If there is light in a block, light neighbors
					*/
					if(n.light > NO_LIGHT_MAX){
						v3s16 pos(x,y,z);
						lightNeighbors(pos, modified_blocks);
					}
				}
			}
		}
		/*m_dout<<"X";
		m_dout.flush();*/
	}
	//m_dout<<DTIME<<"Done"<<std::endl;
	m_dout<<"Done ("<<getTimestamp()<<")"<<std::endl;
}

/*
	This is called after changing a node from transparent to opaque.
	The lighting value of the node should be left as-is after changing
	other values. This sets the lighting value to 0.
*/
void Map::nodeAddedUpdate(v3s16 p, f32 lightwas)
{
	m_dout<<DTIME<<"Map::nodeAddedUpdate()"<<std::endl;

	core::list<v3s16> light_sources;
	core::map<v3s16, MapBlock*> modified_blocks;
	//MapNode n = getNode(p);

	/*
		From this node to nodes underneath:
		If lighting is sunlight (1.0), unlight neighbours and
		set lighting to 0.
		Else discontinue.
	*/

	bool node_under_sunlight = true;
	
	v3s16 toppos = p + v3s16(0,1,0);

	/*
		If there is a node at top and it doesn't have sunlight,
		there has not been any sunlight going down.

		Otherwise there probably is.
	*/
	try{
		MapNode topnode = getNode(toppos);

		if(topnode.light < 0.999)
			node_under_sunlight = false;
	}
	catch(InvalidPositionException &e)
	{
	}

	// Add the block of the added node to modified_blocks
	v3s16 blockpos = getNodeBlockPos(p);
	MapBlock * blockref = getBlockNoCreate(blockpos);
	MapBlock *block = blockref;
	assert(block != NULL);
	modified_blocks.insert(blockpos, block);
	
	// Unlight neighbours of node.
	// This means setting light of all consequent dimmer nodes
	// to 0.
	
	if(isValidPosition(p) == false)
		throw;
		
	unLightNeighbors(p, lightwas, light_sources, modified_blocks);

	MapNode n = getNode(p);
	n.light = 0;
	setNode(p, n);
	
	if(node_under_sunlight)
	{
		s16 y = p.Y - 1;
		for(;; y--){
			//m_dout<<DTIME<<"y="<<y<<std::endl;
			v3s16 n2pos(p.X, y, p.Z);
			
			MapNode n2;
			try{
				n2 = getNode(n2pos);
			}
			catch(InvalidPositionException &e)
			{
				break;
			}

			if(n2.light >= 0.999){
				//m_dout<<DTIME<<"doing"<<std::endl;
				unLightNeighbors(n2pos, n2.light, light_sources, modified_blocks);
				n2.light = 0;
				setNode(n2pos, n2);
			}
			else
				break;
		}
	}
	
	core::list<v3s16>::Iterator j = light_sources.begin();
	for(; j != light_sources.end(); j++)
	{
		lightNeighbors(*j, modified_blocks);
	}

	core::map<v3s16, MapBlock * >::Iterator i = modified_blocks.getIterator();
	for(; i.atEnd() == false; i++)
	{
		i.getNode()->getValue()->updateMesh();
	}
}

/*
*/
void Map::removeNodeAndUpdate(v3s16 p)
{
	m_dout<<DTIME<<"Map::removeNodeAndUpdate()"<<std::endl;
	
	core::map<v3s16, MapBlock*> modified_blocks;

	bool node_under_sunlight = true;
	
	v3s16 toppos = p + v3s16(0,1,0);
	
	/*
		If there is a node at top and it doesn't have sunlight,
		there will be no sunlight going down.
	*/
	try{
		MapNode topnode = getNode(toppos);

		if(topnode.light < 0.999)
			node_under_sunlight = false;
	}
	catch(InvalidPositionException &e)
	{
	}

	/*
		Unlight neighbors (in case the node is a light source)
	*/
	core::list<v3s16> light_sources;
	unLightNeighbors(p, getNode(p).light,
			light_sources, modified_blocks);

	/*
		Remove the node
	*/
	MapNode n;
	n.d = MATERIAL_AIR;
	n.light = LIGHT_MIN;
	setNode(p, n);
	
	/*
		Recalculate lighting
	*/
	core::list<v3s16>::Iterator j = light_sources.begin();
	for(; j != light_sources.end(); j++)
	{
		lightNeighbors(*j, modified_blocks);
	}

	// Add the block of the removed node to modified_blocks
	v3s16 blockpos = getNodeBlockPos(p);
	MapBlock * blockref = getBlockNoCreate(blockpos);
	MapBlock *block = blockref;
	assert(block != NULL);
	modified_blocks.insert(blockpos, block);

	if(node_under_sunlight)
	{
		s16 ybottom = propagateSunlight(p, modified_blocks);
		/*m_dout<<DTIME<<"Node was under sunlight. "
				"Propagating sunlight";
		m_dout<<DTIME<<" -> ybottom="<<ybottom<<std::endl;*/
		s16 y = p.Y;
		for(; y >= ybottom; y--)
		{
			v3s16 p2(p.X, y, p.Z);
			/*m_dout<<DTIME<<"lighting neighbors of node ("
					<<p2.X<<","<<p2.Y<<","<<p2.Z<<")"
					<<std::endl;*/
			lightNeighbors(p2, modified_blocks);
		}
	}
	else
	{
		// Set the lighting of this node to 0
		try{
			MapNode n = getNode(p);
			n.light = 0;
			setNode(p, n);
		}
		catch(InvalidPositionException &e)
		{
			throw;
		}
	}

	// Get the brightest neighbour node and propagate light from it
	v3s16 n2p = getBrightestNeighbour(p);
	try{
		MapNode n2 = getNode(n2p);
		lightNeighbors(n2p, modified_blocks);
	}
	catch(InvalidPositionException &e)
	{
	}

	core::map<v3s16, MapBlock * >::Iterator i = modified_blocks.getIterator();
	for(; i.atEnd() == false; i++)
	{
		i.getNode()->getValue()->updateMesh();
	}
}

core::aabbox3d<s16> Map::getDisplayedBlockArea()
{
	camera_mutex.Lock();
	core::aabbox3d<s16> box_nodes(floatToInt(camera_position));
	camera_mutex.Unlock();
	
	g_range_mutex.Lock();
	s16 d = g_viewing_range_nodes;
	g_range_mutex.Unlock();
	box_nodes.MinEdge -= v3s16(d,d,d);
	box_nodes.MaxEdge += v3s16(d,d,d);

	return core::aabbox3d<s16>(
			getNodeBlockPos(box_nodes.MinEdge),
			getNodeBlockPos(box_nodes.MaxEdge));
}

void Map::renderMap(video::IVideoDriver* driver,
	video::SMaterial *materials)
{
	//m_dout<<DTIME<<"Rendering map..."<<std::endl;
	
	/*
		Collect all blocks that are in the view range

		Should not optimize more here as we want to auto-update
		all changed nodes in viewing range at the next step.
	*/

	s16 viewing_range_nodes;
	bool viewing_range_all;
	{
		JMutexAutoLock lock(g_range_mutex);
		viewing_range_nodes = g_viewing_range_nodes;
		viewing_range_all = g_viewing_range_all;
	}

	/*
		Get all blocks and draw all visible ones
	*/

	u32 facecount = 0;
	
	core::map<v2s16, MapSector*>::Iterator si;

	si = m_sectors.getIterator();
	for(; si.atEnd() == false; si++)
	{
		MapSector *sector = si.getNode()->getValue();
		core::list< MapBlock * > sectorblocks = sector->getBlocks();
		core::list< MapBlock * >::Iterator i;
		for(i=sectorblocks.begin(); i!=sectorblocks.end(); i++)
		{
			MapBlock *block = *i;

			/*
				Compare block position to camera position, skip
				if not seen on display
			*/
			
			v3s16 blockpos_nodes = block->getPosRelative();

			// Block center position
			v3f blockpos(
					((float)blockpos_nodes.X + MAP_BLOCKSIZE/2) * BS,
					((float)blockpos_nodes.Y + MAP_BLOCKSIZE/2) * BS,
					((float)blockpos_nodes.Z + MAP_BLOCKSIZE/2) * BS
			);

			camera_mutex.Lock();
			// Block position relative to camera
			v3f blockpos_relative = blockpos - camera_position;

			// Distance in camera direction (+=front, -=back)
			f32 dforward = blockpos_relative.dotProduct(camera_direction);

			camera_mutex.Unlock();

			// Total distance
			f32 d = blockpos_relative.getLength();
			
			if(viewing_range_all == false)
			{
				// If block is far away, don't draw it
				if(d > viewing_range_nodes * BS)
					continue;
			}
			
			// Maximum radius of a block
			f32 block_max_radius = 0.5*1.44*1.44*MAP_BLOCKSIZE*BS;
			
			// If block is (nearly) touching the camera, don't
			// bother checking further
			if(d > block_max_radius * 1.5)
			{
				// Cosine of the angle between the camera direction
				// and the block direction (camera_direction is an unit vector)
				f32 cosangle = dforward / d;
				
				// Compensate for the size of the block
				// (as the block has to be shown even if it's a bit off FOV)
				// This is an estimate.
				cosangle += block_max_radius / dforward;

				// If block is not in the field of view, skip it
				if(cosangle < cos(FOV_ANGLE/2))
					continue;
			}
			
			/*
				Draw the faces of the block
			*/
			
			{
				JMutexAutoLock lock(block->mesh_mutex);

				// Cancel if block has no mesh
				if(block->mesh == NULL)
					continue;

				u32 c = block->mesh->getMeshBufferCount();

				for(u32 i=0; i<c; i++)
				{
					scene::IMeshBuffer *buf = block->mesh->getMeshBuffer(i);
					driver->setMaterial(buf->getMaterial());
					driver->drawMeshBuffer(buf);
				}
			}
		}
	}

	static s32 oldfacecount = 0;
	// Cast needed for msvc
	if(abs((long)(facecount - oldfacecount)) > 1000){
		m_dout<<DTIME<<"Map: Rendered "
				<<facecount<<" faces"<<std::endl;
		oldfacecount = facecount;
	}

#if 0	
	core::list< MapBlock * > blocks_displayed;
	core::list< MapBlock * >::Iterator bi;
	
	bool viewing_range_all;
	{
		JMutexAutoLock lock(g_range_mutex);
		viewing_range_all = g_viewing_range_all;
	}
	
	if(viewing_range_all)
	{
		/*
			Get all blocks
		*/
		core::map<v2s16, MapSector*>::Iterator si;

		si = m_sectors.getIterator();
		for(; si.atEnd() == false; si++)
		{
			MapSector *sector = si.getNode()->getValue();
			core::list< MapBlock * > sectorblocks = sector->getBlocks();
			core::list< MapBlock * >::Iterator i;
			for(i=sectorblocks.begin(); i!=sectorblocks.end(); i++){
				blocks_displayed.push_back(*i);
			}
		}
	}
	else{
		/*
			Get visible blocks
		*/
		core::aabbox3d<s16> box_blocks = getDisplayedBlockArea();

		for(s16 y=box_blocks.MaxEdge.Y; y>=box_blocks.MinEdge.Y; y--){
			for(s16 z=box_blocks.MinEdge.Z; z<=box_blocks.MaxEdge.Z; z++){
				for(s16 x=box_blocks.MinEdge.X; x<=box_blocks.MaxEdge.X; x++)
				{
					/*
						Add block to list
					*/
					try{
						MapBlock * blockref = getBlockNoCreate(v3s16(x,y,z));
						blocks_displayed.push_back(blockref);
					}
					catch(InvalidPositionException &e)
					{
					}
				}
			}
		}
	}

	/*u8 material_in_use;
	material_in_use = MATERIAL_AIR;*/

	if(blocks_displayed.getSize() == 0)
		return;
	
	/*
		Draw all displayed faces
	*/

	u32 facecount = 0;
	
	for(bi = blocks_displayed.begin(); bi != blocks_displayed.end(); bi++){
		MapBlock *block = *bi;
		
		/*
			Compare block position to camera position, skip
			if not seen on display
		*/
		
		v3s16 blockpos_nodes = block->getPosRelative();

		// Block center position
		v3f blockpos(
				((float)blockpos_nodes.X + MAP_BLOCKSIZE/2) * BS,
				((float)blockpos_nodes.Y + MAP_BLOCKSIZE/2) * BS,
				((float)blockpos_nodes.Z + MAP_BLOCKSIZE/2) * BS
		);

		camera_mutex.Lock();
		// Block position relative to camera
		v3f blockpos_relative = blockpos - camera_position;

		// Distance in camera direction (+=front, -=back)
		f32 dforward = blockpos_relative.dotProduct(camera_direction);

		camera_mutex.Unlock();

		// Total distance
		f32 d = blockpos_relative.getLength();
		
		// Maximum radius of a block
		f32 block_max_radius = 0.5*1.44*1.44*MAP_BLOCKSIZE*BS;
		
		// If block is (nearly) touching the camera, don't
		// bother checking further
		if(d > block_max_radius * 1.5)
		{
			// Cosine of the angle between the camera direction
			// and the block direction (camera_direction is an unit vector)
			f32 cosangle = dforward / d;
			
			// Compensate for the size of the block
			// (as the block has to be shown even if it's a bit off FOV)
			// This is an estimate.
			cosangle += block_max_radius / dforward;

			// If block is not in the field of view, skip it
			if(cosangle < cos(FOV_ANGLE/2))
				continue;
		}
		
		/*
			Draw the faces of the block
		*/
		
		{
			JMutexAutoLock lock(block->mesh_mutex);

			// Cancel if block has no mesh
			if(block->mesh == NULL)
				continue;

			u32 c = block->mesh->getMeshBufferCount();

			for(u32 i=0; i<c; i++)
			{
				scene::IMeshBuffer *buf = block->mesh->getMeshBuffer(i);
				driver->setMaterial(buf->getMaterial());
				driver->drawMeshBuffer(buf);
			}
		}
	}
	
	static s32 oldfacecount = 0;
	// Cast needed for msvc
	if(abs((long)(facecount - oldfacecount)) > 1000){
		m_dout<<DTIME<<"Map: Rendered "
				<<facecount<<" faces"<<std::endl;
		oldfacecount = facecount;
	}
	
	
	//m_dout<<DTIME<<"done"<<std::endl;
#endif
}

void getFacePositions(core::list<v3s16> &list, u16 d)
{
	if(d == 0)
	{
		list.push_back(v3s16(0,0,0));
		return;
	}

	// Take the top and bottom face without any borders
	// -d+1<x<d-1, y=+-d, -d+1<z<d-1
	for(s16 x=-d+1; x<=d-1; x++)
	for(s16 z=-d+1; z<=d-1; z++)
	{
		list.push_back(v3s16(x,d,z));
		list.push_back(v3s16(x,-d,z));
	}
	
	// Take the back and front face
	// -d<x<d, -d<y<d, z=+-d
	for(s16 x=-d; x<=d; x++)
	for(s16 y=-d; y<=d; y++)
	{
		list.push_back(v3s16(x,y,d));
		list.push_back(v3s16(x,y,-d));
	}

	// Take the left and right face without the already added blocks
	// x=+-d, -d<y<d, -d+1<z<d-1
	for(s16 y=-d; y<=d; y++)
	for(s16 z=-d+1; z<=d-1; z++)
	{
		list.push_back(v3s16(d,y,z));
		list.push_back(v3s16(-d,y,z));
	}

}

/*
	Returns true if updated something
*/
bool Map::updateChangedVisibleArea()
{
	/*
		Update the lighting and the faces of changed blocks that are
		to be displayed.
	*/

	camera_mutex.Lock();
	v3s16 center_nodepos = floatToInt(camera_position);
	camera_mutex.Unlock();

	v3s16 center = getNodeBlockPos(center_nodepos);

	g_range_mutex.Lock();
	s16 d_max_gen = g_generating_range_nodes / MAP_BLOCKSIZE;
	s16 d_max_view = g_viewing_range_nodes / MAP_BLOCKSIZE;
	g_range_mutex.Unlock();
	
	core::list< MapBlock * > blocks_changed;
	
	// Go from the closest to the farthest
	// Take update_max blocks at maximum
	u16 updated_max = 10;
	u16 updated_count = 0;
	// Allow queued_max calls to getBlock returning AsyncQueuedException
	u16 queued_max = 5;
	u16 queued_count = 0;
	for(s16 d=0; d <= core::max_(d_max_gen, d_max_view); d++)
	{
		core::list<v3s16> list;
		getFacePositions(list, d);

		core::list<v3s16>::Iterator li;
		for(li=list.begin(); li!=list.end(); li++)
		{
			v3s16 pos = *li + center;
			try
			{
				bool generate = true;
				if(d > d_max_gen)
					generate = false;
				if(queued_count > queued_max)
					generate = false;

				MapBlock * block = getBlock(pos, generate);
				if(block->getChangedFlag())
				{
					/*// Print initial message when first is found
					if(blocks_changed.empty() == true){
						m_dout<<DTIME<<"Updating changed MapBlocks at ";
					}*/

					v3s16 p = block->getPosRelative();
					/*m_dout<<"("<<p.X<<","<<p.Y<<","<<p.Z<<") ";
					m_dout.flush();*/

					blocks_changed.push_back(block);
					block->resetChangedFlag();

					updated_count++;
					if(updated_count >= updated_max)
						goto done;
				}
			}
			catch(InvalidPositionException &e)
			{
			}
			catch(ClientNotReadyException &e)
			{
				// Client is not connected and initialized, cancel
				goto done;
			}
			catch(AsyncQueuedException &e)
			{
				queued_count++;
			}
		}
	}
done:
	
	// Quit if nothing has changed
	if(blocks_changed.empty()){
		return false;
	}

	//m_dout<<std::endl;

	m_dout<<DTIME<<"Map::updateChangedVisibleArea(): "
			"Updating "<<blocks_changed.getSize()
			<<" blocks"<<std::endl;

	core::map<v3s16, MapBlock*> modified_blocks;

	//m_dout<<DTIME<<"Updating lighting"<<std::endl;

	updateLighting(blocks_changed, modified_blocks);

	//m_dout<<DTIME<<"Updating face cache of changed blocks"<<std::endl;

	core::map<v3s16, MapBlock * >::Iterator i = modified_blocks.getIterator();
	for(; i.atEnd() == false; i++)
	{
		MapBlock *block = i.getNode()->getValue();
		block->updateMesh();
		
		/*m_dout<<"X";
		m_dout.flush();*/
	}

	//m_dout<<"Map::uCVA Done ("<<getTimestamp()<<")"<<std::endl;
	//m_dout<<DTIME<<"Map::updateChangedVisibleArea(): Done."<<std::endl;

	return true;
}

/*
	Updates usage timers
*/
void Map::timerUpdate(float dtime)
{
	core::map<v2s16, MapSector*>::Iterator si;

	si = m_sectors.getIterator();
	for(; si.atEnd() == false; si++)
	{
		MapSector *sector = si.getNode()->getValue();
		sector->usage_timer += dtime;
	}
}

void Map::deleteSectors(core::list<v2s16> &list)
{
	core::list<v2s16>::Iterator j;
	for(j=list.begin(); j!=list.end(); j++)
	{
		MapSector *sector = m_sectors[*j];
		/*
			If sector is in sector cache, remove it from there
		*/
		if(m_sector_cache == sector)
		{
			m_sector_cache = NULL;
		}
		/*
			Remove from map and delete
		*/
		m_sectors.remove(*j);
		delete sector;
	}
}

u32 Map::deleteUnusedSectors(float timeout)
{
	core::list<v2s16> sector_deletion_queue;
	core::map<v2s16, MapSector*>::Iterator i = m_sectors.getIterator();
	for(; i.atEnd() == false; i++)
	{
		MapSector *sector = i.getNode()->getValue();
		/*
			Delete sector from memory if it hasn't been used in a long time
		*/
		if(sector->usage_timer > timeout)
		{
			sector_deletion_queue.push_back(i.getNode()->getKey());
		}
	}
	deleteSectors(sector_deletion_queue);
	return sector_deletion_queue.getSize();
}

/*
	ServerMap
*/

ServerMap::ServerMap(std::string savedir):
	Map(dout_server),
	m_heightmap(NULL)
	//m_heightmap(new UnlimitedHeightmap(16, 42.0, 0.6, 0.0))
	//m_heightmap(16, 42.0, 0.6, 0.0)
	//m_heightmap(32, 66.0, 0.6, 0.0)
{
	m_savedir = savedir;
	m_map_saving_enabled = false;
	
	try
	{
		if(fs::PathExists(m_savedir))
		{
			// Load master heightmap
			loadMasterHeightmap();
			
			// Load sector (0,0) and throw and exception on fail
			if(loadSectorFull(v2s16(0,0)) == false)
				throw LoadError("Failed to load sector (0,0)");

			std::cout<<DTIME<<"Server: Successfully loaded master "
					"heightmap and sector (0,0) from "<<savedir<<
					", assuming valid save directory."
					<<std::endl;

			m_map_saving_enabled = true;
			return;
		}
		else{
			m_map_saving_enabled = true;
		}
	}
	catch(std::exception &e)
	{
		std::cout<<DTIME<<"Server: Failed to load map from "<<savedir
				<<", exception: "<<e.what()<<std::endl;
		std::cout<<DTIME<<"Please remove the map or fix it."<<std::endl;
		std::cout<<DTIME<<"WARNING: Map saving will be disabled."<<std::endl;
	}

	std::cout<<DTIME<<"Initializing new map."<<std::endl;
	m_heightmap = new UnlimitedHeightmap(16, 42.0, 0.6, 0.0);
	// Initially write whole map
	save(false);
}

ServerMap::~ServerMap()
{
	try
	{
		if(m_map_saving_enabled)
		{
			//save(false);
			// Save only changed parts
			save(true);
			std::cout<<DTIME<<"Server: saved map to "<<m_savedir<<std::endl;
		}
		else
		{
			std::cout<<DTIME<<"Server: map not saved"<<std::endl;
		}
	}
	catch(std::exception &e)
	{
		std::cout<<DTIME<<"Server: Failed to save map to "<<m_savedir
				<<", exception: "<<e.what()<<std::endl;
	}
	
	if(m_heightmap != NULL)
		delete m_heightmap;
}

MapSector * ServerMap::getSector(v2s16 p2d, bool generate)
{
	// Check that it doesn't exist already
	try{
		return getSectorNoGenerate(p2d);
	}
	catch(InvalidPositionException &e)
	{
		/*
			Try to load the sector from disk.
			We count it as generating a sector.
		*/
		if(generate == true)
		{
			if(loadSectorFull(p2d) == true)
			{
				return getSectorNoGenerate(p2d);
			}
		}

		/*
			If there is no heightmap or we don't want to generate
			anything, re-throw the exception.
		*/
		if(generate == false || m_heightmap == NULL)
			throw e;
	}

	/*m_dout<<DTIME<<"Map::getSector(("<<p2d.X<<","<<p2d.Y
			<<")): generating"<<std::endl;*/

	//m_dout<<DTIME<<"Generating heightmap..."<<std::endl;

	f32 corners[4] = {
		m_heightmap->getGroundHeight(p2d+v2s16(0,0)),
		m_heightmap->getGroundHeight(p2d+v2s16(1,0)),
		m_heightmap->getGroundHeight(p2d+v2s16(1,1)),
		m_heightmap->getGroundHeight(p2d+v2s16(0,1)),
	};
	
	HeightmapBlockGenerator *gen;
	gen = new HeightmapBlockGenerator(p2d, &m_hwrapper);
	/*DummyHeightmap dummyheightmap;
	gen = new HeightmapBlockGenerator(p2d, &dummyheightmap);*/
	gen->m_heightmap->generateContinued(2.0, 0.2, corners);
	//sector->getHeightmap()->generateContinued(0.0, 0.0, corners);
	
	MapSector *sector = new MapSector(this, p2d, gen);

	sector->setHeightmap(gen->m_heightmap);

	m_sectors.insert(p2d, sector);

	/*m_dout<<DTIME<<"Map::getSector(("<<p2d.X<<","<<p2d.Y
			<<")): generated. Now returning."<<std::endl;*/
	
	return sector;
}

void ServerMap::createDir(std::string path)
{
	if(fs::CreateDir(path) == false)
	{
		m_dout<<DTIME<<"ServerMap: Failed to create directory "
				<<"\""<<path<<"\""<<std::endl;
		throw BaseException("ServerMap failed to create directory");
	}
}

std::string ServerMap::getSectorSubDir(v2s16 pos)
{
	char cc[9];
	snprintf(cc, 9, "%.4x%.4x",
			(unsigned int)pos.X&0xffff,
			(unsigned int)pos.Y&0xffff);

	return std::string(cc);
}

std::string ServerMap::getSectorDir(v2s16 pos)
{
	return m_savedir + "/sectors/" + getSectorSubDir(pos);
}

v2s16 ServerMap::getSectorPos(std::string dirname)
{
	if(dirname.size() != 8)
		throw InvalidFilenameException("Invalid sector directory name");
	unsigned int x, y;
	int r = sscanf(dirname.c_str(), "%4x%4x", &x, &y);
	if(r != 2)
		throw InvalidFilenameException("Invalid sector directory name");
	v2s16 pos((s16)x, (s16)y);
	return pos;
}

v3s16 ServerMap::getBlockPos(std::string sectordir, std::string blockfile)
{
	v2s16 p2d = getSectorPos(sectordir);

	if(blockfile.size() != 4){
		throw InvalidFilenameException("Invalid block filename");
	}
	unsigned int y;
	int r = sscanf(blockfile.c_str(), "%4x", &y);
	if(r != 1)
		throw InvalidFilenameException("Invalid block filename");
	return v3s16(p2d.X, y, p2d.Y);
}

// Debug helpers
#define ENABLE_SECTOR_SAVING 1
#define ENABLE_SECTOR_LOADING 1
#define ENABLE_BLOCK_SAVING 1
#define ENABLE_BLOCK_LOADING 1

void ServerMap::save(bool only_changed)
{
	if(m_map_saving_enabled == false)
	{
		std::cout<<DTIME<<"WARNING: Not saving map, saving disabled."<<std::endl;
		return;
	}
	
	if(only_changed == false)
		std::cout<<DTIME<<"ServerMap: Saving whole map, this can take time."
				<<std::endl;
	
	saveMasterHeightmap();
	
	u32 sector_meta_count = 0;
	u32 block_count = 0;
	core::map<v2s16, MapSector*>::Iterator i = m_sectors.getIterator();
	for(; i.atEnd() == false; i++)
	{
		MapSector *sector = i.getNode()->getValue();
		
		if(ENABLE_SECTOR_SAVING)
		{
			if(sector->differs_from_disk || only_changed == false)
			{
				saveSectorMeta(sector);
				sector_meta_count++;
			}
		}
		if(ENABLE_BLOCK_SAVING)
		{
			core::list<MapBlock*> blocks = sector->getBlocks();
			core::list<MapBlock*>::Iterator j;
			for(j=blocks.begin(); j!=blocks.end(); j++)
			{
				MapBlock *block = *j;
				if(block->getChangedFlag() || only_changed == false)
				{
					saveBlock(block);
					block_count++;
				}
			}
		}
	}
	
	u32 deleted_count = 0;
	deleted_count = deleteUnusedSectors
			(SERVERMAP_DELETE_UNUSED_SECTORS_TIMEOUT);
	
	/*
		Only print if something happened or saved whole map
	*/
	if(only_changed == false || sector_meta_count != 0
			|| block_count != 0 || deleted_count != 0)
	{
		std::cout<<DTIME<<"ServerMap: Written: "
				<<sector_meta_count<<" sector metadata files, "
				<<block_count<<" block files, "
				<<deleted_count<<" sectors unloaded from memory."
				<<std::endl;
	}
}

void ServerMap::loadAll()
{
	std::cout<<DTIME<<"ServerMap: Loading map..."<<std::endl;

	loadMasterHeightmap();

	std::vector<fs::DirListNode> list = fs::GetDirListing(m_savedir+"/sectors/");

	std::cout<<DTIME<<"There are "<<list.size()<<" sectors."<<std::endl;
	
	s32 counter = 0;
	s32 printed_counter = -100000;
	s32 count = list.size();

	std::vector<fs::DirListNode>::iterator i;
	for(i=list.begin(); i!=list.end(); i++)
	{
		if(counter > printed_counter + 10)
		{
			std::cout<<DTIME<<counter<<"/"<<count<<std::endl;
			printed_counter = counter;
		}
		counter++;

		if(ENABLE_SECTOR_LOADING)
		{
			// We want directories
			if(i->dir == false)
				continue;
			try{
				loadSectorMeta(i->name);
			}
			catch(InvalidFilenameException &e)
			{
				// This catches unknown crap in directory
			}
		}
		if(ENABLE_BLOCK_LOADING)
		{
			std::vector<fs::DirListNode> list2 = fs::GetDirListing
					(m_savedir+"/sectors/"+i->name);
			std::vector<fs::DirListNode>::iterator i2;
			for(i2=list2.begin(); i2!=list2.end(); i2++)
			{
				// We want files
				if(i2->dir)
					continue;
				try{
					loadBlock(i->name, i2->name);
				}
				catch(InvalidFilenameException &e)
				{
					// This catches unknown crap in directory
				}
			}
		}
	}
	std::cout<<DTIME<<"ServerMap: Map loaded."<<std::endl;
}

void ServerMap::saveMasterHeightmap()
{
	createDir(m_savedir);
	
	std::string fullpath = m_savedir + "/master_heightmap";
	std::ofstream o(fullpath.c_str(), std::ios_base::binary);
	if(o.good() == false)
		throw FileNotGoodException("Cannot open master heightmap");
	
	// Format used for writing
	u8 version = SER_FMT_VER_HIGHEST;

	SharedBuffer<u8> hmdata = m_heightmap->serialize(version);
	/*
		[0] u8 serialization version
		[1] X master heightmap
	*/
	u32 fullsize = 1 + hmdata.getSize();
	SharedBuffer<u8> data(fullsize);

	data[0] = version;
	memcpy(&data[1], *hmdata, hmdata.getSize());

	o.write((const char*)*data, fullsize);
}

void ServerMap::loadMasterHeightmap()
{
	std::string fullpath = m_savedir + "/master_heightmap";
	std::ifstream is(fullpath.c_str(), std::ios_base::binary);
	if(is.good() == false)
		throw FileNotGoodException("Cannot open master heightmap");
	
	u8 version;
	is.read((char*)&version, 1);

	if(m_heightmap != NULL)
		delete m_heightmap;
	m_heightmap = UnlimitedHeightmap::deSerialize(is, version);
}

void ServerMap::saveSectorMeta(MapSector *sector)
{
	// Format used for writing
	u8 version = SER_FMT_VER_HIGHEST;
	// Get destination
	v2s16 pos = sector->getPos();
	createDir(m_savedir);
	createDir(m_savedir+"/sectors");
	std::string dir = getSectorDir(pos);
	createDir(dir);
	/*
		Save heightmap
	*/
	{
		std::string fullpath = dir + "/heightmap";
		std::ofstream o(fullpath.c_str(), std::ios_base::binary);
		if(o.good() == false)
			throw FileNotGoodException("Cannot open master heightmap");

		/*
			[0] u8 serialization version
			[1] heightmap data
		*/
		o.write((char*)&version, 1);
		
		u32 heightmap_size =
				FixedHeightmap::serializedLength(version, MAP_BLOCKSIZE);
		
		SharedBuffer<u8> data(heightmap_size);
		sector->getHeightmap()->serialize(*data, version);

		o.write((const char*)*data, heightmap_size);
		
		sector->differs_from_disk = false;
	}
}

void ServerMap::loadSectorMeta(std::string dirname)
{
	// Get destination
	v2s16 p2d = getSectorPos(dirname);
	std::string dir = m_savedir + "/sectors/" + dirname;
	/*
		Load heightmap
	*/
	{
		std::string fullpath = dir + "/heightmap";
		std::ifstream is(fullpath.c_str(), std::ios_base::binary);
		if(is.good() == false)
			throw FileNotGoodException("Cannot open sector heightmap");
		
		/*
			[0] u8 serialization version
			[1] heightmap data
		*/
		u8 version = SER_FMT_VER_INVALID;
		is.read((char*)&version, 1);
		
		u32 heightmap_size =
				FixedHeightmap::serializedLength(version, MAP_BLOCKSIZE);
		
		SharedBuffer<u8> data(heightmap_size);
		is.read((char*)*data, heightmap_size);

		/*u32 usedsize;
		FixedHeightmap *hm = FixedHeightmap::deSerialize
				(*data, heightmap_size, usedsize, &m_hwrapper, version);*/
		FixedHeightmap *hm = new FixedHeightmap(&m_hwrapper, p2d, MAP_BLOCKSIZE);
		hm->deSerialize(*data, version);

		HeightmapBlockGenerator *gen;
		gen = new HeightmapBlockGenerator(hm);

		try{
			MapSector *sector = getSectorNoGenerate(p2d);
			sector->changeBlockGen(gen);
			sector->setHeightmap(gen->m_heightmap);
			sector->differs_from_disk = false;
		}
		catch(InvalidPositionException &e)
		{
			MapSector *sector = new MapSector(this, p2d, gen);
			sector->setHeightmap(gen->m_heightmap);
			m_sectors.insert(p2d, sector);
			sector->differs_from_disk = false;
		}
	}
}

bool ServerMap::loadSectorFull(v2s16 p2d)
{
	std::string sectorsubdir = getSectorSubDir(p2d);

	if(ENABLE_SECTOR_LOADING)
	{
		try{
			loadSectorMeta(sectorsubdir);
		}
		catch(InvalidFilenameException &e)
		{
			return false;
		}
		catch(FileNotGoodException &e)
		{
			return false;
		}
		catch(std::exception &e)
		{
			return false;
		}
	}
	if(ENABLE_BLOCK_LOADING)
	{
		std::vector<fs::DirListNode> list2 = fs::GetDirListing
				(m_savedir+"/sectors/"+sectorsubdir);
		std::vector<fs::DirListNode>::iterator i2;
		for(i2=list2.begin(); i2!=list2.end(); i2++)
		{
			// We want files
			if(i2->dir)
				continue;
			try{
				loadBlock(sectorsubdir, i2->name);
			}
			catch(InvalidFilenameException &e)
			{
				// This catches unknown crap in directory
			}
		}
	}
	return true;
}

#if 0
bool ServerMap::deFlushSector(v2s16 p2d)
{
	// See if it already exists in memory
	try{
		MapSector *sector = getSectorNoGenerate(p2d);
		return true;
	}
	catch(InvalidPositionException &e)
	{
		/*
			Try to load the sector from disk.
		*/
		if(loadSectorFull(p2d) == true)
		{
			return true;
		}
	}
	return false;
}
#endif

void ServerMap::saveBlock(MapBlock *block)
{
	// Format used for writing
	u8 version = SER_FMT_VER_HIGHEST;
	// Get destination
	v3s16 p3d = block->getPos();
	v2s16 p2d(p3d.X, p3d.Z);
	createDir(m_savedir);
	createDir(m_savedir+"/sectors");
	std::string dir = getSectorDir(p2d);
	createDir(dir);
	
	// Block file is map/sectors/xxxxxxxx/xxxx
	char cc[5];
	snprintf(cc, 5, "%.4x", (unsigned int)p3d.Y&0xffff);
	std::string fullpath = dir + "/" + cc;
	std::ofstream o(fullpath.c_str(), std::ios_base::binary);
	if(o.good() == false)
		throw FileNotGoodException("Cannot open block data");

	/*
		[0] u8 serialization version
		[1] data
	*/
	o.write((char*)&version, 1);
	
	u32 block_size =
			MapBlock::serializedLength(version);
	
	SharedBuffer<u8> data(block_size);
	block->serialize(*data, version);

	o.write((const char*)*data, block_size);
	
	// We just wrote it to the disk
	block->resetChangedFlag();
}

void ServerMap::loadBlock(std::string sectordir, std::string blockfile)
{
	// Block file is map/sectors/xxxxxxxx/xxxx
	std::string fullpath = m_savedir+"/sectors/"+sectordir+"/"+blockfile;
	std::ifstream is(fullpath.c_str(), std::ios_base::binary);
	if(is.good() == false)
		throw FileNotGoodException("Cannot open block file");

	v3s16 p3d = getBlockPos(sectordir, blockfile);
	v2s16 p2d(p3d.X, p3d.Z);
	
	u8 version = SER_FMT_VER_INVALID;
	is.read((char*)&version, 1);

	u32 block_size = MapBlock::serializedLength(version);
	SharedBuffer<u8> data(block_size);
	is.read((char*)*data, block_size);

	// This will always return a sector because we're the server
	MapSector *sector = getSector(p2d, true);
	MapBlock *block = NULL;
	try{
		block = sector->getBlockNoCreate(p3d.Y);
		block->deSerialize(*data, version);
	}
	catch(InvalidPositionException &e)
	{
		block = sector->createBlankBlockNoInsert(p3d.Y);
		block->deSerialize(*data, version);
		sector->insertBlock(block);
	}
	// We just loaded it from the disk
	block->resetChangedFlag();
}

/*
	ClientMap
*/

ClientMap::ClientMap(
		Client *client,
		video::SMaterial *materials,
		scene::ISceneNode* parent,
		scene::ISceneManager* mgr,
		s32 id
):
	Map(dout_client),
	scene::ISceneNode(parent, mgr, id),
	m_client(client),
	m_materials(materials)
{
	/*m_box = core::aabbox3d<f32>(0,0,0,
			map->getW()*BS, map->getH()*BS, map->getD()*BS);*/
	/*m_box = core::aabbox3d<f32>(0,0,0,
			map->getSizeNodes().X * BS,
			map->getSizeNodes().Y * BS,
			map->getSizeNodes().Z * BS);*/
	m_box = core::aabbox3d<f32>(-BS*1000000,-BS*1000000,-BS*1000000,
			BS*1000000,BS*1000000,BS*1000000);
}

MapSector * ClientMap::getSector(v2s16 p2d, bool generate)
{
	// Check that it doesn't exist already
	try{
		MapSector *sector = getSectorNoGenerate(p2d);
		return sector;
	}
	catch(InvalidPositionException &e)
	{
		if(generate == false)
			throw e;
	}

	/*
		Create a ClientBlockGenerator to fill the MapSector's
		blocks from data fetched by the client from the server.
	*/
	ClientBlockGenerator *gen;
	gen = new ClientBlockGenerator(m_client);
	
	MapSector *sector = new MapSector(this, p2d, gen);

	m_sectors.insert(p2d, sector);

	return sector;
}


