#include "client.h"
#include "utility.h"
#include <iostream>
#include "clientserver.h"
#include "jmutexautolock.h"
#include "main.h"
#include <sstream>

#ifdef _WIN32
	#include <windows.h>
	#define sleep_ms(x) Sleep(x)
#else
	#include <unistd.h>
	#define sleep_ms(x) usleep(x*1000)
#endif

/*
	FIXME: This thread can access the environment at any time
*/

void * ClientUpdateThread::Thread()
{
	ThreadStarted();

	while(getRun())
	{
		m_client->AsyncProcessData();
		sleep_ms(200);
	}

	return NULL;
}

Client::Client(IrrlichtDevice *device, video::SMaterial *materials):
	m_thread(this),
	m_env(new ClientMap(this, materials,
			device->getSceneManager()->getRootSceneNode(),
			device->getSceneManager(), 666),
			dout_client),
	m_con(PROTOCOL_ID, 512, 20.0),
	m_device(device)
	//m_server_ser_ver(SER_FMT_VER_INVALID)
{
	m_fetchblock_mutex.Init();
	m_incoming_queue_mutex.Init();
	m_env_mutex.Init();
	m_con_mutex.Init();

	m_thread.Start();
	
	{
		JMutexAutoLock envlock(m_env_mutex);
		//m_env.getMap().StartUpdater();

		Player *player = new Player(true,
				device->getSceneManager()->getRootSceneNode(),
				device,
				0);
		//f32 y = BS*2 + m_env.getMap().getGroundHeight(v2s16(0,0));
		f32 y = BS*2 + BS*20;
		player->setPosition(v3f(0, y, 0));
		//player->setPosition(v3f(-33*BS, -1*BS, 43*BS)); // DEBUG
		m_env.addPlayer(player);
	}
}

Client::~Client()
{
	m_thread.setRun(false);
	while(m_thread.IsRunning())
		sleep_ms(100);
}

void Client::connect(Address address)
{
	JMutexAutoLock lock(m_con_mutex);
	m_con.setTimeoutMs(0);
	m_con.Connect(address);
}

bool Client::connectedAndInitialized()
{
	JMutexAutoLock lock(m_con_mutex);

	if(m_con.Connected() == false)
		return false;
	
	// Caller handles con::PeerNotFoundException
	if(m_con.GetPeer(PEER_ID_SERVER)->serialization_version
			== SER_FMT_VER_INVALID)
		return false;
	
	return true;
}

void Client::step(float dtime)
{
	ReceiveAll();
	
	{
		JMutexAutoLock lock(m_con_mutex);
		m_con.RunTimeouts(dtime);
	}

	bool connected = connectedAndInitialized();

	if(connected == false)
	{
		static float counter = -0.001;
		counter -= dtime;
		if(counter <= 0.0)
		{
			counter = 2.0;

			JMutexAutoLock envlock(m_env_mutex);
			
			Player *myplayer = m_env.getLocalPlayer();
			assert(myplayer != NULL);
	
			// Send TOSERVER_INIT
			// [0] u16 TOSERVER_INIT
			// [2] u8 SER_FMT_VER_HIGHEST
			// [3] u8[20] player_name
			SharedBuffer<u8> data(2+1+20);
			writeU16(&data[0], TOSERVER_INIT);
			writeU8(&data[2], SER_FMT_VER_HIGHEST);
			memcpy(&data[3], myplayer->getName(), 20);
			// Send as unreliable
			Send(0, data, false);
		}
		return;
	}

	/*
		Do stuff if connected
	*/
	
	{
		JMutexAutoLock lock(m_env_mutex);
		m_env.step(dtime);
	}

	{
		/*
			TODO: Fetch some nearby blocks
		*/
		fetchBlocks();
	}

	{
		static float counter = 0.0;
		counter += dtime;
		if(counter >= 10)
		{
			counter = 0.0;
			JMutexAutoLock lock(m_con_mutex);
			// connectedAndInitialized() is true, peer exists.
			con::Peer *peer = m_con.GetPeer(PEER_ID_SERVER);
			std::cout<<DTIME<<"Client: avg_rtt="<<peer->avg_rtt<<std::endl;
		}
	}
	{
		// Update at reasonable intervals (0.2s)
		static float counter = 0.0;
		counter += dtime;
		if(counter >= 0.2)
		{
			counter = 0.0;
			sendPlayerPos();
		}
	}
	/*{
		// Update at some interval in case of connection drops
		// or changes in info
		static float counter = 0.0;
		counter += dtime;
		if(counter >= 10.0)
		{
			counter = 0.0;
			sendPlayerInfo();
		}
	}*/

	/*
		Clear old entries from fetchblock history
	*/
	{
		JMutexAutoLock lock(m_fetchblock_mutex);
		
		core::list<v3s16> remove_queue;
		core::map<v3s16, float>::Iterator i;
		i = m_fetchblock_history.getIterator();
		for(; i.atEnd() == false; i++)
		{
			float value = i.getNode()->getValue();
			value += dtime;
			i.getNode()->setValue(value);
			if(value >= 60.0)
				remove_queue.push_back(i.getNode()->getKey());
		}
		core::list<v3s16>::Iterator j;
		j = remove_queue.begin();
		for(; j != remove_queue.end(); j++)
		{
			m_fetchblock_history.remove(*j);
		}
	}
}

void Client::ReceiveAll()
{
	for(;;){
		try{
			Receive();
		}
		catch(con::NoIncomingDataException &e)
		{
			break;
		}
		catch(con::InvalidIncomingDataException &e)
		{
			dout_client<<DTIME<<"Client::ReceiveAll(): "
					"InvalidIncomingDataException: what()="
					<<e.what()<<std::endl;
		}
	}
}

void Client::Receive()
{
	u32 data_maxsize = 10000;
	Buffer<u8> data(data_maxsize);
	u16 sender_peer_id;
	u32 datasize;
	{
		JMutexAutoLock lock(m_con_mutex);
		datasize = m_con.Receive(sender_peer_id, *data, data_maxsize);
	}
	ProcessData(*data, datasize, sender_peer_id);
}

/*
	sender_peer_id given to this shall be quaranteed to be a valid peer
*/
void Client::ProcessData(u8 *data, u32 datasize, u16 sender_peer_id)
{
	// Ignore packets that don't even fit a command
	if(datasize < 2)
		return;

	ToClientCommand command = (ToClientCommand)readU16(&data[0]);
	
	/*
		If this check is removed, be sure to change the queue
		system to know the ids
	*/
	if(sender_peer_id != PEER_ID_SERVER)
	{
		dout_client<<DTIME<<"Client::ProcessData(): Discarding data not "
				"coming from server: peer_id="<<sender_peer_id
				<<std::endl;
		return;
	}

	con::Peer *peer;
	{
		JMutexAutoLock lock(m_con_mutex);
		// All data is coming from the server
		// PeerNotFoundException is handled by caller.
		peer = m_con.GetPeer(PEER_ID_SERVER);
	}

	u8 ser_version = peer->serialization_version;

	// Execute fast commands straight away

	if(command == TOCLIENT_INIT)
	{
		if(datasize < 3)
			return;

		u8 deployed = data[2];
		
		dout_client<<DTIME<<"Client: TOCLIENT_INIT received with "
				"deployed="<<((int)deployed&0xff)<<std::endl;

		assert(deployed >= SER_FMT_VER_LOWEST);
		assert(deployed <= SER_FMT_VER_HIGHEST);
		
		peer->serialization_version = deployed;

		return;
	}
	
	if(ser_version == SER_FMT_VER_INVALID)
	{
		dout_client<<DTIME<<"WARNING: Client: Server serialization"
				" format invalid or not initialized."
				" Skipping incoming command="<<command<<std::endl;
		return;
	}
	
	if(command == TOCLIENT_REMOVENODE)
	{
		if(datasize < 8)
			return;
		v3s16 p;
		p.X = readS16(&data[2]);
		p.Y = readS16(&data[4]);
		p.Z = readS16(&data[6]);
		try
		{
			JMutexAutoLock envlock(m_env_mutex);
			TimeTaker t("removeNodeAndUpdate", m_device);
			m_env.getMap().removeNodeAndUpdate(p);
		}
		catch(InvalidPositionException &e)
		{
		}
	}
	else if(command == TOCLIENT_ADDNODE)
	{
		if(datasize < 8)
			return;

		v3s16 p;
		p.X = readS16(&data[2]);
		p.Y = readS16(&data[4]);
		p.Z = readS16(&data[6]);
		
		if(datasize < 8 + MapNode::serializedLength(ser_version))
			return;

		MapNode n;
		n.deSerialize(&data[8], ser_version);
		
		try
		{
			JMutexAutoLock envlock(m_env_mutex);
			u8 light = m_env.getMap().getNode(p).getLight();
			m_env.getMap().setNode(p, n);
			TimeTaker t("nodeAddedUpdate", m_device);
			m_env.getMap().nodeAddedUpdate(p, light);
		}
		catch(InvalidPositionException &e)
		{
		}
	}
	else if(command == TOCLIENT_PLAYERPOS)
	{
		u16 our_peer_id;
		{
			JMutexAutoLock lock(m_con_mutex);
			our_peer_id = m_con.GetPeerID();
		}
		// Cancel if we don't have a peer id
		if(our_peer_id == PEER_ID_NEW){
			dout_client<<DTIME<<"TOCLIENT_PLAYERPOS cancelled: "
					"we have no peer id"
					<<std::endl;
			return;
		}

		{ //envlock
			JMutexAutoLock envlock(m_env_mutex);

			u32 player_count = (datasize-2) / (2+12+12);
			u32 start = 2;
			for(u32 i=0; i<player_count; i++)
			{
				u16 peer_id = readU16(&data[start]);


				// Don't update the position of the local player
				if(peer_id == our_peer_id)
				{
					/*dout_client<<DTIME<<"Client: TOCLIENT_PLAYERPOS: "
							"player peer_id="<<peer_id;
					dout_client<<" is ours"<<std::endl;
					dout_client<<std::endl;*/

					start += (2+12+12);
					continue;
				}
				else
				{
					/*dout_client<<DTIME<<"Client: TOCLIENT_PLAYERPOS: "
							"player peer_id="<<peer_id;
					dout_client<<std::endl;*/
				}


				Player *player = m_env.getPlayer(peer_id);

				// Create a player if it doesn't exist
				if(player == NULL)
				{
					player = new Player(false,
							m_device->getSceneManager()->getRootSceneNode(),
							m_device,
							peer_id);
					player->peer_id = peer_id;
					m_env.addPlayer(player);
					dout_client<<DTIME<<"Client: Adding new player "
							<<peer_id<<std::endl;
				}

				//player->timeout_counter = 0.0;
				
				v3s32 ps = readV3S32(&data[start+2]);
				v3s32 ss = readV3S32(&data[start+2+12]);
				v3f position((f32)ps.X/100., (f32)ps.Y/100., (f32)ps.Z/100.);
				v3f speed((f32)ss.X/100., (f32)ss.Y/100., (f32)ss.Z/100.);
				player->setPosition(position);
				player->speed = speed;

				start += (2+12+12);
			}
		} //envlock
	}
	else if(command == TOCLIENT_PLAYERINFO)
	{
		u16 our_peer_id;
		{
			JMutexAutoLock lock(m_con_mutex);
			our_peer_id = m_con.GetPeerID();
		}
		// Cancel if we don't have a peer id
		if(our_peer_id == PEER_ID_NEW){
			dout_client<<DTIME<<"TOCLIENT_PLAYERINFO cancelled: "
					"we have no peer id"
					<<std::endl;
			return;
		}
		
		//std::cout<<DTIME<<"Client: Server reports players:"<<std::endl;

		{ //envlock
			JMutexAutoLock envlock(m_env_mutex);
			
			u32 item_size = 2+PLAYERNAME_SIZE;
			u32 player_count = (datasize-2) / item_size;
			u32 start = 2;
			// peer_ids
			core::list<u16> players_alive;
			for(u32 i=0; i<player_count; i++)
			{
				// Make sure the name ends in '\0'
				data[start+2+PLAYERNAME_SIZE-1] = 0;

				u16 peer_id = readU16(&data[start]);

				players_alive.push_back(peer_id);
				
				/*std::cout<<DTIME<<"peer_id="<<peer_id
						<<" name="<<((char*)&data[start+2])<<std::endl;*/

				// Don't update the info of the local player
				if(peer_id == our_peer_id)
				{
					start += item_size;
					continue;
				}

				Player *player = m_env.getPlayer(peer_id);

				// Create a player if it doesn't exist
				if(player == NULL)
				{
					player = new Player(false,
							m_device->getSceneManager()->getRootSceneNode(),
							m_device,
							peer_id);
					player->peer_id = peer_id;
					m_env.addPlayer(player);
					dout_client<<DTIME<<"Client: Adding new player "
							<<peer_id<<std::endl;
				}
				
				player->updateName((char*)&data[start+2]);

				start += item_size;
			}
			
			/*
				Remove those players from the environment that
				weren't listed by the server.
			*/
			//std::cout<<DTIME<<"Removing dead players"<<std::endl;
			core::list<Player*> players = m_env.getPlayers();
			core::list<Player*>::Iterator ip;
			for(ip=players.begin(); ip!=players.end(); ip++)
			{
				bool is_alive = false;
				core::list<u16>::Iterator i;
				for(i=players_alive.begin(); i!=players_alive.end(); i++)
				{
					if((*ip)->peer_id == *i)
					{
						is_alive = true;
						break;
					}
				}
				/*std::cout<<DTIME<<"peer_id="<<((*ip)->peer_id)
						<<" is_alive="<<is_alive<<std::endl;*/
				if(is_alive)
					continue;
				std::cout<<DTIME<<"Removing dead player "<<(*ip)->peer_id
						<<std::endl;
				m_env.removePlayer((*ip)->peer_id);
			}
		} //envlock
	}
	else if(command == TOCLIENT_OPT_BLOCK_NOT_FOUND)
	{
		// Ignore too small packet
		if(datasize < 8)
			return;
			
		v3s16 p;
		p.X = readS16(&data[2]);
		p.Y = readS16(&data[4]);
		p.Z = readS16(&data[6]);
		
		/*
			Handle histories
		*/
		{
			JMutexAutoLock lock(m_fetchblock_mutex);
			
			// Add to opt-not-found history
			m_opt_not_found_history.insert(p, 0.0);

			// Remove from fetch history
			if(m_fetchblock_history.find(p) != NULL)
			{
				m_fetchblock_history.remove(p);
			}
		}
	}
	// Default to queueing it (for slow commands)
	else
	{
		JMutexAutoLock lock(m_incoming_queue_mutex);
		
		IncomingPacket packet(data, datasize);
		m_incoming_queue.push_back(packet);
	}
}

/*
	Returns true if there was something in queue
*/
bool Client::AsyncProcessData()
{
	try
	{

	con::Peer *peer;
	{
		JMutexAutoLock lock(m_con_mutex);
		// All data is coming from the server
		peer = m_con.GetPeer(PEER_ID_SERVER);
	}
	
	u8 ser_version = peer->serialization_version;

getdata:
	IncomingPacket packet = getPacket();
	u8 *data = packet.m_data;
	u32 datasize = packet.m_datalen;
	
	// An empty packet means queue is empty
	if(data == NULL){
		return false;
	}
	
	if(datasize < 2)
		goto getdata;
	
	ToClientCommand command = (ToClientCommand)readU16(&data[0]);

	if(command == TOCLIENT_BLOCKDATA)
	{
		// Ignore too small packet
		if(datasize < 8)
			goto getdata;
		/*if(datasize < 8 + MapBlock::serializedLength(ser_version))
			goto getdata;*/
			
		v3s16 p;
		p.X = readS16(&data[2]);
		p.Y = readS16(&data[4]);
		p.Z = readS16(&data[6]);
		
		/*dout_client<<DTIME<<"Client: Thread: BLOCKDATA for ("
				<<p.X<<","<<p.Y<<","<<p.Z<<"): ";*/
		
		std::string datastring((char*)&data[8], datasize-8);
		std::istringstream istr(datastring, std::ios_base::binary);
		
		MapSector *sector;
		MapBlock *block;
		
		{ //envlock
			JMutexAutoLock envlock(m_env_mutex);
			
			v2s16 p2d(p.X, p.Z);
			sector = m_env.getMap().emergeSector(p2d);
			
			v2s16 sp = sector->getPos();
			if(sp != p2d)
			{
				std::cout<<"ERROR: Got sector with getPos()="
						<<"("<<sp.X<<","<<sp.Y<<"), tried to get"
						<<"("<<p2d.X<<","<<p2d.Y<<")"<<std::endl;
			}

			assert(sp == p2d);
			//assert(sector->getPos() == p2d);
			
			try{
				block = sector->getBlockNoCreate(p.Y);
				/*
					Update an existing block
				*/
				//std::cout<<"Updating"<<std::endl;
				block->deSerialize(istr, ser_version);
				//block->setChangedFlag();
			}
			catch(InvalidPositionException &e)
			{
				/*
					Create a new block
				*/
				//std::cout<<"Creating new"<<std::endl;
				block = new MapBlock(&m_env.getMap(), p);
				block->deSerialize(istr, ser_version);
				sector->insertBlock(block);
				//block->setChangedFlag();
			}
		} //envlock
		
		
		// Old version has zero lighting, update it.
		// This doesn't work perfectly but who cares.
		if(ser_version == 0 || ser_version == 1)
		{
			core::list<MapBlock*> blocks_changed;
			blocks_changed.push_back(block);
			core::map<v3s16, MapBlock*> modified_blocks;
			m_env.getMap().updateLighting(blocks_changed, modified_blocks);
		}

		/*
			Update Mesh
		*/
		block->updateMesh();
		
		/*
			Remove from history
		*/
		{
			JMutexAutoLock lock(m_fetchblock_mutex);
			
			if(m_fetchblock_history.find(p) != NULL)
			{
				m_fetchblock_history.remove(p);
			}
		}
		
	}
	else
	{
		dout_client<<DTIME<<"WARNING: Client: Ignoring unknown command "
				<<command<<std::endl;
	}

	goto getdata;

	} //try
	catch(con::PeerNotFoundException &e)
	{
		dout_client<<DTIME<<"Client::AsyncProcessData(): Cancelling: The server"
				" connection doesn't exist (a timeout or not yet connected?)"<<std::endl;
		return false;
	}
}

void Client::Send(u16 channelnum, SharedBuffer<u8> data, bool reliable)
{
	JMutexAutoLock lock(m_con_mutex);
	m_con.Send(PEER_ID_SERVER, channelnum, data, reliable);
}

/*u8 Client::getServerSerVer()
{
	JMutexAutoLock lock(m_con_mutex);
	con::Peer *peer;
	try{
		// All data is coming from the server
		peer = m_con.GetPeer(PEER_ID_SERVER);
		//peer = m_con.GetPeer(sender_peer_id);
	}
	catch(con::PeerNotFoundException &e)
	{
		std::cout<<DTIME<<"Client::ProcessData(): PeerNotFoundException"
				<<std::endl;
		return;
	}

	return peer->serialization_version;
}*/

void Client::fetchBlock(v3s16 p, u8 flags)
{
	if(connectedAndInitialized() == false)
		throw ClientNotReadyException
		("ClientNotReadyException: connectedAndInitialized() == false");

	/*std::cout<<"Client::fetchBlock(): Sending GETBLOCK for ("
			<<p.X<<","<<p.Y<<","<<p.Z<<")"<<std::endl;*/

	JMutexAutoLock conlock(m_con_mutex);

	SharedBuffer<u8> data(9);
	writeU16(&data[0], TOSERVER_GETBLOCK);
	writeS16(&data[2], p.X);
	writeS16(&data[4], p.Y);
	writeS16(&data[6], p.Z);
	writeU8(&data[8], flags);
	m_con.Send(PEER_ID_SERVER, 1, data, true);
}

/*
	Calls fetchBlock() on some nearby missing blocks.

	Returns when any of various network load indicators go over limit.

	Does nearly the same thing as the old updateChangedVisibleArea()
*/
void Client::fetchBlocks()
{
	if(connectedAndInitialized() == false)
		throw ClientNotReadyException
		("ClientNotReadyException: connectedAndInitialized() == false");

	JMutexAutoLock envlock(m_env_mutex);
	
	v3s16 center_nodepos = Map::floatToInt(camera_position);

	v3s16 center = Map::getNodeBlockPos(center_nodepos);

	g_range_mutex.Lock();
	s16 d_max_gen = g_forcedfetch_range_nodes / MAP_BLOCKSIZE;
	s16 d_max_view = g_viewing_range_nodes / MAP_BLOCKSIZE;
	bool viewing_range_all = g_viewing_range_all;
	g_range_mutex.Unlock();
	
	u8 ser_version = SER_FMT_VER_INVALID;
	{
		con::Peer *peer = m_con.GetPeer(PEER_ID_SERVER);
		ser_version = peer->serialization_version;
	}
	
	s16 d_tryfetch = d_max_view;

	if(viewing_range_all)
	{
		d_tryfetch = VIEWING_RANGE_NODES_MAX;
	}
	
	// If server version is old, limit range, because otherwise it
	// would have to generate huge amounts of map because of our
	// viewing range.
	if(ser_version <= 2)
	{
		d_tryfetch = d_max_gen;
	}
	
	for(s16 d=0; d <= d_tryfetch; d++)
	{
		core::list<v3s16> list;
		getFacePositions(list, d);
		
		bool optional = d > d_max_gen;

		core::list<v3s16>::Iterator li;
		for(li=list.begin(); li!=list.end(); li++)
		{
			v3s16 p = *li + center;
			
			// Limit the generating area vertically to half
			if(abs(p.Y - center.Y) > d_max_gen / 2)
				optional = true;
			
			/*
				Fetch throttling
			*/
			{
				JMutexAutoLock lock(m_fetchblock_mutex);
				
				// Don't allow too many simultaneous fetches
				if(m_fetchblock_history.size() >= MAX_SIMULTANEOUS_BLOCK_FETCHES)
					return;
			}

			/*
				Check if map already has this block;
				-> if not, check if it is in the fetch history
				-> if not, send a fetch command.
			*/
			MapBlock *block;
			try
			{
				block = m_env.getMap().getBlockNoCreate(p);
				// Already has
				continue;
			}
			catch(InvalidPositionException &e)
			{
				// Doesn't have.
				// Continue outside of catch
			}
			
			/*
				Handle histories
			*/
			{
				JMutexAutoLock lock(m_fetchblock_mutex);
				
				/*
					Check if fetch command was already sent for this block.
					If so, cancel.
				*/
				if(m_fetchblock_history.find(p) != NULL)
					continue;
					
				/*
					Check if the block was not found optionally before
				*/
				if(m_opt_not_found_history.find(p) != NULL)
				{
					// If it is optional, cancel.
					if(optional)
					{
						continue;
					}
					/*
						If it is not optional and it was not found
						optionally before, remove it from the not-found
						history.
					*/
					else
					{
						m_opt_not_found_history.remove(p);
					}
				}
			}

			/*std::cout<<"Client::fetchBlocks(): Calling fetchBlock() for ("
					<<p.X<<","<<p.Y<<","<<p.Z<<")"<<std::endl;*/
			
			/*
				Send fetch command
			*/
			
			u8 flags = 0;
			if(optional)
				flags |= TOSERVER_GETBLOCK_FLAG_OPTIONAL;
			
			fetchBlock(p, flags);
			
			/*
				Add to history
			*/
			{
				JMutexAutoLock lock(m_fetchblock_mutex);
				m_fetchblock_history.insert(p, 0.0);
			}
		}
	}
/*fetched_enough:
	return;*/
}

bool Client::isFetchingBlocks()
{
	JMutexAutoLock conlock(m_con_mutex);
	con::Peer *peer = m_con.GetPeerNoEx(PEER_ID_SERVER);
	// Not really fetching but can't fetch more.
	if(peer == NULL) return true;

	con::Channel *channel = &(peer->channels[1]);
	/*
		NOTE: Channel 0 should always be used for fetching blocks,
		      and for nothing else.
	*/
	if(channel->incoming_reliables.size() > 0)
		return true;
	if(channel->outgoing_reliables.size() > 0)
		return true;
	return false;
}

IncomingPacket Client::getPacket()
{
	JMutexAutoLock lock(m_incoming_queue_mutex);
	
	core::list<IncomingPacket>::Iterator i;
	// Refer to first one
	i = m_incoming_queue.begin();

	// If queue is empty, return empty packet
	if(i == m_incoming_queue.end()){
		IncomingPacket packet;
		return packet;
	}
	
	// Pop out first packet and return it
	IncomingPacket packet = *i;
	m_incoming_queue.erase(i);
	return packet;
}

void Client::removeNode(v3s16 nodepos)
{
	if(connectedAndInitialized() == false){
		dout_client<<DTIME<<"Client::removeNode() cancelled (not connected)"
				<<std::endl;
		return;
	}
	
	// Test that the position exists
	try{
		JMutexAutoLock envlock(m_env_mutex);
		m_env.getMap().getNode(nodepos);
	}
	catch(InvalidPositionException &e)
	{
		dout_client<<DTIME<<"Client::removeNode() cancelled (doesn't exist)"
				<<std::endl;
		return;
	}

	SharedBuffer<u8> data(8);
	writeU16(&data[0], TOSERVER_REMOVENODE);
	writeS16(&data[2], nodepos.X);
	writeS16(&data[4], nodepos.Y);
	writeS16(&data[6], nodepos.Z);
	Send(0, data, true);
}

void Client::addNode(v3s16 nodepos, MapNode n)
{
	if(connectedAndInitialized() == false){
		dout_client<<DTIME<<"Client::addNode() cancelled (not connected)"
				<<std::endl;
		return;
	}
	
	// Test that the position exists
	try{
		JMutexAutoLock envlock(m_env_mutex);
		m_env.getMap().getNode(nodepos);
	}
	catch(InvalidPositionException &e)
	{
		dout_client<<DTIME<<"Client::addNode() cancelled (doesn't exist)"
				<<std::endl;
		return;
	}

	con::Peer *peer = m_con.GetPeer(PEER_ID_SERVER);
	u8 ser_version = peer->serialization_version;

	u8 datasize = 8 + MapNode::serializedLength(ser_version);
	SharedBuffer<u8> data(datasize);
	writeU16(&data[0], TOSERVER_ADDNODE);
	writeS16(&data[2], nodepos.X);
	writeS16(&data[4], nodepos.Y);
	writeS16(&data[6], nodepos.Z);
	n.serialize(&data[8], ser_version);
	Send(0, data, true);
}

void Client::sendPlayerPos()
{
	JMutexAutoLock envlock(m_env_mutex);
	
	Player *myplayer = m_env.getLocalPlayer();
	if(myplayer == NULL)
		return;
	
	u16 our_peer_id;
	{
		JMutexAutoLock lock(m_con_mutex);
		our_peer_id = m_con.GetPeerID();
	}
	
	// Set peer id if not set already
	if(myplayer->peer_id == PEER_ID_NEW)
		myplayer->peer_id = our_peer_id;
	// Check that an existing peer_id is the same as the connection's
	assert(myplayer->peer_id == our_peer_id);
	
	v3f pf = myplayer->getPosition();
	v3s32 position(pf.X*100, pf.Y*100, pf.Z*100);
	v3f sf = myplayer->speed;
	v3s32 speed(sf.X*100, sf.Y*100, sf.Z*100);

	SharedBuffer<u8> data(2+12+12);
	writeU16(&data[0], TOSERVER_PLAYERPOS);
	writeV3S32(&data[2], position);
	writeV3S32(&data[2+12], speed);
	// Send as unreliable
	Send(0, data, false);
}

/*void Client::sendPlayerInfo()
{
	JMutexAutoLock envlock(m_env_mutex);
	
	Player *myplayer = m_env.getLocalPlayer();
	if(myplayer == NULL)
		return;
	
	SharedBuffer<u8> data(2+PLAYERNAME_SIZE);
	writeU16(&data[0], TOSERVER_PLAYERINFO);
	snprintf((char*)&data[2], PLAYERNAME_SIZE, "%s", myplayer->getName());
	// Send as reliable
	Send(0, data, true);
}*/

void Client::updateCamera(v3f pos, v3f dir)
{
	JMutexAutoLock envlock(m_env_mutex);
	m_env.getMap().updateCamera(pos, dir);
	camera_position = pos;
	camera_direction = dir;
}

MapNode Client::getNode(v3s16 p)
{
	JMutexAutoLock envlock(m_env_mutex);
	return m_env.getMap().getNode(p);
}

/*f32 Client::getGroundHeight(v2s16 p)
{
	JMutexAutoLock envlock(m_env_mutex);
	return m_env.getMap().getGroundHeight(p);
}*/

bool Client::isNodeUnderground(v3s16 p)
{
	JMutexAutoLock envlock(m_env_mutex);
	return m_env.getMap().isNodeUnderground(p);
}

Player * Client::getLocalPlayer()
{
	JMutexAutoLock envlock(m_env_mutex);
	return m_env.getLocalPlayer();
}

core::list<Player*> Client::getPlayers()
{
	JMutexAutoLock envlock(m_env_mutex);
	return m_env.getPlayers();
}

void Client::printDebugInfo(std::ostream &os)
{
	JMutexAutoLock lock1(m_fetchblock_mutex);
	JMutexAutoLock lock2(m_incoming_queue_mutex);

	os<<"m_incoming_queue.getSize()="<<m_incoming_queue.getSize()
		<<", m_fetchblock_history.size()="<<m_fetchblock_history.size()
		<<", m_opt_not_found_history.size()="<<m_opt_not_found_history.size()
		<<std::endl;
}
	

