/*
(c) 2010 Perttu Ahola <celeron55@gmail.com>
*/

#include "heightmap.h"
#include "serialization.h"

bool g_heightmap_debugprint = false;

f32 FixedHeightmap::avgNeighbours(v2s16 p, s16 d)
{
	//printf("avgNeighbours((%i,%i), %i): ", p.X, p.Y, d);

	v2s16 dirs[4] = {
		v2s16(1,0),
		v2s16(0,1),
		v2s16(-1,0),
		v2s16(0,-1)
	};
	f32 sum = 0.0;
	f32 count = 0.0;
	for(u16 i=0; i<4; i++){
		v2s16 p2 = p + dirs[i] * d;
		f32 n = getGroundHeightParent(p2);
		if(n < GROUNDHEIGHT_VALID_MINVALUE)
			continue;
		sum += n;
		count += 1.0;
		//printf("(%i,%i)=%f ", p2.X, p2.Y, n);
	}
	//printf("\n");
	assert(count > 0.001);
	return sum / count;
}

f32 FixedHeightmap::avgDiagNeighbours(v2s16 p, s16 d)
{
	/*printf("avgDiagNeighbours((%i,%i), %i): ",
		p.X, p.Y, d);*/
		
	v2s16 dirs[4] = {
		v2s16(1,1),
		v2s16(-1,-1),
		v2s16(-1,1),
		v2s16(1,-1)
	};
	f32 sum = 0.0;
	f32 count = 0.0;
	for(u16 i=0; i<4; i++){
		v2s16 p2 = p + dirs[i] * d;
		f32 n = getGroundHeightParent(p2);
		if(n < GROUNDHEIGHT_VALID_MINVALUE)
			continue;
		sum += n;
		count += 1.0;
		//printf("(%i,%i)=%f ", p2.X, p2.Y, n);
	}
	//printf("\n");
	assert(count > 0.001);
	return sum / count;
}

/*
	Adds a point to transform into a diamond pattern
	a = 2, 4, 8, 16, ...
*/
void FixedHeightmap::makeDiamond(v2s16 center, s16 a, f32 randmax)
{
	f32 n = avgDiagNeighbours(center, a/2);
	// Add (-1.0...1.0) * randmax
	n += ((float)rand() / (float)(RAND_MAX/2) - 1.0)*randmax;
	setGroundHeightParent(center, n);
}

/*
	Adds points to transform into a diamond pattern
	a = 2, 4, 8, 16, ...
*/
void FixedHeightmap::makeDiamonds(s16 a, f32 randmax)
{
	s16 count_y = (H-1) / a;
	s16 count_x = (W-1) / a;
	for(s32 yi=0; yi<count_y; yi++){
		for(s32 xi=0; xi<count_x; xi++){
			v2s16 center(xi*a + a/2, yi*a + a/2);
			makeDiamond(center, a, randmax);
		}
	}
}

/*
	Adds a point to transform into a square pattern
	a = 2, 4, 8, 16, ...
*/
void FixedHeightmap::makeSquare(v2s16 center, s16 a, f32 randmax)
{
	f32 n = avgNeighbours(center, a/2);
	// Add (-1.0...1.0) * randmax
	n += ((float)rand() / (float)(RAND_MAX/2) - 1.0)*randmax;
	setGroundHeightParent(center, n);
}

/*
	Adds points to transform into a square pattern
	a = 2, 4, 8, 16, ...
*/
void FixedHeightmap::makeSquares(s16 a, f32 randmax)
{
	/*
		This is a bit tricky; the points have to be kind of
		interlaced. (but not exactly)
		
		On even rows (0, 2, 4, ...) the nodes are horizontally
		at a*n+a/2 positions, vertically at a*n.

		On odd rows (1, 3, 5, ...) the nodes are horizontally
		at a*n positions, vertically at a*n+a/2.
	*/

	s16 count_y = (H-1) / a + 1;
	for(s32 yi=0; yi<=count_y; yi++){
		s16 count_x;
		// Even rows
		count_x = (W-1) / a;
		for(s32 xi=0; xi<count_x; xi++){
			v2s16 center(xi*a + a/2, yi*a);
			makeSquare(center, a, randmax);
		}
		if(yi >= count_y - 1)
			break;
		// odd rows
		count_x = (W-1) / a + 1;
		for(s32 xi=0; xi<count_x; xi++){
			v2s16 center(xi*a, yi*a + a/2);
			makeSquare(center, a, randmax);
		}
	}
}

void FixedHeightmap::DiamondSquare(f32 randmax, f32 randfactor)
{
	u16 a;
	if(W < H)
		a = W-1;
	else
		a = H-1;
	
	// Check that a is a power of two
	if((a & (a-1)) != 0)
		throw;
	
	while(a >= 2)
	{
		makeDiamonds(a, randmax);
		if(HEIGHTMAP_DEBUGPRINT){
			printf("MakeDiamonds a=%i result:\n", a);
			print();
		}
		makeSquares(a, randmax);
		if(HEIGHTMAP_DEBUGPRINT){
			printf("MakeSquares a=%i result:\n", a);
			print();
		}
		a /= 2;
		randmax *= randfactor;
	}
}

u32 FixedHeightmap::serializedLength(u8 version, u16 blocksize)
{
	if(!ser_ver_supported(version))
		throw VersionMismatchException("ERROR: FixedHeightmap format not supported");
	
	// Any version
	{
		/*// [0] s32 blocksize
		// [4] v2s16 pos_on_master
		// [8] s32 data[W*H] (W=H=blocksize+1)
		return 4 + 4 + (blocksize+1)*(blocksize+1)*4;*/

		// [8] s32 data[W*H] (W=H=blocksize+1)
		return (blocksize+1)*(blocksize+1)*4;
	}
}

u32 FixedHeightmap::serializedLength(u8 version)
{
	return serializedLength(version, m_blocksize);
}

void FixedHeightmap::serialize(u8 *dest, u8 version)
{
	//std::cout<<"FixedHeightmap::serialize"<<std::endl;

	if(!ser_ver_supported(version))
		throw VersionMismatchException("ERROR: FixedHeightmap format not supported");
	
	// Any version
	{
		/*writeU32(&dest[0], m_blocksize);
		writeV2S16(&dest[4], m_pos_on_master);
		u32 nodecount = W*H;
		for(u32 i=0; i<nodecount; i++)
		{
			writeS32(&dest[8+i*4], (s32)(m_data[i]*1000.0));
		}*/

		u32 nodecount = W*H;
		for(u32 i=0; i<nodecount; i++)
		{
			writeS32(&dest[i*4], (s32)(m_data[i]*1000.0));
		}
	}
}

void FixedHeightmap::deSerialize(u8 *source, u8 version)
{
	/*std::cout<<"FixedHeightmap::deSerialize m_blocksize="
			<<m_blocksize<<std::endl;*/

	if(!ser_ver_supported(version))
		throw VersionMismatchException("ERROR: FixedHeightmap format not supported");
	
	// Any version
	{
		u32 nodecount = (m_blocksize+1)*(m_blocksize+1);
		for(u32 i=0; i<nodecount; i++)
		{
			m_data[i] = ((f32)readS32(&source[i*4]))/1000.0;
		}

		/*printf("source[0,1,2,3]=%x,%x,%x,%x\n",
				(int)source[0]&0xff,
				(int)source[1]&0xff,
				(int)source[2]&0xff,
				(int)source[3]&0xff);
				
		std::cout<<"m_data[0]="<<m_data[0]<<", "
				<<"readS32(&source[0])="<<readS32(&source[0])
				<<std::endl;
		std::cout<<"m_data[4*4]="<<m_data[4*4]<<", "
				<<"readS32(&source[4*4])="<<readS32(&source[4*4])
				<<std::endl;*/
	}
}

void FixedHeightmap::generateContinued(f32 randmax, f32 randfactor,
		f32 *corners)
{
	if(HEIGHTMAP_DEBUGPRINT){
		std::cout<<"FixedHeightmap("<<m_pos_on_master.X
				<<","<<m_pos_on_master.Y
				<<")::generateContinued()"<<std::endl;
	}
	/*
		TODO: Implement changing neighboring heightmaps when needed
	*/

	// Works only with blocksize=2,4,8,16,32,64,...
	s16 a = m_blocksize;
	
	// Check that a is a power of two
	if((a & (a-1)) != 0)
		throw;
	
	// Overwrite with GROUNDHEIGHT_NOTFOUND_SETVALUE
	for(s16 y=0; y<=a; y++){
		for(s16 x=0; x<=a; x++){
			v2s16 p(x,y);
			setGroundHeight(p, GROUNDHEIGHT_NOTFOUND_SETVALUE);
		}
	}

#if 1
	/*
		Seed borders from master heightmap
		TODO: Check that this works properly
	*/
	struct SeedSpec
	{
		v2s16 neighbour_start;
		v2s16 heightmap_start;
		v2s16 dir;
	};

	SeedSpec seeds[4] =
	{
		{ // Z- edge on X-axis
			v2s16(0, -1), // neighbour_start
			v2s16(0, 0), // heightmap_start
			v2s16(1, 0) // dir
		},
		{ // Z+ edge on X-axis
			v2s16(0, m_blocksize),
			v2s16(0, m_blocksize),
			v2s16(1, 0)
		},
		{ // X- edge on Z-axis
			v2s16(-1, 0),
			v2s16(0, 0),
			v2s16(0, 1)
		},
		{ // X+ edge on Z-axis
			v2s16(m_blocksize, 0),
			v2s16(m_blocksize, 0),
			v2s16(0, 1)
		},
	};

	for(s16 i=0; i<4; i++){
		v2s16 npos = seeds[i].neighbour_start + m_pos_on_master * m_blocksize;
		v2s16 hpos = seeds[i].heightmap_start;
		for(s16 s=0; s<m_blocksize+1; s++){
			f32 h = m_master->getGroundHeight(npos, false);
			//std::cout<<"h="<<h<<std::endl;
			if(h < GROUNDHEIGHT_VALID_MINVALUE)
				continue;
				//break;
			setGroundHeight(hpos, h);
			hpos += seeds[i].dir;
			npos += seeds[i].dir;
		}
	}
	
	if(HEIGHTMAP_DEBUGPRINT){
		std::cout<<"borders seeded:"<<std::endl;
		print();
	}
#endif

	/*
		Fill with corners[] (if not already set)
	*/
	v2s16 dirs[4] = {
		v2s16(0,0),
		v2s16(1,0),
		v2s16(1,1),
		v2s16(0,1),
	};
	for(u16 i=0; i<4; i++){
		v2s16 npos = dirs[i] * a;
		// Don't replace already seeded corners
		f32 h = getGroundHeight(npos);
		if(h > GROUNDHEIGHT_VALID_MINVALUE)
			continue;
		setGroundHeight(dirs[i] * a, corners[i]);
	}
	
	if(HEIGHTMAP_DEBUGPRINT){
		std::cout<<"corners filled:"<<std::endl;
		print();
	}

	/*std::cout<<"Seeded heightmap:"<<std::endl;
	print();*/

	DiamondSquare(randmax, randfactor);
}


/*
	UnlimitedHeightmap
*/

void UnlimitedHeightmap::print()
{
	s16 minx =  10000;
	s16 miny =  10000;
	s16 maxx = -10000;
	s16 maxy = -10000;
	core::map<v2s16, FixedHeightmap*>::Iterator i;
	i = m_heightmaps.getIterator();
	if(i.atEnd()){
		printf("UnlimitedHeightmap::print(): empty.\n");
		return;
	}
	for(; i.atEnd() == false; i++)
	{
		v2s16 p = i.getNode()->getValue()->getPosOnMaster();
		if(p.X < minx) minx = p.X;
		if(p.Y < miny) miny = p.Y;
		if(p.X > maxx) maxx = p.X;
		if(p.Y > maxy) maxy = p.Y;
	}
	minx = minx * m_blocksize;
	miny = miny * m_blocksize;
	maxx = (maxx+1) * m_blocksize;
	maxy = (maxy+1) * m_blocksize;
	printf("UnlimitedHeightmap::print(): from (%i,%i) to (%i,%i)\n",
			minx, miny, maxx, maxy);
	for(s32 y=miny; y<=maxy; y++){
		for(s32 x=minx; x<=maxx; x++){
			f32 n = getGroundHeight(v2s16(x,y), false);
			if(n < GROUNDHEIGHT_VALID_MINVALUE)
				printf("  -   ");
			else
				printf("% -5.1f ", getGroundHeight(v2s16(x,y), false));
		}
		printf("\n");
	}
}
	
FixedHeightmap * UnlimitedHeightmap::getHeightmap(v2s16 p, bool generate)
{
	/*
		We want to check that all neighbours of the wanted heightmap
		exist.
		This is because generating the neighboring heightmaps will
		modify the current one.
	*/
	
	// Go through all neighbors (corners also) and the current one.
	// If one doesn't exist:
	//   If generate==true, generate it
	//   If generate==false, return false.
	for(s16 x=p.X-1; x<=p.X+1; x++)
	for(s16 y=p.Y-1; y<=p.Y+1; y++)
	{
		v2s16 p(x,y);

		// Check if exists
		core::map<v2s16, FixedHeightmap*>::Node *n = m_heightmaps.find(p);
		if(n != NULL)
			continue;
		
		// Doesn't exist

		// Fail if generate==false
		if(generate==false)
			throw InvalidPositionException
			("Child heightmap or one of its neighbors doesn't exist");
			
		// Generate it

		FixedHeightmap *heightmap = new FixedHeightmap(this, p, m_blocksize);

		m_heightmaps.insert(p, heightmap);
		
		f32 corners[] = {m_basevalue, m_basevalue, m_basevalue, m_basevalue};
		heightmap->generateContinued(m_randmax, m_randfactor, corners);
	}

	core::map<v2s16, FixedHeightmap*>::Node *n = m_heightmaps.find(p);

	if(n != NULL)
	{
		return n->getValue();
	}
	else
	{
		assert(0);
		throw InvalidPositionException
		("Something went really wrong in UnlimitedHeightmap::getHeightmap");
	}
}

f32 UnlimitedHeightmap::getGroundHeight(v2s16 p, bool generate)
{
	v2s16 heightmappos = getNodeHeightmapPos(p);
	v2s16 relpos = p - heightmappos*m_blocksize;
	try{
		FixedHeightmap * href = getHeightmap(heightmappos, generate);
		f32 h = href->getGroundHeight(relpos);
		if(h > GROUNDHEIGHT_VALID_MINVALUE)
			return h;
	}
	catch(InvalidPositionException){}
	
	/*
		OK, wasn't there.
		
		Mercilessly try to get it somewhere.
	*/
#if 1
	if(relpos.X == 0){
		try{
			FixedHeightmap * href = getHeightmap(
					heightmappos-v2s16(1,0), false);
			f32 h = href->getGroundHeight(v2s16(m_blocksize, relpos.Y));
			if(h > GROUNDHEIGHT_VALID_MINVALUE)
				return h;
		}
		catch(InvalidPositionException){}
	}
	if(relpos.Y == 0){
		try{
			FixedHeightmap * href = getHeightmap(
					heightmappos-v2s16(0,1), false);
			f32 h = href->getGroundHeight(v2s16(relpos.X, m_blocksize));
			if(h > GROUNDHEIGHT_VALID_MINVALUE)
				return h;
		}
		catch(InvalidPositionException){}
	}
	if(relpos.X == 0 && relpos.Y == 0){
		try{
			FixedHeightmap * href = getHeightmap(
					heightmappos-v2s16(1,1), false);
			f32 h = href->getGroundHeight(v2s16(m_blocksize, m_blocksize));
			if(h > GROUNDHEIGHT_VALID_MINVALUE)
				return h;
		}
		catch(InvalidPositionException){}
	}
#endif
	return GROUNDHEIGHT_NOTFOUND_SETVALUE;
}

void UnlimitedHeightmap::setGroundHeight(v2s16 p, f32 y, bool generate)
{
	v2s16 heightmappos = getNodeHeightmapPos(p);
	v2s16 relpos = p - heightmappos*m_blocksize;
	/*std::cout<<"UnlimitedHeightmap::setGroundHeight(("
			<<p.X<<","<<p.Y<<"), "<<y<<"): "
			<<"heightmappos=("<<heightmappos.X<<","
			<<heightmappos.Y<<") relpos=("
			<<relpos.X<<","<<relpos.Y<<")"
			<<std::endl;*/
	try{
		FixedHeightmap * href = getHeightmap(heightmappos, generate);
		href->setGroundHeight(relpos, y);
	}catch(InvalidPositionException){}
	// Update in neighbour heightmap if it's at border
	if(relpos.X == 0){
		try{
			FixedHeightmap * href = getHeightmap(
					heightmappos-v2s16(1,0), generate);
			href->setGroundHeight(v2s16(m_blocksize, relpos.Y), y);
		}catch(InvalidPositionException){}
	}
	if(relpos.Y == 0){
		try{
			FixedHeightmap * href = getHeightmap(
					heightmappos-v2s16(0,1), generate);
			href->setGroundHeight(v2s16(relpos.X, m_blocksize), y);
		}catch(InvalidPositionException){}
	}
	if(relpos.X == m_blocksize && relpos.Y == m_blocksize){
		try{
			FixedHeightmap * href = getHeightmap(
					heightmappos-v2s16(1,1), generate);
			href->setGroundHeight(v2s16(m_blocksize, m_blocksize), y);
		}catch(InvalidPositionException){}
	}
}


SharedBuffer<u8> UnlimitedHeightmap::serialize(u8 version)
{
	//std::cout<<"UnlimitedHeightmap::serialize()"<<std::endl;

	if(!ser_ver_supported(version))
		throw VersionMismatchException("ERROR: UnlimitedHeightmap format not supported");
	
	// Any version
	{
		/*
			[0] u16 blocksize
			[2] s32 randmax*1000
			[6] s32 randfactor*1000
			[10] s32 basevalue*1000
			[14] u32 heightmap_count
			[18] X * (v2s16 pos + heightmap)
		*/
		u32 heightmap_size =
				FixedHeightmap::serializedLength(version, m_blocksize);
		u32 heightmap_count = m_heightmaps.size();

		//std::cout<<"heightmap_size="<<heightmap_size<<std::endl;

		u32 datasize = 2+4+4+4+4+heightmap_count*(4+heightmap_size);
		SharedBuffer<u8> data(datasize);

		writeU16(&data[0], m_blocksize);
		writeU32(&data[2], (s32)(m_randmax*1000.0));
		writeU32(&data[6], (s32)(m_randfactor*1000.0));
		writeU32(&data[10], (s32)(m_basevalue*1000.0));
		writeU32(&data[14], heightmap_count);

		core::map<v2s16, FixedHeightmap*>::Iterator j;
		j = m_heightmaps.getIterator();
		u32 i=0;
		for(; j.atEnd() == false; j++)
		{
			FixedHeightmap *hm = j.getNode()->getValue();
			v2s16 pos = j.getNode()->getKey();
			writeV2S16(&data[18+i*(4+heightmap_size)], pos);
			hm->serialize(&data[18+i*(4+heightmap_size)+4], version);
			i++;
		}
		return data;
	}
}

/*UnlimitedHeightmap * UnlimitedHeightmap::deSerialize(u8 *data, u32 maxsize,
		u32 &usedsize, u8 version)*/
UnlimitedHeightmap * UnlimitedHeightmap::deSerialize
		(std::istream &istr, u8 version)
{
	if(!ser_ver_supported(version))
		throw VersionMismatchException("ERROR: UnlimitedHeightmap format not supported");
	
	// Any version
	{
		/*
			[0] u16 blocksize
			[2] s32 randmax*1000
			[6] s32 randfactor*1000
			[10] s32 basevalue*1000
			[14] u32 heightmap_count
			[18] X * (v2s16 pos + heightmap)
		*/
		SharedBuffer<u8> data(18);
		istr.read((char*)*data, 18);
		if(istr.gcount() != 18)
			throw SerializationError
					("UnlimitedHeightmap::deSerialize: no enough input data");
		s16 blocksize = readU16(&data[0]);
		f32 randmax = (f32)readU32(&data[2]) / 1000.0;
		f32 randfactor = (f32)readU32(&data[6]) / 1000.0;
		f32 basevalue = (f32)readU32(&data[10]) / 1000.0;
		u32 heightmap_count = readU32(&data[14]);

		/*std::cout<<"UnlimitedHeightmap::deSerialize():"
				<<" blocksize="<<blocksize
				<<" heightmap_count="<<heightmap_count
				<<std::endl;*/

		u32 heightmap_size =
				FixedHeightmap::serializedLength(version, blocksize);

		//std::cout<<"heightmap_size="<<heightmap_size<<std::endl;

		UnlimitedHeightmap *hm = new UnlimitedHeightmap
				(blocksize, randmax, randfactor, basevalue);

		for(u32 i=0; i<heightmap_count; i++)
		{
			//std::cout<<"i="<<i<<std::endl;
			SharedBuffer<u8> data(4+heightmap_size);
			istr.read((char*)*data, 4+heightmap_size);
			if(istr.gcount() != (s32)(4+heightmap_size)){
				delete hm;
				throw SerializationError
						("UnlimitedHeightmap::deSerialize: no enough input data");
			}
			v2s16 pos = readV2S16(&data[0]);
			/*u32 usedsize;
			FixedHeightmap *f = FixedHeightmap::deSerialize(
					&data[4],
					heightmap_size,
					usedsize,
					hm,
					version);*/
			FixedHeightmap *f = new FixedHeightmap(hm, pos, blocksize);
			f->deSerialize(&data[4], version);
			hm->m_heightmaps.insert(pos, f);
		}
		return hm;
	}
}


