#include "server.h"
#include "utility.h"
#include <iostream>
#include "clientserver.h"
#include "map.h"
#include "jmutexautolock.h"
#include "main.h"
#include "constants.h"

void * ServerThread::Thread()
{
	ThreadStarted();

	while(getRun())
	{
		try{
			m_server->AsyncRunStep();
		
			//dout_server<<"Running m_server->Receive()"<<std::endl;
			m_server->Receive();
		}
		catch(con::NoIncomingDataException &e)
		{
		}
		/*catch(std::exception &e)
		{
			derr_server<<"ServerThread: Some exception: "
					<<e.what()<<std::endl;
		}*/
	}
	

	return NULL;
}

void * EmergeThread::Thread()
{
	ThreadStarted();

	bool debug=false;

	/*
		Get block info from queue, emerge them and send them
		to clients.

		After queue is empty, exit.
	*/
	while(getRun())
	{
		QueuedBlockEmerge *qptr = m_server->m_emerge_queue.pop();
		if(qptr == NULL)
			break;
		
		SharedPtr<QueuedBlockEmerge> q(qptr);

		v3s16 &p = q->pos;
		
		//derr_server<<"EmergeThread::Thread(): running"<<std::endl;
		
		/*
			Try to emerge it from somewhere.

			If it is only wanted as optional, only loading from disk
			will be allowed.
		*/
		
		/*
			Check if any peer wants it as non-optional. In that case it
			will be generated.

			Also decrement the emerge queue count in clients.
		*/

		bool optional = true;

		{
			core::map<u16, u8>::Iterator i;
			for(i=q->peer_ids.getIterator(); i.atEnd()==false; i++)
			{
				//u16 peer_id = i.getNode()->getKey();

				// Check flags
				u8 flags = i.getNode()->getValue();
				if((flags & TOSERVER_GETBLOCK_FLAG_OPTIONAL) == false)
					optional = false;
				
			}
		}

		/*std::cerr<<"EmergeThread: p="
				<<"("<<p.X<<","<<p.Y<<","<<p.Z<<") "
				<<"optional="<<optional<<std::endl;*/
		
		ServerMap &map = ((ServerMap&)m_server->m_env.getMap());
			
		core::map<v3s16, MapBlock*> changed_blocks;
		core::map<v3s16, MapBlock*> lighting_invalidated_blocks;

		MapBlock *block = NULL;
		bool got_block = true;
		try{
			bool only_from_disk = false;
			
			if(optional)
				only_from_disk = true;

			JMutexAutoLock(m_server->m_env_mutex);
			
			block = map.emergeBlock(
					p,
					only_from_disk,
					changed_blocks,
					lighting_invalidated_blocks);
			
			// If it is a dummy, block was not found on disk
			if(block->isDummy())
			{
				//std::cerr<<"EmergeThread: Got a dummy block"<<std::endl;
				got_block = false;
			}
		}
		catch(InvalidPositionException &e)
		{
			// Block not found.
			// This happens when position is over limit.
			got_block = false;
		}
		
		core::map<v3s16, MapBlock*> modified_blocks;
			
		if(got_block)
		{
			if(debug && changed_blocks.size() > 0)
			{
				dout_server<<DTIME<<"Got changed_blocks: ";
				for(core::map<v3s16, MapBlock*>::Iterator i = changed_blocks.getIterator();
						i.atEnd() == false; i++)
				{
					MapBlock *block = i.getNode()->getValue();
					v3s16 p = block->getPos();
					dout_server<<"("<<p.X<<","<<p.Y<<","<<p.Z<<") ";
				}
				dout_server<<std::endl;
			}

			/*
				Collect a list of blocks that have been modified in
				addition to the fetched one.
			*/

			// Add all the "changed blocks"
			for(core::map<v3s16, MapBlock*>::Iterator i = changed_blocks.getIterator();
					i.atEnd() == false; i++)
			{
				MapBlock *block = i.getNode()->getValue();
				modified_blocks.insert(block->getPos(), block);
			}

			// Update lighting without locking the environment mutex,
			// add modified blocks to changed blocks
			map.updateLighting(lighting_invalidated_blocks, modified_blocks);
		}

		/*
			Set sent status of modified blocks on clients
		*/
	
		// NOTE: Server's clients are also behind the connection mutex
		JMutexAutoLock lock(m_server->m_con_mutex);

		/*
			Add the originally fetched block to the modified list
		*/
		if(got_block)
		{
			modified_blocks.insert(p, block);
		}
		
		/*
			Set the modified blocks unsent for all the clients
		*/
		
		for(core::map<u16, RemoteClient*>::Iterator
				i = m_server->m_clients.getIterator();
				i.atEnd() == false; i++)
		{
			RemoteClient *client = i.getNode()->getValue();
			
			if(modified_blocks.size() > 0)
			{
				// Remove block from sent history
				client->SetBlocksNotSent(modified_blocks);
			}
			
			// Decrement emerge queue count of client
			client->BlockEmerged();
		}
	}

	return NULL;
}

void RemoteClient::SendBlocks(Server *server, float dtime)
{
	/*
		TODO

		Find what blocks to send to the client next, and send them.

		Throttling is based on limiting the amount of blocks "flying"
		at a given time.

		Generating blocks is still handled by TOSERVER_GETBLOCK
		and the emerge thread.
		- TODO: Handle them here too
	*/

	// Can't send anything without knowing version
	if(serialization_version == SER_FMT_VER_INVALID)
	{
		std::cerr<<"RemoteClient::SendBlocks(): Not sending, no version."
				<<std::endl;
		return;
	}

	{
		JMutexAutoLock lock(m_blocks_sending_mutex);
		
		if(m_blocks_sending.size() >= MAX_SIMULTANEOUS_BLOCK_SENDS)
		{
			//std::cout<<"Not sending any blocks, Queue full."<<std::endl;
			return;
		}
	}

	Player *player = server->m_env.getPlayer(peer_id);

	v3f playerpos = player->getPosition();
	v3f playerspeed = player->getSpeed();

	v3s16 center_nodepos = Map::floatToInt(playerpos);

	v3s16 center = getNodeBlockPos(center_nodepos);

	/*
		Find out what block the player is going to next and set
		center to it.

		Don't react to speeds under the initial value of highest_speed
	*/
	/*f32 highest_speed = 0.1 * BS;
	v3s16 dir(0,0,0);
	if(abs(playerspeed.X) > highest_speed)
	{
		highest_speed = playerspeed.X;
		if(playerspeed.X > 0)
			dir = v3s16(1,0,0);
		else
			dir = v3s16(-1,0,0);
	}
	if(abs(playerspeed.Y) > highest_speed)
	{
		highest_speed = playerspeed.Y;
		if(playerspeed.Y > 0)
			dir = v3s16(0,1,0);
		else
			dir = v3s16(0,-1,0);
	}
	if(abs(playerspeed.Z) > highest_speed)
	{
		highest_speed = playerspeed.Z;
		if(playerspeed.Z > 0)
			dir = v3s16(0,0,1);
		else
			dir = v3s16(0,0,-1);
	}

	center += dir;*/
	
	/*
		Calculate the starting value of the block finder radius.

		The radius shall be the last used value minus the
		maximum moved distance.
	*/
	/*s16 d_start = m_last_block_find_d;
	if(max_moved >= d_start)
	{
		d_start = 0;
	}
	else
	{
		d_start -= max_moved;
	}*/
	
	s16 last_nearest_unsent_d;
	s16 d_start;
	{
		JMutexAutoLock lock(m_blocks_sent_mutex);
		if(m_last_center != center)
		{
			m_nearest_unsent_d = 0;
			m_last_center = center;
		}
		d_start = m_nearest_unsent_d;
		last_nearest_unsent_d = m_nearest_unsent_d;
	}

	// Serialization version used
	//u8 ser_version = serialization_version;

	//bool has_incomplete_blocks = false;
	
	/*
		TODO: Get this from somewhere
		TODO: Values more than 7 make placing and removing blocks very
		      sluggish when the map is being generated. This is
			  because d is looped every time from 0 to d_max if no
			  blocks are found for sending.
	*/
	//s16 d_max = 7;
	s16 d_max = 9;

	//TODO: Get this from somewhere (probably a bigger value)
	s16 d_max_gen = 4;
	
	//std::cout<<"Starting from "<<d_start<<std::endl;

	for(s16 d = d_start; d <= d_max; d++)
	{
		//std::cout<<"RemoteClient::SendBlocks(): d="<<d<<std::endl;
		
		//if(has_incomplete_blocks == false)
		{
			JMutexAutoLock lock(m_blocks_sent_mutex);
			/*
				If m_nearest_unsent_d was changed by the EmergeThread
				(it can change it to 0 through SetBlockNotSent),
				update our d to it.
				Else update m_nearest_unsent_d
			*/
			if(m_nearest_unsent_d != last_nearest_unsent_d)
			{
				d = m_nearest_unsent_d;
			}
			else
			{
				m_nearest_unsent_d = d;
			}
			last_nearest_unsent_d = m_nearest_unsent_d;
		}

		/*
			Get the border/face dot coordinates of a "d-radiused"
			box
		*/
		core::list<v3s16> list;
		getFacePositions(list, d);
		
		core::list<v3s16>::Iterator li;
		for(li=list.begin(); li!=list.end(); li++)
		{
			v3s16 p = *li + center;
			
			/*
				Send throttling
				- Don't allow too many simultaneous transfers

				Also, don't send blocks that are already flying.
			*/
			{
				JMutexAutoLock lock(m_blocks_sending_mutex);
				
				if(m_blocks_sending.size() >= MAX_SIMULTANEOUS_BLOCK_SENDS)
				{
					/*std::cout<<"Not sending more blocks. Queue full. "
							<<m_blocks_sending.size()
							<<std::endl;*/
					return;
				}

				if(m_blocks_sending.find(p) != NULL)
					continue;
			}
			
			/*
				Do not go over-limit
			*/
			if(p.X < -MAP_GENERATION_LIMIT / MAP_BLOCKSIZE
			|| p.X > MAP_GENERATION_LIMIT / MAP_BLOCKSIZE
			|| p.Y < -MAP_GENERATION_LIMIT / MAP_BLOCKSIZE
			|| p.Y > MAP_GENERATION_LIMIT / MAP_BLOCKSIZE
			|| p.Z < -MAP_GENERATION_LIMIT / MAP_BLOCKSIZE
			|| p.Z > MAP_GENERATION_LIMIT / MAP_BLOCKSIZE)
				continue;

			bool generate = d <= d_max_gen;
		
			// Limit the generating area vertically to half
			if(abs(p.Y - center.Y) > d_max_gen / 2)
				generate = false;
			
			/*
				Don't send already sent blocks
			*/
			{
				JMutexAutoLock lock(m_blocks_sent_mutex);
				
				if(m_blocks_sent.find(p) != NULL)
					continue;
			}
					
			/*
				Check if map has this block
			*/
			MapBlock *block = NULL;
			try
			{
				block = server->m_env.getMap().getBlockNoCreate(p);
			}
			catch(InvalidPositionException &e)
			{
			}
			
			bool surely_not_found_on_disk = false;
			if(block != NULL)
			{
				/*if(block->isIncomplete())
				{
					has_incomplete_blocks = true;
					continue;
				}*/

				if(block->isDummy())
				{
					surely_not_found_on_disk = true;
				}
			}

			/*
				If block has been marked to not exist on disk (dummy)
				and generating new ones is not wanted, skip block. TODO
			*/
			if(generate == false && surely_not_found_on_disk == true)
			{
				// get next one.
				continue;
			}

			/*
				Add inexistent block to emerge queue.
			*/
			if(block == NULL || surely_not_found_on_disk)
			{
				// Block not found.
				//TODO: Get value from somewhere
				//TODO: Balance between clients
				//if(server->m_emerge_queue.size() < 1)
				if(m_num_blocks_in_emerge_queue == 0)
				{
					// Add it to the emerge queue and trigger the thread
					
					u8 flags = 0;
					if(generate == false)
						flags |= TOSERVER_GETBLOCK_FLAG_OPTIONAL;
					
					{
						JMutexAutoLock nbeqlock(m_nbeq_mutex);
						m_num_blocks_in_emerge_queue++;
					}

					server->m_emerge_queue.addBlock(peer_id, p, flags);
					server->m_emergethread.trigger();
				}
				
				// get next one.
				continue;
			}

			/*
				Send block
			*/
			
			/*std::cout<<"RemoteClient::SendBlocks(): d="<<d<<", p="
				<<"("<<p.X<<","<<p.Y<<","<<p.Z<<")"
				<<" sending queue size: "<<m_blocks_sending.size()<<std::endl;*/

			server->SendBlockNoLock(peer_id, block, serialization_version);
			
			/*
				Add to history
			*/
			SentBlock(p);
		}
	}

	// Don't add anything here. The loop breaks by returning.
}

void RemoteClient::GotBlock(v3s16 p)
{
	JMutexAutoLock lock(m_blocks_sending_mutex);
	JMutexAutoLock lock2(m_blocks_sent_mutex);
	if(m_blocks_sending.find(p) != NULL)
		m_blocks_sending.remove(p);
	else
		std::cout<<"RemoteClient::GotBlock(): Didn't find in"
				" m_blocks_sending"<<std::endl;
	m_blocks_sent.insert(p, true);
}

void RemoteClient::SentBlock(v3s16 p)
{
	JMutexAutoLock lock(m_blocks_sending_mutex);
	if(m_blocks_sending.size() > 15)
	{
		std::cout<<"RemoteClient::SentBlock(): "
				<<"m_blocks_sending.size()="
				<<m_blocks_sending.size()<<std::endl;
	}
	if(m_blocks_sending.find(p) == NULL)
		m_blocks_sending.insert(p, 0.0);
	else
		std::cout<<"RemoteClient::SentBlock(): Sent block"
				" already in m_blocks_sending"<<std::endl;
}

void RemoteClient::SetBlockNotSent(v3s16 p)
{
	JMutexAutoLock sendinglock(m_blocks_sending_mutex);
	JMutexAutoLock sentlock(m_blocks_sent_mutex);

	m_nearest_unsent_d = 0;
	
	if(m_blocks_sending.find(p) != NULL)
		m_blocks_sending.remove(p);
	if(m_blocks_sent.find(p) != NULL)
		m_blocks_sent.remove(p);
}

void RemoteClient::SetBlocksNotSent(core::map<v3s16, MapBlock*> &blocks)
{
	JMutexAutoLock sendinglock(m_blocks_sending_mutex);
	JMutexAutoLock sentlock(m_blocks_sent_mutex);

	m_nearest_unsent_d = 0;
	
	for(core::map<v3s16, MapBlock*>::Iterator
			i = blocks.getIterator();
			i.atEnd()==false; i++)
	{
		v3s16 p = i.getNode()->getKey();

		if(m_blocks_sending.find(p) != NULL)
			m_blocks_sending.remove(p);
		if(m_blocks_sent.find(p) != NULL)
			m_blocks_sent.remove(p);
	}
}

void RemoteClient::BlockEmerged()
{
	JMutexAutoLock lock(m_nbeq_mutex);
	assert(m_num_blocks_in_emerge_queue > 0);
	m_num_blocks_in_emerge_queue--;
}

/*
	PlayerInfo
*/

PlayerInfo::PlayerInfo()
{
	name[0] = 0;
}

void PlayerInfo::PrintLine(std::ostream *s)
{
	(*s)<<id<<": \""<<name<<"\" ("
			<<position.X<<","<<position.Y
			<<","<<position.Z<<") ";
	address.print(s);
	(*s)<<" avg_rtt="<<avg_rtt;
	(*s)<<std::endl;
}

u32 PIChecksum(core::list<PlayerInfo> &l)
{
	core::list<PlayerInfo>::Iterator i;
	u32 checksum = 1;
	u32 a = 10;
	for(i=l.begin(); i!=l.end(); i++)
	{
		checksum += a * (i->id+1);
		checksum ^= 0x435aafcd;
		a *= 10;
	}
	return checksum;
}

/*
	Server
*/

Server::Server(std::string mapsavedir):
	m_env(new ServerMap(mapsavedir), dout_server),
	m_con(PROTOCOL_ID, 512, 20.0, this),
	m_thread(this),
	m_emergethread(this)
{
	m_env_mutex.Init();
	m_con_mutex.Init();
	m_step_dtime_mutex.Init();
	m_step_dtime = 0.0;
}

Server::~Server()
{
	// Stop threads
	stop();

	JMutexAutoLock clientslock(m_con_mutex);

	for(core::map<u16, RemoteClient*>::Iterator
		i = m_clients.getIterator();
		i.atEnd() == false; i++)
	{
		u16 peer_id = i.getNode()->getKey();

		// Delete player
		{
			JMutexAutoLock envlock(m_env_mutex);
			m_env.removePlayer(peer_id);
		}
		
		// Delete client
		delete i.getNode()->getValue();
	}
}

void Server::start(unsigned short port)
{
	// Stop thread if already running
	m_thread.stop();
	
	// Initialize connection
	m_con.setTimeoutMs(50);
	m_con.Serve(port);

	// Start thread
	m_thread.setRun(true);
	m_thread.Start();
	
	dout_server<<"Server started on port "<<port<<std::endl;
}

void Server::stop()
{
	// Stop threads (set run=false first so both start stopping)
	m_thread.setRun(false);
	m_emergethread.setRun(false);
	m_thread.stop();
	m_emergethread.stop();
	
	dout_server<<"Server threads stopped"<<std::endl;
}

void Server::step(float dtime)
{
	{
		JMutexAutoLock lock(m_step_dtime_mutex);
		m_step_dtime += dtime;
	}
}

void Server::AsyncRunStep()
{
	float dtime;
	{
		JMutexAutoLock lock1(m_step_dtime_mutex);
		dtime = m_step_dtime;
		m_step_dtime = 0.0;
	}
	{
		// Process connection's timeouts
		JMutexAutoLock lock2(m_con_mutex);
		m_con.RunTimeouts(dtime);
	}
	{
		// Step environment
		// This also runs Map's timers
		JMutexAutoLock lock(m_env_mutex);
		m_env.step(dtime);
	}
	
	/*
		Do background stuff
	*/
	
	// Periodically print some info
	{
		static float counter = 0.0;
		counter += dtime;
		if(counter >= 30.0)
		{
			counter = 0.0;

			JMutexAutoLock lock2(m_con_mutex);

			for(core::map<u16, RemoteClient*>::Iterator
				i = m_clients.getIterator();
				i.atEnd() == false; i++)
			{
				//u16 peer_id = i.getNode()->getKey();
				RemoteClient *client = i.getNode()->getValue();
				client->PrintInfo(std::cout);
			}
		}
	}

	// Send blocks to clients
	SendBlocks(dtime);
	
	{
		static float counter = 0.0;
		counter += dtime;
		if(counter >= 0.2)
		{
			counter = 0.0;

			SendPlayerPositions();
		}
	}
	{
		static float counter = 0.0;
		counter += dtime;
		if(counter >= 10.0)
		{
			counter = 0.0;

			SendPlayerInfos();
		}
	}

	{
		// Save map
		static float counter = 0.0;
		counter += dtime;
		if(counter >= SERVER_MAP_SAVE_INTERVAL)
		{
			counter = 0.0;

			JMutexAutoLock lock(m_env_mutex);
			// Save only changed parts
			m_env.getMap().save(true);
		}
	}
}

void Server::Receive()
{
	u32 data_maxsize = 10000;
	Buffer<u8> data(data_maxsize);
	u16 peer_id;
	u32 datasize;
	try{
		{
			JMutexAutoLock lock(m_con_mutex);
			datasize = m_con.Receive(peer_id, *data, data_maxsize);
		}
		ProcessData(*data, datasize, peer_id);
	}
	catch(con::InvalidIncomingDataException &e)
	{
		derr_server<<"Server::Receive(): "
				"InvalidIncomingDataException: what()="
				<<e.what()<<std::endl;
	}
	catch(con::PeerNotFoundException &e)
	{
		// The peer has been disconnected.
		// Find the associated player and remove it.

		JMutexAutoLock envlock(m_env_mutex);

		dout_server<<"ServerThread: peer_id="<<peer_id
				<<" has apparently closed connection. "
				<<"Removing player."<<std::endl;

		m_env.removePlayer(peer_id);
	}
}

void Server::ProcessData(u8 *data, u32 datasize, u16 peer_id)
{
	// Let free access to the environment and the connection.
	// Blocking them doesn't matter too much because this is the
	// server and latency is not such a problem.
	JMutexAutoLock envlock(m_env_mutex);
	JMutexAutoLock conlock(m_con_mutex);
	
	con::Peer *peer;
	try{
		peer = m_con.GetPeer(peer_id);
	}
	catch(con::PeerNotFoundException &e)
	{
		derr_server<<DTIME<<"Server::ProcessData(): Cancelling: peer "
				<<peer_id<<" not found"<<std::endl;
		return;
	}
	
	//u8 peer_ser_ver = peer->serialization_version;
	u8 peer_ser_ver = getClient(peer->id)->serialization_version;

	try
	{

	if(datasize < 2)
		return;

	ToServerCommand command = (ToServerCommand)readU16(&data[0]);
	
	if(command == TOSERVER_INIT)
	{
		// [0] u16 TOSERVER_INIT
		// [2] u8 SER_FMT_VER_HIGHEST
		// [3] u8[20] player_name

		if(datasize < 3)
			return;

		// First byte after command is maximum supported
		// serialization version
		u8 client_max = data[2];
		u8 our_max = SER_FMT_VER_HIGHEST;
		// Use the highest version supported by both
		u8 deployed = core::min_(client_max, our_max);
		// If it's lower than the lowest supported, give up.
		if(deployed < SER_FMT_VER_LOWEST)
			deployed = SER_FMT_VER_INVALID;

		//peer->serialization_version = deployed;
		getClient(peer->id)->serialization_version = deployed;

		/*
			Set up player
		*/

		Player *player = m_env.getPlayer(peer_id);

		// Check if player doesn't exist
		if(player == NULL)
			throw con::InvalidIncomingDataException
				("Server::ProcessData(): INIT: Player doesn't exist");

		// update name if it was supplied
		if(datasize >= 20+3)
		{
			data[20+3-1] = 0;
			player->updateName((const char*)&data[3]);
		}

		// Now answer with a TOCLIENT_INIT
		
		SharedBuffer<u8> reply(3);
		writeU16(&reply[0], TOCLIENT_INIT);
		writeU8(&reply[2], deployed);
		// Send as reliable
		m_con.Send(peer_id, 0, reply, true);

		return;
	}


	if(peer_ser_ver == SER_FMT_VER_INVALID)
	{
		derr_server<<DTIME<<"Server::ProcessData(): Cancelling: Peer"
				" serialization format invalid or not initialized."
				" Skipping incoming command="<<command<<std::endl;
		return;
	}
	
	Player *player = m_env.getPlayer(peer_id);

	if(player == NULL){
		derr_server<<"Server::ProcessData(): Cancelling: "
				"No player for peer_id="<<peer_id
				<<std::endl;
		return;
	}

	if(command == TOSERVER_GETBLOCK)
	{
		// Check for too short data
		if(datasize < 8)
			return;
		/*
			The client is asking a MapBlock.
			Get block data and send it.
			If not found, generate if not optional.
		*/

		/*
			[0] u16 command
			[2] u16 X
			[4] u16 Y
			[6] u16 Z
			[8] u8 flags (old versions leave this out)
		*/

		v3s16 p;
		p.X = readS16(&data[2]);
		p.Y = readS16(&data[4]);
		p.Z = readS16(&data[6]);
		
		u8 flags = 0;
		if(datasize >= 9)
			flags = data[8];

		/*dout_server<<"Server::ProcessData(): GETBLOCK"
				<<std::endl;*/
		/*std::cout<<"Server::ProcessData(): GETBLOCK"
				<<"("<<p.X<<","<<p.Y<<","<<p.Z<<")"<<std::endl;*/

		MapBlock *block;
		try{
			block = m_env.getMap().getBlockNoCreate(p);
		}
		catch(InvalidPositionException &e)
		{
			std::cerr<<"WARNING: GETBLOCK triggered emergethread disbled"
					<<std::endl;
#if 0
			// Block not found.
			// Add it to the emerge queue, trigger the thread and quit.
			m_emerge_queue.addBlock(peer_id, p, flags);

			/*derr_server<<"Queued block "
					"("<<p.X<<","<<p.Y<<","<<p.Z<<")."
					" triggering thread."
					<<std::endl;*/

			m_emergethread.trigger();
#endif
			return;
		}

		SharedBuffer<u8> blockdata = block->serialize(peer_ser_ver);

		/*std::cout<<"Block serialized: size="<<blockdata.getSize()<<std::endl;
		for(u32 i=0; i<blockdata.getSize(); i++)
		{
			printf("%x ", (unsigned int)blockdata[i] & 0xff);
		}
		std::cout<<std::endl;*/

		u32 replysize = 8 + blockdata.getSize();

		//std::cout<<"replysize="<<replysize<<std::endl;

		SharedBuffer<u8> reply(replysize);
		writeU16(&reply[0], TOCLIENT_BLOCKDATA);
		writeS16(&reply[2], p.X);
		writeS16(&reply[4], p.Y);
		writeS16(&reply[6], p.Z);
		memcpy(&reply[8], *blockdata, blockdata.getSize());

		// Send as reliable
		m_con.Send(peer_id, 1, reply, true);
	}
	else if(command == TOSERVER_REMOVENODE)
	{
		if(datasize < 8)
			return;
			
		v3s16 p;
		p.X = readS16(&data[2]);
		p.Y = readS16(&data[4]);
		p.Z = readS16(&data[6]);

		core::map<v3s16, MapBlock*> modified_blocks;

		try{
			m_env.getMap().removeNodeAndUpdate(p, modified_blocks);
		}
		catch(InvalidPositionException &e)
		{
			derr_server<<"Server: Ignoring REMOVENODE: Node not found"
					<<std::endl;
			return;
		}

		u32 replysize = 8;
		SharedBuffer<u8> reply(replysize);
		writeU16(&reply[0], TOCLIENT_REMOVENODE);
		writeS16(&reply[2], p.X);
		writeS16(&reply[4], p.Y);
		writeS16(&reply[6], p.Z);
		// Send as reliable
		m_con.SendToAll(0, reply, true);
	}
	else if(command == TOSERVER_ADDNODE)
	{
		if(datasize < 8 + MapNode::serializedLength(peer_ser_ver))
			return;

		v3s16 p;
		p.X = readS16(&data[2]);
		p.Y = readS16(&data[4]);
		p.Z = readS16(&data[6]);

		core::map<v3s16, MapBlock*> modified_blocks;

		MapNode n;
		n.deSerialize(&data[8], peer_ser_ver);
		try{
			u8 light = m_env.getMap().getNode(p).getLight();
			m_env.getMap().setNode(p, n);
			m_env.getMap().nodeAddedUpdate(p, light, modified_blocks);
		}
		catch(InvalidPositionException &e)
		{
			derr_server<<"Server: Ignoring ADDNODE: Node not found"
					<<std::endl;
			return;
		}

		u32 replysize = 8 + MapNode::serializedLength(peer_ser_ver);
		SharedBuffer<u8> reply(replysize);
		writeU16(&reply[0], TOCLIENT_ADDNODE);
		writeS16(&reply[2], p.X);
		writeS16(&reply[4], p.Y);
		writeS16(&reply[6], p.Z);
		n.serialize(&reply[8], peer_ser_ver);
		// Send as reliable
		m_con.SendToAll(0, reply, true);
	}
	else if(command == TOSERVER_PLAYERPOS)
	{
		if(datasize < 2+12+12)
			return;

		v3s32 ps = readV3S32(&data[2]);
		v3s32 ss = readV3S32(&data[2+12]);
		
		v3f position((f32)ps.X/100., (f32)ps.Y/100., (f32)ps.Z/100.);
		v3f speed((f32)ss.X/100., (f32)ss.Y/100., (f32)ss.Z/100.);
		player->setPosition(position);
		player->speed = speed;

		/*dout_server<<"Server::ProcessData(): Moved player "<<peer_id
				<<" to ("<<position.X
				<<","<<position.Y
				<<","<<position.Z
				<<")"<<std::endl;*/
	}
	else if(command == TOSERVER_GOTBLOCKS)
	{
		if(datasize < 2+1)
			return;
		
		/*
			[0] u16 command
			[2] u8 count
			[3] v3s16 pos_0
			[3+6] v3s16 pos_1
			...
		*/

		u16 count = data[2];
		for(u16 i=0; i<count; i++)
		{
			if((s16)datasize < 2+1+(i+1)*6)
				throw con::InvalidIncomingDataException
					("GOTBLOCKS length is too short");
			v3s16 p = readV3S16(&data[2+1+i*6]);
			/*std::cout<<"Server: GOTBLOCKS ("
					<<p.X<<","<<p.Y<<","<<p.Z<<")"<<std::endl;*/
			RemoteClient *client = getClient(peer_id);
			client->GotBlock(p);
		}
	}
	/*else if(command == TOSERVER_PLAYERINFO)
	{
		if(datasize < 2+PLAYERNAME_SIZE)
			return;

		player->updateName((char*)&data[2]);
	}*/
	else
	{
		derr_server<<"WARNING: Server::ProcessData(): Ingoring "
				"unknown command "<<command<<std::endl;
	}
	
	} //try
	catch(SendFailedException &e)
	{
		derr_server<<"Server::ProcessData(): SendFailedException: "
				<<"what="<<e.what()
				<<std::endl;
	}
}

/*void Server::Send(u16 peer_id, u16 channelnum,
		SharedBuffer<u8> data, bool reliable)
{
	JMutexAutoLock lock(m_con_mutex);
	m_con.Send(peer_id, channelnum, data, reliable);
}*/

void Server::SendBlockNoLock(u16 peer_id, MapBlock *block, u8 ver)
{
	/*
		Create a packet with the block in the right format
	*/
	
	SharedBuffer<u8> blockdata = block->serialize(ver);
	u32 replysize = 8 + blockdata.getSize();
	SharedBuffer<u8> reply(replysize);
	v3s16 p = block->getPos();
	writeU16(&reply[0], TOCLIENT_BLOCKDATA);
	writeS16(&reply[2], p.X);
	writeS16(&reply[4], p.Y);
	writeS16(&reply[6], p.Z);
	memcpy(&reply[8], *blockdata, blockdata.getSize());
	
	/*
		Send packet
	*/
	m_con.Send(peer_id, 1, reply, true);
}

void Server::SendBlock(u16 peer_id, MapBlock *block, u8 ver)
{
	JMutexAutoLock conlock(m_con_mutex);
	
	SendBlockNoLock(peer_id, block, ver);
}

core::list<PlayerInfo> Server::getPlayerInfo()
{
	JMutexAutoLock envlock(m_env_mutex);
	JMutexAutoLock conlock(m_con_mutex);
	
	core::list<PlayerInfo> list;

	core::list<Player*> players = m_env.getPlayers();
	
	core::list<Player*>::Iterator i;
	for(i = players.begin();
			i != players.end(); i++)
	{
		PlayerInfo info;

		Player *player = *i;
		try{
			con::Peer *peer = m_con.GetPeer(player->peer_id);
			info.id = peer->id;
			info.address = peer->address;
			info.avg_rtt = peer->avg_rtt;
		}
		catch(con::PeerNotFoundException &e)
		{
			// Outdated peer info
			info.id = 0;
			info.address = Address(0,0,0,0,0);
			info.avg_rtt = 0.0;
		}

		snprintf(info.name, PLAYERNAME_SIZE, "%s", player->getName());
		info.position = player->getPosition();

		list.push_back(info);
	}

	return list;
}

void Server::peerAdded(con::Peer *peer)
{
	dout_server<<"Server::peerAdded(): peer->id="
			<<peer->id<<std::endl;
	
	// Connection is already locked when this is called.
	//JMutexAutoLock lock(m_con_mutex);
	
	// Error check
	core::map<u16, RemoteClient*>::Node *n;
	n = m_clients.find(peer->id);
	if(n != NULL)
		throw ProgrammingError
			("Server::peerAdded(): Client already exists");

	// Create client
	RemoteClient *client = new RemoteClient();
	client->peer_id = peer->id;
	m_clients.insert(client->peer_id, client);

	// Create player
	{
		JMutexAutoLock envlock(m_env_mutex);
		
		Player *player = m_env.getPlayer(peer->id);

		if(player != NULL)
			throw ProgrammingError
				("Server::peerAdded(): Player already exists");

		player = new Player(false);
		player->peer_id = peer->id;
		m_env.addPlayer(player);
	}
}

void Server::deletingPeer(con::Peer *peer, bool timeout)
{
	dout_server<<"Server::deletingPeer(): peer->id="
			<<peer->id<<", timeout="<<timeout<<std::endl;
	
	// Connection is already locked when this is called.
	//JMutexAutoLock lock(m_con_mutex);

	// Error check
	core::map<u16, RemoteClient*>::Node *n;
	n = m_clients.find(peer->id);
	if(n == NULL)
		throw ProgrammingError
			("Server::deletingPeer(): No such client");
	
	// Delete player
	{
		JMutexAutoLock envlock(m_env_mutex);
		m_env.removePlayer(peer->id);
	}
	
	// Delete client
	delete m_clients[peer->id];
	m_clients.remove(peer->id);
}

void Server::SendPlayerPositions()
{
	JMutexAutoLock envlock(m_env_mutex);
	
	core::list<Player*> players = m_env.getPlayers();
	
	u32 player_count = players.getSize();
	u32 datasize = 2+(2+12+12)*player_count;

	SharedBuffer<u8> data(datasize);
	writeU16(&data[0], TOCLIENT_PLAYERPOS);
	
	u32 start = 2;
	core::list<Player*>::Iterator i;
	for(i = players.begin();
			i != players.end(); i++)
	{
		Player *player = *i;
		
		v3f pf = player->getPosition();
		v3s32 position(pf.X*100, pf.Y*100, pf.Z*100);
		v3f sf = player->speed;
		v3s32 speed(sf.X*100, sf.Y*100, sf.Z*100);

		writeU16(&data[start], player->peer_id);
		writeV3S32(&data[start+2], position);
		writeV3S32(&data[start+2+12], speed);
		start += 2+12+12;
	}

	JMutexAutoLock conlock(m_con_mutex);

	// Send as unreliable
	m_con.SendToAll(0, data, false);
}

void Server::SendPlayerInfos()
{
	JMutexAutoLock envlock(m_env_mutex);
	
	core::list<Player*> players = m_env.getPlayers();
	
	u32 player_count = players.getSize();
	u32 datasize = 2+(2+PLAYERNAME_SIZE)*player_count;

	SharedBuffer<u8> data(datasize);
	writeU16(&data[0], TOCLIENT_PLAYERINFO);
	
	u32 start = 2;
	core::list<Player*>::Iterator i;
	for(i = players.begin();
			i != players.end(); i++)
	{
		Player *player = *i;
		
		writeU16(&data[start], player->peer_id);
		snprintf((char*)&data[start+2], PLAYERNAME_SIZE, "%s", player->getName());
		start += 2+PLAYERNAME_SIZE;
	}

	JMutexAutoLock conlock(m_con_mutex);

	// Send as reliable
	m_con.SendToAll(0, data, true);
}

void Server::SendBlocks(float dtime)
{
	//std::cout<<"Server::SendBlocks(): BEGIN"<<std::endl;

	JMutexAutoLock conlock(m_con_mutex);
	JMutexAutoLock envlock(m_env_mutex);

	for(core::map<u16, RemoteClient*>::Iterator
		i = m_clients.getIterator();
		i.atEnd() == false; i++)
	{
		RemoteClient *client = i.getNode()->getValue();
		assert(client->peer_id == i.getNode()->getKey());
		
		if(client->serialization_version == SER_FMT_VER_INVALID)
			continue;

		//std::cout<<"Server::SendBlocks(): sending blocks for client "<<client->peer_id<<std::endl;
		
		//u16 peer_id = client->peer_id;
		client->SendBlocks(this, dtime);
	}

	//std::cout<<"Server::SendBlocks(): END"<<std::endl;
}

RemoteClient* Server::getClient(u16 peer_id)
{
	//JMutexAutoLock lock(m_con_mutex);
	core::map<u16, RemoteClient*>::Node *n;
	n = m_clients.find(peer_id);
	if(n == NULL)
		throw ProgrammingError
			("Server::getClient(): No such client");
	return n->getValue();
}


