#include "client.h"
#include "utility.h"
#include <iostream>
#include "clientserver.h"
#include "jmutexautolock.h"
#include "main.h"
#include <sstream>

#ifdef _WIN32
	#include <windows.h>
	#define sleep_ms(x) Sleep(x)
#else
	#include <unistd.h>
	#define sleep_ms(x) usleep(x*1000)
#endif

/*
	FIXME: This thread can access the environment at any time
*/

void * ClientUpdateThread::Thread()
{
	ThreadStarted();

	DSTACK(__FUNCTION_NAME);

	try
	{
	
		while(getRun())
		{
			m_client->asyncStep();

			bool was = m_client->AsyncProcessData();

			if(was == false)
				sleep_ms(200);
		}

	}
	/*
		This is what has to be done in threads to get suitable debug info
	*/
	catch(std::exception &e)
	{
		dstream<<std::endl<<DTIME<<"An unhandled exception occurred: "
				<<e.what()<<std::endl;
		assert(0);
	}

	return NULL;
}

Client::Client(IrrlichtDevice *device, video::SMaterial *materials):
	m_thread(this),
	m_env(new ClientMap(this, materials,
			device->getSceneManager()->getRootSceneNode(),
			device->getSceneManager(), 666),
			dout_client),
	m_con(PROTOCOL_ID, 512, 20.0, this),
	m_device(device),
	camera_position(0,0,0),
	camera_direction(0,0,1),
	m_server_ser_ver(SER_FMT_VER_INVALID),
	m_step_dtime(0.0)
{
	//m_fetchblock_mutex.Init();
	m_incoming_queue_mutex.Init();
	m_env_mutex.Init();
	m_con_mutex.Init();
	m_step_dtime_mutex.Init();

	m_thread.Start();
	
	{
		JMutexAutoLock envlock(m_env_mutex);
		//m_env.getMap().StartUpdater();

		Player *player = new LocalPlayer();

		//f32 y = BS*(2 + m_env.getMap().getGroundHeight(v2s16(0,0)));
		f32 y = BS*2 + BS*20;
		player->setPosition(v3f(0, y, 0));
		//player->setPosition(v3f(0, y, 30900*BS)); // DEBUG
		m_env.addPlayer(player);
	}
}

Client::~Client()
{
	m_thread.setRun(false);
	while(m_thread.IsRunning())
		sleep_ms(100);
}

void Client::connect(Address address)
{
	DSTACK(__FUNCTION_NAME);
	JMutexAutoLock lock(m_con_mutex);
	m_con.setTimeoutMs(0);
	m_con.Connect(address);
}

bool Client::connectedAndInitialized()
{
	JMutexAutoLock lock(m_con_mutex);

	if(m_con.Connected() == false)
		return false;
	
	if(m_server_ser_ver == SER_FMT_VER_INVALID)
		return false;
	
	return true;
}

void Client::step(float dtime)
{
	DSTACK(__FUNCTION_NAME);
	// Limit a bit
	if(dtime > 2.0)
		dtime = 2.0;
	
	{
		//TimeTaker timer("ReceiveAll()", m_device);
		ReceiveAll();
	}
	
	{
		JMutexAutoLock lock(m_con_mutex);
		m_con.RunTimeouts(dtime);
	}

	{
		static float counter = -0.001;
		counter -= dtime;
		if(counter <= 0.0)
		{
			counter = 10.0;

			JMutexAutoLock lock(m_env_mutex);

			core::list<v3s16> deleted_blocks;
	
			// Delete sector blocks
			//TODO: Get timeout from somewhere
			u32 num = m_env.getMap().deleteUnusedSectors
					(300, true, &deleted_blocks);

			if(num > 0)
			{
				dstream<<DTIME<<"Client: Deleted blocks of "<<num
						<<" unused sectors"<<std::endl;
				
				/*
					Send info to server
				*/

				// Env is locked so con can be locked.
				JMutexAutoLock lock(m_con_mutex);
				
				core::list<v3s16>::Iterator i = deleted_blocks.begin();
				core::list<v3s16> sendlist;
				for(;;)
				{
					if(sendlist.size() == 255 || i == deleted_blocks.end())
					{
						if(sendlist.size() == 0)
							break;
						/*
							[0] u16 command
							[2] u8 count
							[3] v3s16 pos_0
							[3+6] v3s16 pos_1
							...
						*/
						u32 replysize = 2+1+6*sendlist.size();
						SharedBuffer<u8> reply(replysize);
						writeU16(&reply[0], TOSERVER_DELETEDBLOCKS);
						reply[2] = sendlist.size();
						u32 k = 0;
						for(core::list<v3s16>::Iterator
								j = sendlist.begin();
								j != sendlist.end(); j++)
						{
							writeV3S16(&reply[2+1+6*k], *j);
							k++;
						}
						m_con.Send(PEER_ID_SERVER, 1, reply, true);

						if(i == deleted_blocks.end())
							break;

						sendlist.clear();
					}

					sendlist.push_back(*i);
					i++;
				}
			}
		}
	}

	bool connected = connectedAndInitialized();

	if(connected == false)
	{
		static float counter = -0.001;
		counter -= dtime;
		if(counter <= 0.0)
		{
			counter = 2.0;

			JMutexAutoLock envlock(m_env_mutex);
			
			Player *myplayer = m_env.getLocalPlayer();
			assert(myplayer != NULL);
	
			// Send TOSERVER_INIT
			// [0] u16 TOSERVER_INIT
			// [2] u8 SER_FMT_VER_HIGHEST
			// [3] u8[20] player_name
			SharedBuffer<u8> data(2+1+20);
			writeU16(&data[0], TOSERVER_INIT);
			writeU8(&data[2], SER_FMT_VER_HIGHEST);
			memcpy(&data[3], myplayer->getName(), 20);
			// Send as unreliable
			Send(0, data, false);
		}

		// Not connected, return
		return;
	}

	/*
		Do stuff if connected
	*/
	
	{
		JMutexAutoLock lock(m_env_mutex);
		m_env.step(dtime);
	}

	{
		// Fetch some nearby blocks
		fetchBlocks();
	}

	{
		static float counter = 0.0;
		counter += dtime;
		if(counter >= 10)
		{
			counter = 0.0;
			JMutexAutoLock lock(m_con_mutex);
			// connectedAndInitialized() is true, peer exists.
			con::Peer *peer = m_con.GetPeer(PEER_ID_SERVER);
			dstream<<DTIME<<"Client: avg_rtt="<<peer->avg_rtt<<std::endl;
		}
	}
	{
		// Update at reasonable intervals (0.2s)
		static float counter = 0.0;
		counter += dtime;
		if(counter >= 0.2)
		{
			counter = 0.0;
			sendPlayerPos();
		}
	}

#if 0
	/*
		Clear old entries from fetchblock history
	*/
	{
		JMutexAutoLock lock(m_fetchblock_mutex);
		
		core::list<v3s16> remove_queue;
		core::map<v3s16, float>::Iterator i;
		i = m_fetchblock_history.getIterator();
		for(; i.atEnd() == false; i++)
		{
			float value = i.getNode()->getValue();
			value += dtime;
			i.getNode()->setValue(value);
			if(value >= 60.0)
				remove_queue.push_back(i.getNode()->getKey());
		}
		core::list<v3s16>::Iterator j;
		j = remove_queue.begin();
		for(; j != remove_queue.end(); j++)
		{
			m_fetchblock_history.remove(*j);
		}
	}
#endif

	{
		JMutexAutoLock lock(m_step_dtime_mutex);
		m_step_dtime += dtime;
	}
	
	/*
		BEGIN TEST CODE
	*/

	/*
		END OF TEST CODE
	*/
}

float Client::asyncStep()
{
	DSTACK(__FUNCTION_NAME);
	//dstream<<"Client::asyncStep()"<<std::endl;
	
	float dtime;
	{
		JMutexAutoLock lock1(m_step_dtime_mutex);
		dtime = m_step_dtime;
		m_step_dtime = 0.0;
	}

#if 0
	{
		static float counter = 0.0;
		counter += dtime;
		if(counter >= 5.0)
		{
			counter = 0.0;

			/*
				Update heightmap mesh
			*/
			TimeTaker timer1("updateMesh", m_device);
			((ClientMap&)m_env.getMap()).updateMesh();
		}
	}
#endif

	return dtime;
}

// Virtual methods from con::PeerHandler
void Client::peerAdded(con::Peer *peer)
{
	derr_client<<"Client::peerAdded(): peer->id="
			<<peer->id<<std::endl;
}
void Client::deletingPeer(con::Peer *peer, bool timeout)
{
	derr_client<<"Client::deletingPeer(): "
			"Server Peer is getting deleted "
			<<"(timeout="<<timeout<<")"<<std::endl;
}

void Client::ReceiveAll()
{
	DSTACK(__FUNCTION_NAME);
	for(;;)
	{
		try{
			Receive();
		}
		catch(con::NoIncomingDataException &e)
		{
			break;
		}
		catch(con::InvalidIncomingDataException &e)
		{
			dout_client<<DTIME<<"Client::ReceiveAll(): "
					"InvalidIncomingDataException: what()="
					<<e.what()<<std::endl;
		}
		//TODO: Testing
		//break;
	}
}

void Client::Receive()
{
	DSTACK(__FUNCTION_NAME);
	u32 data_maxsize = 10000;
	Buffer<u8> data(data_maxsize);
	u16 sender_peer_id;
	u32 datasize;
	{
		JMutexAutoLock lock(m_con_mutex);
		datasize = m_con.Receive(sender_peer_id, *data, data_maxsize);
	}
	ProcessData(*data, datasize, sender_peer_id);
}

/*
	sender_peer_id given to this shall be quaranteed to be a valid peer
*/
void Client::ProcessData(u8 *data, u32 datasize, u16 sender_peer_id)
{
	DSTACK(__FUNCTION_NAME);
	// Ignore packets that don't even fit a command
	if(datasize < 2)
		return;

	ToClientCommand command = (ToClientCommand)readU16(&data[0]);
	
	/*
		If this check is removed, be sure to change the queue
		system to know the ids
	*/
	if(sender_peer_id != PEER_ID_SERVER)
	{
		dout_client<<DTIME<<"Client::ProcessData(): Discarding data not "
				"coming from server: peer_id="<<sender_peer_id
				<<std::endl;
		return;
	}

	con::Peer *peer;
	{
		JMutexAutoLock lock(m_con_mutex);
		// All data is coming from the server
		// PeerNotFoundException is handled by caller.
		peer = m_con.GetPeer(PEER_ID_SERVER);
	}

	u8 ser_version = m_server_ser_ver;

	// Execute fast commands straight away

	if(command == TOCLIENT_INIT)
	{
		if(datasize < 3)
			return;

		u8 deployed = data[2];
		
		dout_client<<DTIME<<"Client: TOCLIENT_INIT received with "
				"deployed="<<((int)deployed&0xff)<<std::endl;

		/*assert(deployed >= SER_FMT_VER_LOWEST);
		assert(deployed <= SER_FMT_VER_HIGHEST);*/

		if(deployed < SER_FMT_VER_LOWEST
				|| deployed > SER_FMT_VER_HIGHEST)
		{
			derr_client<<DTIME<<"Client: TOCLIENT_INIT: Server sent "
					<<"unsupported ser_fmt_ver"<<std::endl;
			return;
		}
		
		m_server_ser_ver = deployed;

		u32 replysize = 2;
		SharedBuffer<u8> reply(replysize);
		writeU16(&reply[0], TOSERVER_INIT2);
		// Send as reliable
		m_con.Send(PEER_ID_SERVER, 1, reply, true);

		return;
	}
	
	if(ser_version == SER_FMT_VER_INVALID)
	{
		dout_client<<DTIME<<"WARNING: Client: Server serialization"
				" format invalid or not initialized."
				" Skipping incoming command="<<command<<std::endl;
		return;
	}
	
	if(command == TOCLIENT_REMOVENODE)
	{
		if(datasize < 8)
			return;
		v3s16 p;
		p.X = readS16(&data[2]);
		p.Y = readS16(&data[4]);
		p.Z = readS16(&data[6]);
		
		core::map<v3s16, MapBlock*> modified_blocks;

		try
		{
			JMutexAutoLock envlock(m_env_mutex);
			//TimeTaker t("removeNodeAndUpdate", m_device);
			m_env.getMap().removeNodeAndUpdate(p, modified_blocks);
		}
		catch(InvalidPositionException &e)
		{
		}
		
		for(core::map<v3s16, MapBlock * >::Iterator
				i = modified_blocks.getIterator();
				i.atEnd() == false; i++)
		{
			v3s16 p = i.getNode()->getKey();
			m_env.getMap().updateMeshes(p);
		}
	}
	else if(command == TOCLIENT_ADDNODE)
	{
		if(datasize < 8)
			return;

		v3s16 p;
		p.X = readS16(&data[2]);
		p.Y = readS16(&data[4]);
		p.Z = readS16(&data[6]);
		
		if(datasize < 8 + MapNode::serializedLength(ser_version))
			return;

		MapNode n;
		n.deSerialize(&data[8], ser_version);
		
		core::map<v3s16, MapBlock*> modified_blocks;

		try
		{
			JMutexAutoLock envlock(m_env_mutex);
			/*u8 light = m_env.getMap().getNode(p).getLight();
			m_env.getMap().setNode(p, n);
			m_env.getMap().nodeAddedUpdate(p, light, modified_blocks);*/
			m_env.getMap().addNodeAndUpdate(p, n, modified_blocks);
		}
		catch(InvalidPositionException &e)
		{}
		
		for(core::map<v3s16, MapBlock * >::Iterator
				i = modified_blocks.getIterator();
				i.atEnd() == false; i++)
		{
			v3s16 p = i.getNode()->getKey();
			m_env.getMap().updateMeshes(p);
		}
	}
	else if(command == TOCLIENT_PLAYERPOS)
	{
		u16 our_peer_id;
		{
			JMutexAutoLock lock(m_con_mutex);
			our_peer_id = m_con.GetPeerID();
		}
		// Cancel if we don't have a peer id
		if(our_peer_id == PEER_ID_NEW){
			dout_client<<DTIME<<"TOCLIENT_PLAYERPOS cancelled: "
					"we have no peer id"
					<<std::endl;
			return;
		}

		{ //envlock
			JMutexAutoLock envlock(m_env_mutex);
			
			u32 player_size = 2+12+12+4+4;
				
			u32 player_count = (datasize-2) / player_size;
			u32 start = 2;
			for(u32 i=0; i<player_count; i++)
			{
				u16 peer_id = readU16(&data[start]);

				// Don't update the position of the local player
				if(peer_id == our_peer_id)
				{
					start += player_size;
					continue;
				}

				Player *player = m_env.getPlayer(peer_id);

				// Create a player if it doesn't exist
				if(player == NULL)
				{
					player = new RemotePlayer(
							m_device->getSceneManager()->getRootSceneNode(),
							m_device,
							peer_id);
					player->peer_id = peer_id;
					m_env.addPlayer(player);
					dout_client<<DTIME<<"Client: Adding new player "
							<<peer_id<<std::endl;
				}

				v3s32 ps = readV3S32(&data[start+2]);
				v3s32 ss = readV3S32(&data[start+2+12]);
				s32 pitch_i = readS32(&data[start+2+12+12]);
				s32 yaw_i = readS32(&data[start+2+12+12+4]);
				/*dstream<<"Client: got "
						<<"pitch_i="<<pitch_i
						<<" yaw_i="<<yaw_i<<std::endl;*/
				f32 pitch = (f32)pitch_i / 100.0;
				f32 yaw = (f32)yaw_i / 100.0;
				v3f position((f32)ps.X/100., (f32)ps.Y/100., (f32)ps.Z/100.);
				v3f speed((f32)ss.X/100., (f32)ss.Y/100., (f32)ss.Z/100.);
				player->setPosition(position);
				player->setSpeed(speed);
				player->setPitch(pitch);
				player->setYaw(yaw);

				/*dstream<<"Client: player "<<peer_id
						<<" pitch="<<pitch
						<<" yaw="<<yaw<<std::endl;*/

				start += player_size;
			}
		} //envlock
	}
	else if(command == TOCLIENT_PLAYERINFO)
	{
		u16 our_peer_id;
		{
			JMutexAutoLock lock(m_con_mutex);
			our_peer_id = m_con.GetPeerID();
		}
		// Cancel if we don't have a peer id
		if(our_peer_id == PEER_ID_NEW){
			dout_client<<DTIME<<"TOCLIENT_PLAYERINFO cancelled: "
					"we have no peer id"
					<<std::endl;
			return;
		}
		
		//dstream<<DTIME<<"Client: Server reports players:"<<std::endl;

		{ //envlock
			JMutexAutoLock envlock(m_env_mutex);
			
			u32 item_size = 2+PLAYERNAME_SIZE;
			u32 player_count = (datasize-2) / item_size;
			u32 start = 2;
			// peer_ids
			core::list<u16> players_alive;
			for(u32 i=0; i<player_count; i++)
			{
				// Make sure the name ends in '\0'
				data[start+2+PLAYERNAME_SIZE-1] = 0;

				u16 peer_id = readU16(&data[start]);

				players_alive.push_back(peer_id);
				
				/*dstream<<DTIME<<"peer_id="<<peer_id
						<<" name="<<((char*)&data[start+2])<<std::endl;*/

				// Don't update the info of the local player
				if(peer_id == our_peer_id)
				{
					start += item_size;
					continue;
				}

				Player *player = m_env.getPlayer(peer_id);

				// Create a player if it doesn't exist
				if(player == NULL)
				{
					player = new RemotePlayer(
							m_device->getSceneManager()->getRootSceneNode(),
							m_device,
							-1);
					player->peer_id = peer_id;
					m_env.addPlayer(player);
					dout_client<<DTIME<<"Client: Adding new player "
							<<peer_id<<std::endl;
				}
				
				player->updateName((char*)&data[start+2]);

				start += item_size;
			}
			
			/*
				Remove those players from the environment that
				weren't listed by the server.
			*/
			//dstream<<DTIME<<"Removing dead players"<<std::endl;
			core::list<Player*> players = m_env.getPlayers();
			core::list<Player*>::Iterator ip;
			for(ip=players.begin(); ip!=players.end(); ip++)
			{
				// Ingore local player
				if((*ip)->isLocal())
					continue;
				
				// Warn about a special case
				if((*ip)->peer_id == 0)
				{
					dstream<<DTIME<<"WARNING: Client: Removing "
							"dead player with id=0"<<std::endl;
				}

				bool is_alive = false;
				core::list<u16>::Iterator i;
				for(i=players_alive.begin(); i!=players_alive.end(); i++)
				{
					if((*ip)->peer_id == *i)
					{
						is_alive = true;
						break;
					}
				}
				/*dstream<<DTIME<<"peer_id="<<((*ip)->peer_id)
						<<" is_alive="<<is_alive<<std::endl;*/
				if(is_alive)
					continue;
				dstream<<DTIME<<"Removing dead player "<<(*ip)->peer_id
						<<std::endl;
				m_env.removePlayer((*ip)->peer_id);
			}
		} //envlock
	}
	else if(command == TOCLIENT_SECTORMETA)
	{
		/*
			[0] u16 command
			[2] u8 sector count
			[3...] v2s16 pos + sector metadata
		*/
		if(datasize < 3)
			return;

		//dstream<<"Client received TOCLIENT_SECTORMETA"<<std::endl;

		{ //envlock
			JMutexAutoLock envlock(m_env_mutex);
			
			std::string datastring((char*)&data[2], datasize-2);
			std::istringstream is(datastring, std::ios_base::binary);

			u8 buf[4];

			is.read((char*)buf, 1);
			u16 sector_count = readU8(buf);
			
			//dstream<<"sector_count="<<sector_count<<std::endl;

			for(u16 i=0; i<sector_count; i++)
			{
				// Read position
				is.read((char*)buf, 4);
				v2s16 pos = readV2S16(buf);
				/*dstream<<"Client: deserializing sector at "
						<<"("<<pos.X<<","<<pos.Y<<")"<<std::endl;*/
				// Create sector
				assert(m_env.getMap().mapType() == MAPTYPE_CLIENT);
				((ClientMap&)m_env.getMap()).deSerializeSector(pos, is);
			}
		} //envlock
	}
	else if(command == TOCLIENT_INVENTORY)
	{
		if(datasize < 3)
			return;

		{ //envlock
			JMutexAutoLock envlock(m_env_mutex);
			
			std::string datastring((char*)&data[2], datasize-2);
			std::istringstream is(datastring, std::ios_base::binary);
			
			m_env.printPlayers(dstream);

			Player *player = m_env.getLocalPlayer();
			assert(player != NULL);

			player->inventory.deSerialize(is);

			dstream<<"Client got player inventory:"<<std::endl;
			player->inventory.print(dstream);
		}
	}
	// Default to queueing it (for slow commands)
	else
	{
		JMutexAutoLock lock(m_incoming_queue_mutex);
		
		IncomingPacket packet(data, datasize);
		m_incoming_queue.push_back(packet);
	}
}

/*
	Returns true if there was something in queue
*/
bool Client::AsyncProcessData()
{
	DSTACK(__FUNCTION_NAME);
	try
	{

	con::Peer *peer;
	{
		JMutexAutoLock lock(m_con_mutex);
		// All data is coming from the server
		peer = m_con.GetPeer(PEER_ID_SERVER);
	}
	
	u8 ser_version = m_server_ser_ver;

	IncomingPacket packet = getPacket();
	u8 *data = packet.m_data;
	u32 datasize = packet.m_datalen;
	
	// An empty packet means queue is empty
	if(data == NULL){
		return false;
	}
	
	if(datasize < 2)
		return true;
	
	ToClientCommand command = (ToClientCommand)readU16(&data[0]);

	if(command == TOCLIENT_BLOCKDATA)
	{
		// Ignore too small packet
		if(datasize < 8)
			return true;
		/*if(datasize < 8 + MapBlock::serializedLength(ser_version))
			goto getdata;*/
			
		v3s16 p;
		p.X = readS16(&data[2]);
		p.Y = readS16(&data[4]);
		p.Z = readS16(&data[6]);
		
		/*dout_client<<DTIME<<"Client: Thread: BLOCKDATA for ("
				<<p.X<<","<<p.Y<<","<<p.Z<<")"<<std::endl;*/

		/*dstream<<DTIME<<"Client: Thread: BLOCKDATA for ("
				<<p.X<<","<<p.Y<<","<<p.Z<<"): ";*/
		
		std::string datastring((char*)&data[8], datasize-8);
		std::istringstream istr(datastring, std::ios_base::binary);
		
		MapSector *sector;
		MapBlock *block;
		
		{ //envlock
			JMutexAutoLock envlock(m_env_mutex);
			
			v2s16 p2d(p.X, p.Z);
			sector = m_env.getMap().emergeSector(p2d);
			
			v2s16 sp = sector->getPos();
			if(sp != p2d)
			{
				dstream<<"ERROR: Got sector with getPos()="
						<<"("<<sp.X<<","<<sp.Y<<"), tried to get"
						<<"("<<p2d.X<<","<<p2d.Y<<")"<<std::endl;
			}

			assert(sp == p2d);
			//assert(sector->getPos() == p2d);
			
			try{
				block = sector->getBlockNoCreate(p.Y);
				/*
					Update an existing block
				*/
				//dstream<<"Updating"<<std::endl;
				block->deSerialize(istr, ser_version);
				//block->setChangedFlag();
			}
			catch(InvalidPositionException &e)
			{
				/*
					Create a new block
				*/
				//dstream<<"Creating new"<<std::endl;
				block = new MapBlock(&m_env.getMap(), p);
				block->deSerialize(istr, ser_version);
				sector->insertBlock(block);
				//block->setChangedFlag();
			}
		} //envlock
		
		
		// Old version has zero lighting, update it.
		if(ser_version == 0 || ser_version == 1)
		{
			derr_client<<"Client: Block in old format: "
					"Calculating lighting"<<std::endl;
			core::map<v3s16, MapBlock*> blocks_changed;
			blocks_changed.insert(block->getPos(), block);
			core::map<v3s16, MapBlock*> modified_blocks;
			m_env.getMap().updateLighting(blocks_changed, modified_blocks);
		}

		/*
			Update Mesh of this block and blocks at x-, y- and z-
		*/

		m_env.getMap().updateMeshes(block->getPos());
		
		/*
			Acknowledge block.
		*/
		/*
			[0] u16 command
			[2] u8 count
			[3] v3s16 pos_0
			[3+6] v3s16 pos_1
			...
		*/
		u32 replysize = 2+1+6;
		SharedBuffer<u8> reply(replysize);
		writeU16(&reply[0], TOSERVER_GOTBLOCKS);
		reply[2] = 1;
		writeV3S16(&reply[3], p);
		// Send as reliable
		m_con.Send(PEER_ID_SERVER, 1, reply, true);

#if 0
		/*
			Remove from history
		*/
		{
			JMutexAutoLock lock(m_fetchblock_mutex);
			
			if(m_fetchblock_history.find(p) != NULL)
			{
				m_fetchblock_history.remove(p);
			}
			else
			{
				/*
					Acknowledge block.
				*/
				/*
					[0] u16 command
					[2] u8 count
					[3] v3s16 pos_0
					[3+6] v3s16 pos_1
					...
				*/
				u32 replysize = 2+1+6;
				SharedBuffer<u8> reply(replysize);
				writeU16(&reply[0], TOSERVER_GOTBLOCKS);
				reply[2] = 1;
				writeV3S16(&reply[3], p);
				// Send as reliable
				m_con.Send(PEER_ID_SERVER, 1, reply, true);
			}
		}
#endif
	}
	else
	{
		dout_client<<DTIME<<"WARNING: Client: Ignoring unknown command "
				<<command<<std::endl;
	}

	return true;

	} //try
	catch(con::PeerNotFoundException &e)
	{
		dout_client<<DTIME<<"Client::AsyncProcessData(): Cancelling: The server"
				" connection doesn't exist (a timeout or not yet connected?)"<<std::endl;
		return false;
	}
}

void Client::Send(u16 channelnum, SharedBuffer<u8> data, bool reliable)
{
	JMutexAutoLock lock(m_con_mutex);
	m_con.Send(PEER_ID_SERVER, channelnum, data, reliable);
}

void Client::fetchBlock(v3s16 p, u8 flags)
{
	if(connectedAndInitialized() == false)
		throw ClientNotReadyException
		("ClientNotReadyException: connectedAndInitialized() == false");

	/*dstream<<"Client::fetchBlock(): Sending GETBLOCK for ("
			<<p.X<<","<<p.Y<<","<<p.Z<<")"<<std::endl;*/

	JMutexAutoLock conlock(m_con_mutex);

	SharedBuffer<u8> data(9);
	writeU16(&data[0], TOSERVER_GETBLOCK);
	writeS16(&data[2], p.X);
	writeS16(&data[4], p.Y);
	writeS16(&data[6], p.Z);
	writeU8(&data[8], flags);
	m_con.Send(PEER_ID_SERVER, 1, data, true);
}

/*
	Calls fetchBlock() on some nearby missing blocks.

	Returns when any of various network load indicators go over limit.

	Does nearly the same thing as the old updateChangedVisibleArea()
*/
void Client::fetchBlocks()
{
	if(connectedAndInitialized() == false)
		throw ClientNotReadyException
		("ClientNotReadyException: connectedAndInitialized() == false");
}

bool Client::isFetchingBlocks()
{
	JMutexAutoLock conlock(m_con_mutex);
	con::Peer *peer = m_con.GetPeerNoEx(PEER_ID_SERVER);
	// Not really fetching but can't fetch more.
	if(peer == NULL) return true;

	con::Channel *channel = &(peer->channels[1]);
	/*
		NOTE: Channel 0 should always be used for fetching blocks,
		      and for nothing else.
	*/
	if(channel->incoming_reliables.size() > 0)
		return true;
	if(channel->outgoing_reliables.size() > 0)
		return true;
	return false;
}

IncomingPacket Client::getPacket()
{
	JMutexAutoLock lock(m_incoming_queue_mutex);
	
	core::list<IncomingPacket>::Iterator i;
	// Refer to first one
	i = m_incoming_queue.begin();

	// If queue is empty, return empty packet
	if(i == m_incoming_queue.end()){
		IncomingPacket packet;
		return packet;
	}
	
	// Pop out first packet and return it
	IncomingPacket packet = *i;
	m_incoming_queue.erase(i);
	return packet;
}

void Client::removeNode(v3s16 nodepos)
{
	if(connectedAndInitialized() == false){
		dout_client<<DTIME<<"Client::removeNode() cancelled (not connected)"
				<<std::endl;
		return;
	}
	
	// Test that the position exists
	try{
		JMutexAutoLock envlock(m_env_mutex);
		m_env.getMap().getNode(nodepos);
	}
	catch(InvalidPositionException &e)
	{
		dout_client<<DTIME<<"Client::removeNode() cancelled (doesn't exist)"
				<<std::endl;
		return;
	}

	SharedBuffer<u8> data(8);
	writeU16(&data[0], TOSERVER_REMOVENODE);
	writeS16(&data[2], nodepos.X);
	writeS16(&data[4], nodepos.Y);
	writeS16(&data[6], nodepos.Z);
	Send(0, data, true);
}

void Client::addNode(v3s16 nodepos, MapNode n)
{
	if(connectedAndInitialized() == false){
		dout_client<<DTIME<<"Client::addNode() cancelled (not connected)"
				<<std::endl;
		return;
	}
	
	// Test that the position exists
	try{
		JMutexAutoLock envlock(m_env_mutex);
		m_env.getMap().getNode(nodepos);
	}
	catch(InvalidPositionException &e)
	{
		dout_client<<DTIME<<"Client::addNode() cancelled (doesn't exist)"
				<<std::endl;
		return;
	}

	//con::Peer *peer = m_con.GetPeer(PEER_ID_SERVER);
	u8 ser_version = m_server_ser_ver;

	u8 datasize = 8 + MapNode::serializedLength(ser_version);
	SharedBuffer<u8> data(datasize);
	writeU16(&data[0], TOSERVER_ADDNODE);
	writeS16(&data[2], nodepos.X);
	writeS16(&data[4], nodepos.Y);
	writeS16(&data[6], nodepos.Z);
	n.serialize(&data[8], ser_version);
	Send(0, data, true);
}

void Client::sendPlayerPos()
{
	JMutexAutoLock envlock(m_env_mutex);
	
	Player *myplayer = m_env.getLocalPlayer();
	if(myplayer == NULL)
		return;
	
	u16 our_peer_id;
	{
		JMutexAutoLock lock(m_con_mutex);
		our_peer_id = m_con.GetPeerID();
	}
	
	// Set peer id if not set already
	if(myplayer->peer_id == PEER_ID_NEW)
		myplayer->peer_id = our_peer_id;
	// Check that an existing peer_id is the same as the connection's
	assert(myplayer->peer_id == our_peer_id);
	
	v3f pf = myplayer->getPosition();
	v3s32 position(pf.X*100, pf.Y*100, pf.Z*100);
	v3f sf = myplayer->getSpeed();
	v3s32 speed(sf.X*100, sf.Y*100, sf.Z*100);
	s32 pitch = myplayer->getPitch() * 100;
	s32 yaw = myplayer->getYaw() * 100;

	/*
		Format:
		[0] u8 command
		[2] v3s32 position*100
		[2+12] v3s32 speed*100
		[2+12+12] s32 pitch*100
		[2+12+12+4] s32 yaw*100
	*/

	SharedBuffer<u8> data(2+12+12+4+4);
	writeU16(&data[0], TOSERVER_PLAYERPOS);
	writeV3S32(&data[2], position);
	writeV3S32(&data[2+12], speed);
	writeS32(&data[2+12+12], pitch);
	writeS32(&data[2+12+12+4], yaw);

	// Send as unreliable
	Send(0, data, false);
}

/*void Client::sendPlayerInfo()
{
	JMutexAutoLock envlock(m_env_mutex);
	
	Player *myplayer = m_env.getLocalPlayer();
	if(myplayer == NULL)
		return;
	
	SharedBuffer<u8> data(2+PLAYERNAME_SIZE);
	writeU16(&data[0], TOSERVER_PLAYERINFO);
	snprintf((char*)&data[2], PLAYERNAME_SIZE, "%s", myplayer->getName());
	// Send as reliable
	Send(0, data, true);
}*/

void Client::updateCamera(v3f pos, v3f dir)
{
	m_env.getMap().updateCamera(pos, dir);
	camera_position = pos;
	camera_direction = dir;
}

MapNode Client::getNode(v3s16 p)
{
	JMutexAutoLock envlock(m_env_mutex);
	return m_env.getMap().getNode(p);
}

/*f32 Client::getGroundHeight(v2s16 p)
{
	JMutexAutoLock envlock(m_env_mutex);
	return m_env.getMap().getGroundHeight(p);
}*/

bool Client::isNodeUnderground(v3s16 p)
{
	JMutexAutoLock envlock(m_env_mutex);
	return m_env.getMap().isNodeUnderground(p);
}

Player * Client::getLocalPlayer()
{
	JMutexAutoLock envlock(m_env_mutex);
	return m_env.getLocalPlayer();
}

core::list<Player*> Client::getPlayers()
{
	JMutexAutoLock envlock(m_env_mutex);
	return m_env.getPlayers();
}

void Client::printDebugInfo(std::ostream &os)
{
	//JMutexAutoLock lock1(m_fetchblock_mutex);
	JMutexAutoLock lock2(m_incoming_queue_mutex);

	os<<"m_incoming_queue.getSize()="<<m_incoming_queue.getSize()
		//<<", m_fetchblock_history.size()="<<m_fetchblock_history.size()
		//<<", m_opt_not_found_history.size()="<<m_opt_not_found_history.size()
		<<std::endl;
}
	

